/*hdr
**	FILE:		24bit.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		21/3/97
**	DESCRIPTION:	24 bit primitives
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"

/****************************** Classes *************************************************************************************/
#define GetRGB(c)	uchar R = R24(c);	uchar G = G24(c);	uchar B = B24(c);

bool GdcApp24::SetSurface(GdcSurface *d, GdcPalette *p, GdcSurface *a)
{
	if (d AND d->Bits == 24)
	{
		Dest = d;
		Pal = p;
		Ptr = d->Base;
		Alpha = 0;
		return TRUE;
	}
	return FALSE;
}

void GdcApp24::SetPtr(int x, int y)
{
	ASSERT(Dest AND Dest->Base);
	Ptr = Dest->Base + ((y * Dest->Line) + (x * 3));
}

void GdcApp24::IncX()
{
	Ptr += 3;
}

void GdcApp24::IncY()
{
	Ptr += Dest->Line;
}

void GdcApp24::IncPtr(int X, int Y)
{
	Ptr += (Y * Dest->Line) + (X * 3);
}

COLOUR GdcApp24::Get()
{
	return Rgb24(Ptr[2], Ptr[1], Ptr[0]);
}

// 24 bit set sub functions
void GdcApp24Set::Set()
{
	GetRGB(c);
	Ptr[0] = B;
	Ptr[1] = G;
	Ptr[2] = R;
}

void GdcApp24Set::VLine(int height)
{
	GetRGB(c);
	while (height--)
	{
		Ptr[0] = B;
		Ptr[1] = G;
		Ptr[2] = R;
		Ptr += Dest->Line;
	}
}

void GdcApp24Set::Rectangle(int x, int y)
{
	GetRGB(c);
	while (y--)
	{
		for (int n=0; n<x; n++)
		{
			*Ptr++ = B;
			*Ptr++ = G;
			*Ptr++ = R;
		}
		Ptr += Dest->Line - (x * 3);
	}
}

bool GdcApp24Set::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src)
	{
		switch (Src->Bits)
		{
			case 8:
			{
				uchar *s = Src->Base;
				if (SPal)
				{
					COLOUR c[256];

					for (int i=0; i<256; i++)
					{
						GdcRGB *p = (*SPal)[i];
						if (p)
						{
							c[i] = (p->R << 16) | (p->G << 8) | (p->B);
						}
						else
						{
							c[i] = 0;
						}
					}

					for (int y=0; y<Src->y; y++)
					{
						for (int x=0; x<Src->x-1; x++)
						{
							*((COLOUR*)Ptr) = c[*s++];
							Ptr += 3;
						}

						*Ptr++ = R24(c[*s]);
						*Ptr++ = G24(c[*s]);
						*Ptr++ = B24(c[*s++]);

						s += Src->Line - Src->x;
						Ptr += Dest->Line - (Src->x * 3);
					}
				}
				else
				{
					for (int y=0; y<Src->y; y++)
					{
						for (int x=0; x<Src->x; x++)
						{
							*Ptr++ = *s;
							*Ptr++ = *s;
							*Ptr++ = *s++;
						}
						s += Src->Line - Src->x;
						Ptr += Dest->Line - (Src->x * 3);
					}
				}
				break;
			}
			case 16:
			{
				for (int y=0; y<Src->y; y++)
				{
					ushort *s = (ushort*) (Src->Base + (Src->Line * y));
					uchar *d = Ptr;

					for (int y=0; y<Src->x; y++)
					{
						d[0] = ((*s & 0x001F) << 3) | ((*s & 0x001C) >> 2);
						d[1] = ((*s & 0x07E0) >> 3) | ((*s & 0x0600) >> 9);
						d[2] = ((*s & 0xF800) >> 8) | ((*s & 0xE000) >> 13);
												
						s++;
						d += 3;
					}

					Ptr += Dest->Line;
				}
				break;
			}
			case 24:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					MemCpy(Ptr, s, Src->x * 3);
					s += Src->Line;
					Ptr += Dest->Line;
				}
				break;
			}
			case 32:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					uchar *NextS = s + Src->Line;
					uchar *NextD = Ptr + Dest->Line;

					for (int x=0; x<Src->x; x++)
					{
						*Ptr++ = *s++;
						*Ptr++ = *s++;
						*Ptr++ = *s++;
						s++;
					}

					s = NextS;
					Ptr = NextD;
				}
				break;
			}
		}
	}
	return TRUE;
}

// 24 bit or sub functions
void GdcApp24Or::Set()
{
	GetRGB(c);
	Ptr[0] |= B;
	Ptr[1] |= G;
	Ptr[2] |= R;
}

void GdcApp24Or::VLine(int height)
{
	GetRGB(c);
	while (height--)
	{
		Ptr[0] |= B;
		Ptr[1] |= G;
		Ptr[2] |= R;
		Ptr += Dest->Line;
	}
}

void GdcApp24Or::Rectangle(int x, int y)
{
	GetRGB(c);
	while (y--)
	{
		for (int n=0; n<x; n++)
		{
			*Ptr++ |= B;
			*Ptr++ |= G;
			*Ptr++ |= R;
		}
		Ptr += Dest->Line - (x * 3);
	}
}

bool GdcApp24Or::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src)
	{
		switch (Src->Bits)
		{
			case 24:
			{
				uchar *s = Src->Base;
				for (int y=0; y<Src->y; y++)
				{
					MemOr(Ptr, s, Src->x * 3);
					s += Src->Line;
					Ptr += Dest->Line;
				}
				break;
			}
		}
	}
	return TRUE;
}

// 24 bit AND sub functions
void GdcApp24And::Set()
{
	GetRGB(c);
	Ptr[0] &= B;
	Ptr[1] &= G;
	Ptr[2] &= R;
}

void GdcApp24And::VLine(int height)
{
	GetRGB(c);
	while (height--)
	{
		Ptr[0] &= B;
		Ptr[1] &= G;
		Ptr[2] &= R;
		Ptr += Dest->Line;
	}
}

void GdcApp24And::Rectangle(int x, int y)
{
	GetRGB(c);
	while (y--)
	{
		for (int n=0; n<x; n++)
		{
			*Ptr++ &= B;
			*Ptr++ &= G;
			*Ptr++ &= R;
		}
		Ptr += Dest->Line - (x * 3);
	}
}

bool GdcApp24And::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemAnd(Ptr, s, Src->x * 3);
			s += Src->Line;
			Ptr += Dest->Line;
		}
	}
	return TRUE;
}

// 24 bit XOR sub functions
void GdcApp24Xor::Set()
{
	GetRGB(c);
	Ptr[0] ^= B;
	Ptr[1] ^= G;
	Ptr[2] ^= R;
}

void GdcApp24Xor::VLine(int height)
{
	GetRGB(c);
	while (height--)
	{
		Ptr[0] ^= B;
		Ptr[1] ^= G;
		Ptr[2] ^= R;
		Ptr += Dest->Line;
	}
}

void GdcApp24Xor::Rectangle(int x, int y)
{
	GetRGB(c);
	while (y--)
	{
		for (int n=0; n<x; n++)
		{
			*Ptr++ ^= B;
			*Ptr++ ^= G;
			*Ptr++ ^= R;
		}
		Ptr += Dest->Line - (x * 3);
	}
}

bool GdcApp24Xor::Blt(GdcSurface *Src, GdcPalette *SPal, GdcSurface *SrcAlpha)
{
	if (Src AND Src->Bits == Dest->Bits)
	{
		uchar *s = Src->Base;
		for (int y=0; y<Src->y; y++)
		{
			MemXor(Ptr, s, Src->x * 3);
			s += Src->Line;
			Ptr += Dest->Line;
		}
	}
	return TRUE;
}
