/*hdr
**	FILE:		Gdc2.h
**	AUTHOR:		Matthew Allen
**	DATE:		20/2/97
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include <math.h>

#include "defs.h"
#include "file.h"
#include "memdev.h"

#include "gdc2.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////
GdcViewDC::GdcViewDC(BView *view)
{
	View = view;
	ASSERT(View);
	BitDepth = GdcD->GetBits();
}

GdcViewDC::~GdcViewDC()
{
	View = 0;
}

void GdcViewDC::GetOrigin(int &x, int &y)
{
	BPoint p = View->Origin();
	x = p.x;
	y = p.y;
}

void GdcViewDC::SetOrigin(int x, int y)
{
	View->SetOrigin(-x, -y);
}

GdcRegion GdcViewDC::ClipRgn()
{
	BRegion r;
	View->GetClippingRegion(&r);
	BRect rc = r.Frame();
	return GdcRegion(&rc);
}

GdcRegion GdcViewDC::ClipRgn(GdcRegion *Rgn)
{
	if (Rgn)
	{
		BRegion r;
		r.Set(*Rgn);
		View->ConstrainClippingRegion(&r);
	}
	else
	{
		View->ConstrainClippingRegion(NULL);
	}
}

COLOUR GdcViewDC::Colour()
{
	rgb_color Rgb = View->HighColor();
	return CBit(GetBits(), Rgb24(Rgb.red, Rgb.green, Rgb.blue), 24);
}

COLOUR GdcViewDC::Colour(COLOUR c, int Bits)
{
	COLOUR Prev = Colour();

	if (Bits < 1) Bits = GetBits();
	c = CBit(24, c, Bits);
	
	rgb_color Rgb;
	Rgb.red = R24(c);
	Rgb.green = G24(c);
	Rgb.blue = B24(c);
	View->SetHighColor(Rgb);
	
	return Prev;
}

int GdcViewDC::Op()
{
	int Mode = View->DrawingMode();
	switch (Mode)
	{
		case B_OP_ERASE:
		{
			return GDC_AND;
			break;
		}
		case B_OP_OVER:
		{
			return GDC_OR;
			break;
		}
		case B_OP_INVERT:
		{
			return GDC_XOR;
			break;
		}
	}
	return GDC_SET;
}

int GdcViewDC::Op(int NewOp)
{
	int Prev = Op();
	drawing_mode Mode = B_OP_COPY;
	switch (NewOp)
	{
		case GDC_AND:
		{
			Mode = B_OP_ERASE;
			break;
		}
		case GDC_OR:
		{
			Mode = B_OP_OVER;
			break;
		}
		case GDC_XOR:
		{
			Mode = B_OP_INVERT;
			break;
		}
	}
	View->SetDrawingMode(Mode);
	return Prev;
}

int GdcViewDC::X()
{
	BRect r = View->Bounds();
	return r.right - r.left + 1;
}

int GdcViewDC::Y()
{
	BRect r = View->Bounds();
	return r.bottom - r.top + 1;
}

int GdcViewDC::GetBits()
{
	return BitDepth;
}

int GdcViewDC::GetFlags()
{
	return Flags;
}

void GdcViewDC::Set(int x, int y)
{
	View->StrokeLine(BPoint(x, y), BPoint(x, y));
}

COLOUR GdcViewDC::Get(int x, int y)
{
	// you can't get the value of a pixel!
	// fix this Be!!!!!!!!!!
	return 0;
}

void GdcViewDC::HLine(int x1, int x2, int y)
{
	View->StrokeLine(BPoint(x1, y), BPoint(x2, y));
}

void GdcViewDC::VLine(int x, int y1, int y2)
{
	View->StrokeLine(BPoint(x, y1), BPoint(x, y2));
}

void GdcViewDC::Line(int x1, int y1, int x2, int y2)
{
	View->StrokeLine(BPoint(x1, y1), BPoint(x2, y2));
}

void GdcViewDC::Circle(double cx, double cy, double radius)
{
	View->StrokeEllipse(BPoint(cx, cy), radius, radius);
}

void GdcViewDC::FilledCircle(double cx, double cy, double radius)
{
	View->FillEllipse(BPoint(cx, cy), radius, radius);
}

void GdcViewDC::Arc(double cx, double cy, double radius, double start, double end)
{
	View->StrokeArc(BPoint(cx, cy), radius, radius, start, end - start);
}

void GdcViewDC::FilledArc(double cx, double cy, double radius, double start, double end)
{
	View->FillArc(BPoint(cx, cy), radius, radius, start, end - start);
}

void GdcViewDC::Ellipse(double cx, double cy, double x, double y)
{
	View->StrokeEllipse(BPoint(cx, cy), x, y);
}

void GdcViewDC::FilledEllipse(double cx, double cy, double x, double y)
{
	View->FillEllipse(BPoint(cx, cy), x, y);
}

void GdcViewDC::Box(int x1, int y1, int x2, int y2)
{
	View->StrokeRect(BRect(x1, y1, x2, y2));
}

void GdcViewDC::Box(GdcRegion *a)
{
	BRect r;
	if (a)
	{
		r = *a;
	}
	else
	{
		r.left = r.top = 0;
		r.right = X()-1;
		r.bottom = Y()-1;
	}
	
	View->StrokeRect(r);
}

void GdcViewDC::Rectangle(int x1, int y1, int x2, int y2)
{
	View->FillRect(BRect(x1, y1, x2, y2));
}

void GdcViewDC::Rectangle(GdcRegion *a)
{
	BRect r;
	if (a)
	{
		r = *a;
	}
	else
	{
		r.left = r.top = 0;
		r.right = X()-1;
		r.bottom = Y()-1;
	}
	
	View->FillRect(r);
}

void GdcViewDC::Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
			// screen to screen Blt
		}
		else
		{
			// memory to screen Blt
			BBitmap *Bmp = Src->GetBitmap();
			if (Bmp)
			{
				BRect S;
				if (a)
				{
					S = *a;
				}
				else
				{
					S.left = S.top = 0;
					S.right = Src->X() - 1;
					S.bottom = Src->Y() - 1;
				}
				
				BRect D(x, y, x+Src->X()-1, y+Src->Y()-1);
				
				View->DrawBitmap(Bmp, S, D);
			}
		}
	}
}

void GdcViewDC::StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s)
{
}

void GdcViewDC::Polygon(int Points, GdcPt2 *Data)
{
	if (Points > 0 AND Data)
	{
		BPoint *p = NEW(BPoint[Points]);
		if (p)
		{
			for (int i=0; i<Points; i++)
			{
				p[i].x = Data[i].x;
				p[i].y = Data[i].y;
			}
			
			View->StrokePolygon(p, Points);
			DeleteArray(p);
		}
	}
}

void GdcViewDC::Text(GdcFont *pFont, int x, int y, char *Str, int Len)
{
}

void GdcViewDC::Bezier(int Threshold, GdcPt2 *Pt)
{
}

void GdcViewDC::FloodFill(int x, int y, int Mode, COLOUR Border, GdcRegion *Bounds)
{
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
GdcMemDC::GdcMemDC()
{
	Bmp = 0;
}

GdcMemDC::GdcMemDC(GdcBasePrimitives *pDC)
{
	Bmp = 0;

	if (pDC AND Create(pDC->X(), pDC->Y(), pDC->GetBits()))
	{
		Blt(0, 0, pDC);
		if (pDC->Palette())
		{
			GdcPalette *Pal = NEW(GdcPalette(pDC->Palette()));
			if (Pal)
			{
				Palette(Pal, TRUE);
			}
		}
	}
}

GdcMemDC::~GdcMemDC()
{
	DeleteObj(Bmp);
}

bool GdcMemDC::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	bool Status = FALSE;
	color_space Mode = B_RGB24;
	switch (Bits)
	{
		case 8:
		{
			Mode = B_CMAP8;
			break;
		}
		case 15:
		case 16:
		{
			Mode = B_RGB16;
			break;
		}
		case 24:
		{
			Bits = 32;
			Mode = B_RGB32;
			break;
		}
		case 32:
		{
			Mode = B_RGB32;
			break;
		}
	}

	DeleteObj(Bmp);
	Bmp = NEW(BBitmap(BRect(0, 0, x-1, y-1), Mode, FALSE, TRUE));
	if (Bmp)
	{
		pMem = NEW(GdcSurface);
		if (pMem)
		{
			pMem->Base = (uchar*) Bmp->Bits();
			pMem->x = x;
			pMem->y = y;
			pMem->Bits = Bits;
			pMem->Line = Bmp->BytesPerRow();
			pMem->Flags = 0;
			
			// Msg("Base: %p x: %i y: %i Bits: %i Line: %i", pMem->Base, pMem->x, pMem->y, pMem->Bits, pMem->Line);
			Status = TRUE;

			int NewOp = (pApp) ? Op() : GDC_SET;

			if (	(Flags & GDC_OWN_APPLICATOR) AND
				NOT (Flags & GDC_CACHED_APPLICATOR))
			{
				DeleteObj(pApp);
			}

			for (int i=0; i<GDC_CACHE_SIZE; i++)
			{
				DeleteObj(pAppCache[i]);
			}

			if (NewOp < GDC_CACHE_SIZE AND NOT DrawOnAlpha())
			{
				pApp = (pAppCache[NewOp]) ? pAppCache[NewOp] : pAppCache[NewOp] = CreateApplicator(NewOp);
				Flags &= ~GDC_OWN_APPLICATOR;
				Flags |= GDC_CACHED_APPLICATOR;
			}
			else
			{
				pApp = CreateApplicator(NewOp);
				Flags &= ~GDC_CACHED_APPLICATOR;
				Flags |= GDC_OWN_APPLICATOR;
			}

			Clip.ZOff(X()-1, Y()-1);
		}
	}

	return Status;
}

void GdcMemDC::Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
			// screen to memory Blt
		}
		else
		{
			// memory to memory Blt
			GdcBasePrimitives::Blt(x, y, Src, a);
		}
	}
}

void GdcMemDC::StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s)
{
}
