/*hdr
**	FILE:		GdcPrim.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		1/3/97
**	DESCRIPTION:	GDC v2.xx device independent primitives
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/****************************** Includes ************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"

/****************************** Defines *************************************************************************************/

#define LINE_SOLID				0xFFFFFFFF

/****************************** Asm *****************************************************************************************/


/****************************** Functions ***********************************************************************************/

GdcBasePrimitives::GdcBasePrimitives()
{
	LineBits = LINE_SOLID;
	LineMask = 1;
}

GdcBasePrimitives::GdcBasePrimitives(GdcBasePrimitives *pDC)
{
	LineBits = LINE_SOLID;
	LineMask = 1;

	if (pDC AND Create(pDC->X(), pDC->Y(), pDC->GetBits()))
	{
		Blt(0, 0, pDC);
		if (pDC->Palette())
		{
			GdcPalette *Pal = NEW(GdcPalette(pDC->Palette()));
			if (Pal)
			{
				Palette(Pal, TRUE);
			}
		}
	}
}

void GdcBasePrimitives::Set(int x, int y)
{
	#ifdef ADJUST_ORIGIN
	OrgXy(x, y);
	#endif

	if (	x >= Clip.x1 AND
		y >= Clip.y1 AND
		x <= Clip.x2 AND
		y <= Clip.y2)
	{
		pApp->SetPtr(x, y);
		pApp->Set();
		Update(GDC_BITS_CHANGE);
	}
}

COLOUR GdcBasePrimitives::Get(int x, int y)
{
	#ifdef ADJUST_ORIGIN
	OrgXy(x, y);
	#endif

	if (	x >= Clip.x1 AND
		y >= Clip.y1 AND
		x <= Clip.x2 AND
		y <= Clip.y2)
	{
		pApp->SetPtr(x, y);
		return pApp->Get();
	}
	return 0;
}

void GdcBasePrimitives::HLine(int x1, int x2, int y)
{
	#ifdef ADJUST_ORIGIN
	OrgXy(x1, y);
	OrgX(x2);
	#endif

	if (x1 > x2) swap(x1, x2);

	if (x1 < Clip.x1) x1 = Clip.x1;
	if (x2 > Clip.x2) x2 = Clip.x2;
	if (	x1 <= x2 AND
		y >= Clip.y1 AND
		y <= Clip.y2)
	{
		pApp->SetPtr(x1, y);
		if (LineBits == LINE_SOLID)
		{
			pApp->Rectangle(x2 - x1 + 1, 1);
			Update(GDC_BITS_CHANGE);
		}
		else
		{
			for (; x1 <= x2; x1++)
			{
				if (LineMask & LineBits)
				{
					pApp->Set();
				}

				LineMask >>= 1;
				if (NOT LineMask) LineMask = 0x80000000;
				pApp->IncX();
			}
			Update(GDC_BITS_CHANGE);
		}
	}
}

void GdcBasePrimitives::VLine(int x, int y1, int y2)
{
	#ifdef ADJUST_ORIGIN
	OrgXy(x, y1);
	OrgY(y2);
	#endif

	if (y1 > y2) swap(y1, y2);
	if (y1 < Clip.y1) y1 = Clip.y1;
	if (y2 > Clip.y2) y2 = Clip.y2;
	if (	y1 <= y2 AND
		x >= Clip.x1 AND
		x <= Clip.x2)
	{
		pApp->SetPtr(x, y1);
		if (LineBits == LINE_SOLID)
		{
			pApp->VLine(y2 - y1 + 1);
			Update(GDC_BITS_CHANGE);
		}
		else
		{
			for (; y1 <= y2; y1++)
			{
				if (LineMask & LineBits)
				{
					pApp->Set();
				}

				LineMask >>= 1;
				if (NOT LineMask) LineMask = 0x80000000;
				pApp->IncY();
			}
			Update(GDC_BITS_CHANGE);
		}
	}
}

void GdcBasePrimitives::Line(int x1, int y1, int x2, int y2)
{
	if (x1 == x2)
	{
		VLine(x1, y1, y2);
	}
	else if (y1 == y2)
	{
		HLine(x1, x2, y1);
	}
	else if (Clip.Valid())
	{
		#ifdef ADJUST_ORIGIN
		OrgXy(x1, y1);
		OrgXy(x2, y2);
		#endif

		// angled
		if (y1 > y2)
		{
			swap(y1, y2);
			swap(x1, x2);
		}

		GdcRegion Bound(x1, y1, x2, y2);
		
		Bound.Normal();
		if (NOT Bound.Overlap(&Clip)) return;

		double m = (double) (y2-y1) / (x2-x1);
		double b = (double) y1 - (m*x1);

		int xt = (int) (((double)Clip.y1-b)/m);
		int xb = (int) (((double)Clip.y2-b)/m);

		if (y1 < Clip.y1)
		{
			x1 = xt;
			y1 = Clip.y1;
		}

		if (y2 > Clip.y2)
		{
			x2 = xb;
			y2 = Clip.y2;
		}

		int X1, X2;
		if (x1 < x2) { X1 = x1; X2 = x2; } else { X1 = x2; X2 = x1; }

		if (X1 < Clip.x1)
		{
			if (X2 < Clip.x1) return;

			if (x1 < Clip.x1)
			{
				y1 = (int) (((double) Clip.x1 * m) + b);
				x1 = Clip.x1;
			}
			else
			{
				y2 = (int) (((double) Clip.x1 * m) + b);
				x2 = Clip.x1;
			}
		}
		
		if (X2 > Clip.x2)
		{
			if (X1 > Clip.x2) return;

			if (x1 > Clip.x2)
			{
				y1 = ((double) Clip.x2 * m) + b;
				x1 = Clip.x2;
			}
			else
			{
				y2 = ((double) Clip.x2 * m) + b;
				x2 = Clip.x2;
			}
		}			
		
		int dx = abs(x2-x1);
		int dy = abs(y2-y1);
		int EInc, ENoInc, E, Inc = 1, Pos = 1;
	
		// printf("Dx: %i\nDy: %i\n", dx, dy);

		if (dy < dx)
		{
			// flat
			if (x1 > x2)
			{
				swap(x1, x2);
				swap(y1, y2);
			}

			if (y1 > y2)
			{
				Inc = -1;
			}

			EInc = dy + dy;
			E = EInc - dx;
			ENoInc = E - dx;

			pApp->SetPtr(x1, y1);
			if (LineBits == LINE_SOLID)
			{
				for (; x1<=x2; x1++)
				{
					pApp->Set();
					if (E < 0)
					{
						pApp->IncX();
						E += EInc;
					}
					else
					{
						pApp->IncPtr(1, Inc);
						E += ENoInc;
					}
				}
			}
			else
			{
				for (; x1<=x2; x1++)
				{
					if (LineBits & LineMask)
					{
						pApp->Set();
					}
					LineMask >>= 1;
					if (NOT LineMask) LineMask = 0x80000000;

					if (E < 0)
					{
						pApp->IncX();
						E += EInc;
					}
					else
					{
						pApp->IncPtr(1, Inc);
						E += ENoInc;
					}
				}
			}
		}
		else
		{
			if (x1 > x2)
			{
				Inc = -1;
			}
		
			// steep
			EInc = dx + dx;
			E = EInc - dy;
			ENoInc = E - dy;
	
			pApp->SetPtr(x1, y1);
			if (LineBits == LINE_SOLID)
			{
				for (; y1<=y2; y1++)
				{
					pApp->Set();
					if (E < 0)
					{
						pApp->IncY();
						E += EInc;
					}
					else
					{
						pApp->IncPtr(Inc, 1);
						E += ENoInc;
					}
				}
			}
			else
			{
				for (; y1<=y2; y1++)
				{
					if (LineBits & LineMask)
					{
						pApp->Set();
					}
					LineMask >>= 1;
					if (NOT LineMask) LineMask = 0x80000000;

					if (E < 0)
					{
						pApp->IncY();
						E += EInc;
					}
					else
					{
						pApp->IncPtr(Inc, 1);
						E += ENoInc;
					}
				}
			}
		}

		Update(GDC_BITS_CHANGE);
	}
}

void GdcBasePrimitives::Circle(double cx, double cy, double radius)
{
	int d = 3 - (2 * radius);
	int x = 0;
	int y = radius;

	Set(cx, cy + y);
	Set(cx, cy - y);
	Set(cx + y, cy);
	Set(cx - y, cy);

	if (d < 0)
	{
		d += (4 * x) + 6;
	}
	else
	{
		d += (4 * (x - y)) + 10;
		y--;
	}

	x++;

	while (x < y)
	{
		Set(cx + x, cy + y);
		Set(cx - x, cy + y);
		Set(cx + x, cy - y);
		Set(cx - x, cy - y);
		Set(cx + y, cy + x);
		Set(cx - y, cy + x);
		Set(cx + y, cy - x);
		Set(cx - y, cy - x);

		if (d < 0)
		{
			d += (4 * x) + 6;
		}
		else
		{
			d += (4 * (x - y)) + 10;
			y--;
		}

		x++;
	}

	if (x == y)
	{
		Set(cx + x, cy + x);
		Set(cx - x, cy + x);
		Set(cx + x, cy - x);
		Set(cx - x, cy - x);
	}

	Update(GDC_BITS_CHANGE);
}

void GdcBasePrimitives::FilledCircle(double cx, double cy, double radius)
{
	int d = 3 - 2 * radius;
	int x = 0;
	int y = radius;

	// HLine(cx + y, cx - y, cy);
	if (d < 0)
	{
		d += 4 * x + 6;
	}
	else
	{
		HLine(cx, cx, cy + y);
		d += 4 * (x - y) + 10;
		y--;
	}

	while (x < y)
	{
		HLine(cx + y, cx - y, cy + x);
		if (x != 0) HLine(cx + y, cx - y, cy - x);

		if (d < 0)
		{
			d += 4 * x + 6;
		}
		else
		{
			HLine(cx + x, cx - x, cy + y);
			HLine(cx + x, cx - x, cy - y);

			d += 4 * (x - y) + 10;
			y--;
		}

		x++;
	}

	if (x == y)
	{
		HLine(cx + y, cx - y, cy + x);
		HLine(cx + y, cx - y, cy - x);
	}

	Update(GDC_BITS_CHANGE);
}

void GdcBasePrimitives::Box(GdcRegion *a)
{
	if (a)
	{
		GdcRegion b = *a;
		b.Normal();
		if (b.x1 != b.x2 AND b.y1 != b.y2)
		{
			HLine(b.x1, b.x2 - 1, b.y1);
			VLine(b.x2, b.y1, b.y2 - 1);
			HLine(b.x1 + 1, b.x2, b.y2);
			VLine(b.x1, b.y1 + 1, b.y2);
		}
		else
		{
			Set(b.x1, b.y1);
		}
	}
	else
	{
		GdcRegion b(0, 0, X()-1, Y()-1);

		HLine(b.x1, b.x2 - 1, b.y1);
		VLine(b.x2, b.y1, b.y2 - 1);
		HLine(b.x1 + 1, b.x2, b.y2);
		VLine(b.x1, b.y1 + 1, b.y2);
	}
}

void GdcBasePrimitives::Box(int x1, int y1, int x2, int y2)
{
	GdcRegion a(x1, y1, x2, y2);
	Box(&a);
}

void GdcBasePrimitives::Rectangle(GdcRegion *a)
{
	if (a)
	{
		GdcRegion b = a;

		#ifdef ADJUST_ORIGIN
		OrgRgn(b);
		#endif
		b.Normal();
		b.Bound(&Clip);

		if (b.Valid())
		{
			pApp->SetPtr(b.x1, b.y1);
			pApp->Rectangle(b.X(), b.Y());
			Update(GDC_BITS_CHANGE);
		}
	}
	else
	{
		pApp->SetPtr(0, 0);
		pApp->Rectangle(X(), Y());
		Update(GDC_BITS_CHANGE);
	}
}

void GdcBasePrimitives::Rectangle(int x1, int y1, int x2, int y2)
{
	GdcRegion a(x1, y1, x2, y2);
	Rectangle(&a);
}

void GdcBasePrimitives::Ellipse(double cx, double cy, double a, double b)
{
	// TODO: fix this primitive for odd widths and heights
	a = floor(a);
	b = floor(b);
	
	long aSq = a*a;
	long bSq = b*b;
	long two_aSq = aSq+aSq;
	long two_bSq = bSq+bSq;
	long x=0, y=b, two_xBsq = 0, two_yAsq = y * two_aSq, error = -y * aSq;

	if (aSq AND bSq AND error)
	{
		while (two_xBsq <= two_yAsq)
		{
			if (x) Set(cx+x, cy+y);
			if (y) Set(cx-x, cy+y);
			if (x) Set(cx+x, cy-y);
			if (y) Set(cx-x, cy-y);
			x++;
	
			two_xBsq += two_bSq;
			error += two_xBsq - bSq;
			if (error >= 0)
			{
				y--;
				two_yAsq -= two_aSq;
				error -= two_yAsq;
			}
		}
	
		x=a; y=0; two_xBsq = x * two_bSq; two_yAsq = 0; error = -x * bSq;
	
		while (two_xBsq >= two_yAsq)
		{
			if (x) Set(cx+x, cy+y);
			if (x) Set(cx-x, cy+y);
			if (y) Set(cx+x, cy-y);
			if (y) Set(cx-x, cy-y);
			y++;
	
			two_yAsq += two_aSq;
			error += two_yAsq - aSq;
			if (error >= 0)
			{
				x--;
				two_xBsq -= two_bSq;
				error -= two_xBsq;
			}
		}
	}
	Update(GDC_BITS_CHANGE);
}

void GdcBasePrimitives::FilledEllipse(double cx, double cy, double a, double b)
{
	// TODO: fix this primitive for odd widths and heights
	a = floor(a);
	b = floor(b);

	long aSq = a*a;
	long bSq = b*b;
	long two_aSq = aSq+aSq;
	long two_bSq = bSq+bSq;
	long x=0, y=b, two_xBsq = 0, two_yAsq = y * two_aSq, error = -y * aSq;

	if (aSq AND bSq AND error)
	{
		while (two_xBsq <= two_yAsq)
		{
			x++;
			two_xBsq += two_bSq;
			error += two_xBsq - bSq;
			if (error >= 0)
			{
				Line(cx+x-1, cy+y, cx-x+1, cy+y);
				if (y != 0) Line(cx+x-1, cy-y, cx-x+1, cy-y);
	
				y--;
				two_yAsq -= two_aSq;
				error -= two_yAsq;
			}
		}
	
		x=a;
		y=0;
		two_xBsq = x * two_bSq;
		two_yAsq = 0;
		error = -x * bSq;
	
		while (two_xBsq > two_yAsq)
		{
			Line(cx+x, cy+y, cx-x, cy+y);
			if (y != 0) Line(cx+x, cy-y, cx-x, cy-y);
			y++;
	
			two_yAsq += two_aSq;
			error += two_yAsq - aSq;
			if (error >= 0)
			{
				x--;
				two_xBsq -= two_bSq;
				error -= two_xBsq;
			}
		}
	}

	Update(GDC_BITS_CHANGE);
}

struct EDGE {
	int yMin, yMax, x, dWholeX, dX, dY, frac;
};

int nActive, nNextEdge;

void GdcBasePrimitives::Polygon(int nPoints, GdcPt2 *aPoints)
{
	GdcPt2 p0, p1;
	int i, j, gap, x0, x1, y, nEdges;
	EDGE *ET, **GET, **AET;

	/********************************************************************
	 * Add entries to the global edge table.  The global edge table has a
	 * bucket for each scan line in the polygon. Each bucket contains all
	 * the edges whose yMin == yScanline.  Each bucket contains the yMax,
	 * the x coordinate at yMax, and the denominator of the slope (dX)
	 */

	// allocate the tables
	ET = NEW(EDGE[nPoints]);
	GET = NEW(EDGE*[nPoints]);
	AET = NEW(EDGE*[nPoints]);

	if (ET AND GET AND AET)
	{
		for ( i = 0, nEdges = 0; i < nPoints; i++)
		{
			p0 = aPoints[i];
			p1 = aPoints[(i+1) % nPoints];

			// ignore if this is a horizontal edge
			if (p0.y == p1.y) continue;

			// swap points if necessary to ensure p0 contains yMin
			if (p0.y > p1.y)
			{
				p0 = p1;
				p1 = aPoints[i];
			}

			// create the new edge
			ET[nEdges].yMin = p0.y;
			ET[nEdges].yMax = p1.y;
			ET[nEdges].x = p0.x;
			ET[nEdges].dX = p1.x - p0.x;
			ET[nEdges].dY = p1.y - p0.y;
			ET[nEdges].frac = 0;
			GET[nEdges] = &ET[nEdges];
			nEdges++;
		}

		// sort the GET on yMin
		for (gap = 1; gap < nEdges; gap = 3*gap + 1);
		for (gap /= 3; gap > 0; gap /= 3)
			for (i = gap; i < nEdges; i++)
				for (j = i-gap; j >= 0; j -= gap)
				{
					if (GET[j]->yMin <= GET[j+gap]->yMin)
						break;
					EDGE *t = GET[j];
					GET[j] = GET[j+gap];
					GET[j+gap] = t;
				}

		// initialize the active edge table, and set y to first entering edge
		nActive = 0;
		nNextEdge = 0;
		y = GET[nNextEdge]->yMin;

		/* Now process the edges using the scan line algorithm.  Active edges
		will be added to the Active Edge Table (AET), and inactive edges will
		be deleted.  X coordinates will be updated with incremental integer
		arithmetic using the slope (dY / dX) of the edges. */

		while (nNextEdge < nEdges || nActive)
		{
			/* Move from the ET bucket y to the AET those edges whose yMin == y
			(entering edges) */
			while (nNextEdge < nEdges && GET[nNextEdge]->yMin == y)
				AET[nActive++] = GET[nNextEdge++];

			/* Remove from the AET those entries for which yMax == y (leaving
			edges) */
			i = 0;
			while (i < nActive)
			{
				if (AET[i]->yMax == y)
					memcpy(&AET[i], &AET[i+1], sizeof(AET[0]) * (--nActive - i));
				else
					i++;
			}

			/* Now sort the AET on x.  Since the list is usually quite small,
			the sort is implemented as a simple non-recursive shell sort */
			for (gap = 1; gap < nActive; gap = 3*gap + 1);
			for (gap /= 3; gap > 0; gap /= 3)
				for (i = gap; i < nActive; i++)
					for (j = i-gap; j >= 0; j -= gap)
					{
						if (AET[j]->x <= AET[j+gap]->x)
							break;
						EDGE *t = AET[j];
						AET[j] = AET[j+gap];
						AET[j+gap] = t;
					}

			/* Fill in desired pixels values on scan line y by using pairs of x
			coordinates from the AET */

			for (i = 0; i < nActive; i += 2)
			{
				x0 = AET[i]->x;
				x1 = AET[i+1]->x;

				/* Left edge adjustment for positive fraction.  0 is interior. */
				if (AET[i]->frac > 0) x0++;

				// Right edge adjustment for negative fraction.  0 is exterior. */
				if (AET[i+1]->frac <= 0) x1--;

				// Draw interior spans
				if (x1 >= x0)
					HLine(x0, x1, y);
			}

			/* Update all the x coordinates.  Edges are scan converted using a
			modified midpoint algorithm (Bresenham's algorithm reduces to the
			midpoint algorithm for two dimensional lines) */
			for (i = 0; i < nActive; i++)
			{
				EDGE *e = AET[i];
				// update the fraction by dX
				e->frac += e->dX;

				if (e->dX < 0)
					while ( -(e->frac) >= e->dY)
					{
						e->frac += e->dY;
						e->x--;
					}
				else
					while (e->frac >= e->dY)
					{
						e->frac -= e->dY;
						e->x++;
					}
			}
			y++;
		}
	}

	// Release tables
	DeleteArray(ET);
	DeleteArray(GET);
	DeleteArray(AET);
}

void GdcBasePrimitives::Blt(int x, int y, GdcBasePrimitives *Src, GdcRegion *a)
{
	#ifdef ADJUST_ORIGIN
	OrgXy(x, y);
	#endif

	if (Src)
	{
		GdcRegion S;
		if (a)	S = *a;
		else	S.ZOff(Src->X()-1, Src->Y()-1);
		S.Offset(Src->OriginX, Src->OriginY);

		GdcRegion SClip = S;
		SClip.Bound(&Src->Clip);

		if (SClip.Valid())
		{
			GdcRegion D = SClip;
			D.Offset(x-S.x1, y-S.y1);

			GdcRegion DClip = D;
			DClip.Bound(&Clip);

			GdcRegion Re = DClip;
			Re.Offset(S.x1-x, S.y1-y);
			SClip.Bound(&Re);

			if (DClip.Valid() AND SClip.Valid())
			{
				GdcSurface Bits, Alpha;

				Bits.Base =	Src->pMem->Base +
						(SClip.y1 * Src->pMem->Line) +
						(SClip.x1 * Src->pMem->Bits / 8);
				Bits.x = min(SClip.X(), DClip.X());
				Bits.y = min(SClip.Y(), DClip.Y());
				Bits.Line = Src->pMem->Line;
				Bits.Bits = Src->GetBits();
				Bits.Flags = 0;

				if (Src->pAlphaDC)
				{
					GdcSurface *ASurface = Src->pAlphaDC->pMem;
					Alpha = Bits;
					Alpha.Bits = 8;
					Alpha.Line = ASurface->Line;
					Alpha.Base =	ASurface->Base +
							(SClip.y1 * ASurface->Line) +
							(SClip.x1);
				}

				pApp->SetPtr(DClip.x1, DClip.y1);
				pApp->Blt(&Bits, Src->Palette(), (Src->pAlphaDC) ? &Alpha : 0);
				Update(GDC_BITS_CHANGE);

				if (pApp->GetFlags() & GDC_UPDATED_PALETTE)
				{
					Palette(NEW(GdcPalette(pApp->GetPal())));
				}
			}
		}
	}
}

void GdcBasePrimitives::Text(GdcFont *pFont, int x, int y, char *Str, int Len)
{
}

#define sqr(a)			((a)*(a))
#define Distance(a, b)		(sqrt(sqr(a.x-b.x)+sqr(a.y-b.y)))
class FPt2 {
public:
	double x, y;
};

typedef FPt2 BPt;

void GdcBasePrimitives::Bezier(int Threshold, GdcPt2 *Pt)
{
	if (Pt)
	{
		int OldPts = 3;
		int NewPts = 0;
		BPt *BufA = NEW(BPt[1024]);
		BPt *BufB = NEW(BPt[1024]);
		BPt *Old = BufA;
		BPt *New = BufB;

		Threshold = max(Threshold, 1);

		if (NOT Old OR NOT New) return;
		for (int n=0; n<OldPts; n++)
		{
			Old[n].x = Pt[n].x;
			Old[n].y = Pt[n].y;
			#ifdef ADJUST_ORIGIN
			OrgPt(Old[n]);
			#endif
		}

		do
		{
			NewPts = 1;
			New[0] = Old[0];
			for (int i=1; i<OldPts-1; i += 2)
			{
				New[NewPts].x = (Old[i].x + Old[i-1].x) / 2;
				New[NewPts].y = (Old[i].y + Old[i-1].y) / 2;
				NewPts += 2;

				New[NewPts].x = (Old[i].x + Old[i+1].x) / 2;
				New[NewPts].y = (Old[i].y + Old[i+1].y) / 2;
				NewPts--;

				New[NewPts].x = (New[NewPts+1].x + New[NewPts-1].x) / 2;
				New[NewPts].y = (New[NewPts+1].y + New[NewPts-1].y) / 2;
				NewPts += 2;

				New[NewPts++] = Old[i+1];
			}

			BPt *Temp = Old;
			Old = New;
			New = Temp;
			
			OldPts = NewPts;
		}
		while (Distance(Old[0], Old[1]) > Threshold);

		if (Threshold > 1)
		{
			for (int i=0; i<OldPts-1; i += 4)
			{
				Line(	Old[i].x,
					Old[i].y,
					Old[i+1].x,
					Old[i+1].y);
				Line(	Old[i+1].x,
					Old[i+1].y,
					Old[i+3].x,
					Old[i+3].y);
				Line(	Old[i+3].x,
					Old[i+3].y,
					Old[i+4].x,
					Old[i+4].y);
			}
		}
		else
		{
			for (int i=0; i<OldPts; i++)
			{
				Set(	Old[i].x,
					Old[i].y);
			}
		}

		delete [] BufA;
		delete [] BufB;

		Update(GDC_BITS_CHANGE);
	}
}

class PointStack {

	int Used;
	int Size;
	GdcPt2 *Stack;

	bool SetSize(int s)
	{
		GdcPt2 *Next = NEW(GdcPt2[Size + s]);
		if (Next)
		{
			Size += s;
			Used = min(Size, Used);
			memcpy(Next, Stack, sizeof(GdcPt2)*Used);
			DeleteArray(Stack);
			Stack = Next;
			return TRUE;
		}
		return FALSE;
	}

public:
	PointStack()
	{
		Used = 0;
		Size = 1024;
		Stack = NEW(GdcPt2[Size]);
	}

	~PointStack()
	{
		DeleteArray(Stack);
	}

	int GetSize() { return Used; }

	void Push(int x, int y)
	{
		if (Used >= Size)
		{
			SetSize(Size+1024);
		}

		if (Stack)
		{
			Stack[Used].x = x;
			Stack[Used].y = y;
			Used++;
		}
	}

	void Pop(int &x, int &y)
	{
		if (Stack AND Used > 0)
		{
			Used--;
			x = Stack[Used].x;
			y = Stack[Used].y;
		}
	}
};

void GdcBasePrimitives::FloodFill(int StartX, int StartY, int Mode, COLOUR Border, GdcRegion *FillBounds)
{
	COLOUR Seed = Get(StartX, StartY);
	PointStack Ps;
	GdcRegion Bounds;

	Bounds.x1 = X();
	Bounds.y1 = Y();
	Bounds.x2 = 0;
	Bounds.y2 = 0;

	Ps.Push(StartX, StartY);

	switch (Mode)
	{
		case GDC_FILL_TO_DIFFERENT:
		{
			if (Seed != Colour())
			{
				while (Ps.GetSize() > 0)
				{
					bool Above = TRUE;
					bool Below = TRUE;
					int Ox, Oy;
					Ps.Pop(Ox, Oy);
					int x = Ox, y = Oy;

					// move right loop
					COLOUR c = Get(x, y);

					while (x < X() AND c == Seed)
					{
						Set(x, y);
						Bounds.Union(x, y);

						if (y > 0)
						{
							c = Get(x, y - 1);

							if (Above)
							{
								if (c == Seed)
								{
									Ps.Push(x, y - 1);
									Above = FALSE;
								}
							}
							else if (c != Seed)
							{
								Above = TRUE;
							}
						}

						if (y < Y() - 1)
						{
							c = Get(x, y + 1);

							if (Below)
							{
								if (c == Seed)
								{
									Ps.Push(x, y + 1);
									Below = FALSE;
								}
							}
							else if (c != Seed)
							{
								Below = TRUE;
							}
						}

						x++;
						c = Get(x, y);
					}

					// move left loop
					x = Ox;

					Above = NOT ((y > 0) AND (Get(x, y - 1) == Seed));
					Below = NOT ((y < Y() - 1) AND (Get(x, y + 1) == Seed));

					x--;
					c = Get(x, y);

					while (x >= 0 AND c == Seed)
					{
						Set(x, y);
						Bounds.Union(x, y);

						if (y > 0)
						{
							c = Get(x, y - 1);

							if (Above)
							{
								if (c == Seed)
								{
									Ps.Push(x, y - 1);
									Above = FALSE;
								}
							}
							else if (c != Seed)
							{
								Above = TRUE;
							}
						}

						if (y < Y() - 1)
						{
							c = Get(x, y + 1);

							if (Below)
							{
								if (c == Seed)
								{
									Ps.Push(x, y + 1);
									Below = FALSE;
								}
							}
							else if (c != Seed)
							{
								Below = TRUE;
							}
						}

						x--;
						c = Get(x, y);
					}
				}
			}
			break;
		}

		case GDC_FILL_TO_BORDER:
		{
			break;
		}
	}

	if (FillBounds)
	{
		*FillBounds = Bounds;
	}
	Update(GDC_BITS_CHANGE);
}

void GdcBasePrimitives::Arc(double cx, double cy, double radius, double start, double end) {}
void GdcBasePrimitives::FilledArc(double cx, double cy, double radius, double start, double end) {}
void GdcBasePrimitives::StretchBlt(GdcRegion *d, GdcBasePrimitives *Src, GdcRegion *s) {}

