/*hdr
**	FILE:		Intel.h
**	AUTHOR:		Matthew Allen
**	DATE:		11/3/97
**	DESCRIPTION:	Fast core routines for intel based machines
**
**	Copyright (C) 1997, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include "defs.h"
#include "memdev.h"
#include "core.h"

void Add64(ulong *DestH, ulong *DestL, ulong SrcH, ulong SrcL)
{
/*
	asm {
		mov eax, DestL
		mov ebx, SrcL
		mov ecx, SrcH
		mov edx, DestH
		add [eax], ebx
		adc [edx], ecx
	}
*/

	*DestH += SrcH + (*DestL & SrcL & 0x80000000) ? 1 : 0;
	*DestL += SrcL;
}

void MemSet(void *d, int c, uint s)
{
	if (d AND s > 0) memset(d, c, s);
}

void MemCpy(void *d, void *s, uint l)
{
	if (d AND s AND l > 0) memcpy(d, s, l);
}

void MemAnd(void *d, void *s, uint l)
{
	uchar *D = (uchar*) d;
	uchar *S = (uchar*) s;
	if (D AND S)
	{
		for (; l > 0; l--)
		{
			*D++ &= *S++;
		}
	}
}

void MemXor(void *d, void *s, uint l)
{
	uchar *D = (uchar*) d;
	uchar *S = (uchar*) s;
	if (D AND S)
	{
		for (; l > 0; l--)
		{
			*D++ ^= *S++;
		}
	}
}

void MemOr(void *d, void *s, uint l)
{
	uchar *D = (uchar*) d;
	uchar *S = (uchar*) s;
	if (D AND S)
	{
		for (; l > 0; l--)
		{
			*D++ |= *S++;
		}
	}
}

