#ifndef __DEFS_H
#define __DEFS_H

#include <string.h>
#include "assert.h"

#define BEOS					3
#define MAIL_GUI				1

typedef	unsigned char			uchar;
typedef	unsigned short			ushort;
typedef	unsigned int			uint;
typedef	unsigned long			ulong;

#define SUCCESS					1
#define FAIL					0
#define TRUE					1
#define FALSE					0
#define YES						1
#define NO						0

#define abs(a)					(((a) > 0) ? (a) : -(a))
#define min(a,b)				(((a) < (b)) ? (a) : (b))
#define max(a,b)				(((a) > (b)) ? (a) : (b))
#define limit(i,l,u)			(((i)<(l)) ? (l) : (((i)>(u)) ? (u) : (i)))
#define swap(a, b)				{ int n = a; a = b; b = n; }

#define makelong(a, b)			((a)<<16 | (b&0xFFFF))
#define loword(a)				(a&0xFFFF)
#define hiword(a)				(a>>16)

#define NOT						!
#define AND						&&
#define OR						||

#define ZeroObj(obj)			memset(&obj, 0, sizeof(obj))
#define CountOf(array)			(sizeof(array)/sizeof(array[0]))
#define StrEmpty(s)				(s)[0] = 0
#define IsStrEmpty(s)			((NOT s) OR ((s)[0] == 0))
#define ASSERT(b)				assert(b)

extern void Add64(ulong *DestH, ulong *DestL, ulong SrcH, ulong SrcL);
extern void Msg(char *s, ...);

typedef unsigned long long int quad;

// stupid libraries don't have standard string functions
extern int strcmp(char *a, char *b);
extern int stricmp(char *a, char *b);
extern void strupr(char *a);
extern char *NewStr(char *a);
extern int strincmp(char *a, char *b, int i);


class String
{
protected:
	int Length;
	int Alloc;
	char *Data;

public:
// Constructors/Destructors
	String();
	String(char *Src);
	~String();

// Attributes & Operations
	int GetLength();
	bool SetLength(int Len);
	bool IsEmpty();
	void Empty();

	char operator[](int i);
	operator char *() const;

	// overloaded assignment
	String& operator =(String &Str);
	String& operator =(char ch);
	String& operator =(char *Str);

	// string concatenation
	String& operator +=(String &Str);
	String& operator +=(char ch);
	String& operator +=(char *Str);

	friend String operator +(String &Str1, String &Str2);
	friend String operator +(String &Str, char ch);
	friend String operator +(char ch, String &Str);
	friend String operator +(String &Str1, char *Str2);
	friend String operator +(char *Str1, String &Str2);

	// string comparison
	int Compare(char *Str);
	int CompareNoCase(char *Str);

	// simple sub-string extraction
	String Mid(int nFirst, int nCount = -1);
	String Left(int nCount);
	String Right(int nCount);

	// upper/lower/reverse conversion
	void MakeUpper();
	void MakeLower();

	// trimming whitespace (either side)
	void TrimRight();
	void TrimLeft();

	// searching (return starting index, or -1 if not found)
	// look for a single character match
	int Find(char ch);
	int ReverseFind(char ch);

	// look for a specific sub-string
	int Find(char *Sub);

	// simple formatting
	void Format(char *lpszFormat, ...);

	/*
	friend CArchive& AFXAPI operator<<(CArchive& ar, const String& string);
	friend CArchive& AFXAPI operator>>(CArchive& ar, String& string);
	*/
};

bool operator ==(String	&s1, String	&s2);
bool operator ==(String	&s1, char	*s2);
bool operator ==(char	*s1, String	&s2);
bool operator !=(String	&s1, String	&s2);
bool operator !=(String	&s1, char	*s2);
bool operator !=(char	*s1, String	&s2);
bool operator <(String	&s1, String	&s2);
bool operator <(String	&s1, char	*s2);
bool operator <(char	*s1, String	&s2);
bool operator >(String	&s1, String	&s2);
bool operator >(String	&s1, char	*s2);
bool operator >(char	*s1, String	&s2);
bool operator <=(String	&s1, String	&s2);
bool operator <=(String	&s1, char	*s2);
bool operator <=(char	*s1, String	&s2);
bool operator >=(String	&s1, String	&s2);
bool operator >=(String	&s1, char	*s2);
bool operator >=(char	*s1, String	&s2);

String operator +(String	&Str1,	String &Str2);
String operator +(String	&Str,	char   ch);
String operator +(char		ch,	String &Str);
String operator +(String	&Str1,	char   *Str2);
String operator +(char		*Str1,	String &Str2);

#endif