/*hdr
**      FILE:           Filter2.h
**      AUTHOR:         Matthew Allen
**      DATE:           1/4/97
**      DESCRIPTION:    Graphics file filters
**
**      Copyright (C) 1997, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef FILTER2_H
#define FILTER2_H

#include <string.h>
#include "contain.h"
#include "gui.h"
#include "prop.h"

extern int FindHeader(int Offset, char *Str, VFile &f);

class GdcFileFormat : public VFile {
protected:
	int Images;
	int Current;	// current image index (zero based)

	GdcSurface *GetSurface(GdcDCBase *pDC)
	{
		return pDC->pMem;
	}

	GdcRegion *GetClip(GdcDCBase *pDC)
	{
		return &pDC->Clip;
	}

	int GetLineLength(GdcDCBase *pDC)
	{
		return (pDC->pMem) ? pDC->pMem->Line : 0;
	}

public:
	ObjProperties Props;

	GdcFileFormat()
	{
		Images = 1;
		Current = 0;
	}
	virtual ~GdcFileFormat() {}

	virtual bool SetIndex(int Image) { return FALSE; }
	virtual int GetIndex() { return Current; }
	virtual int GetImages() { return Images; }
	virtual bool SetImages(int Images) { return FALSE; }
	virtual bool ReadHeader(bool All) { return FALSE; }
	virtual bool WriteAll() { return FALSE; }

	virtual bool ReadImage(GdcBasePrimitives *pDC) { return FALSE; }
	virtual bool WriteImage(GdcBasePrimitives *pDC) { return FALSE; }
};

class GdcBmp : public GdcFileFormat {

	int ActualBits;
	int ScanSize;
	bool SetSize(long *Info, GdcDCBase *pDC);

public:
	bool ReadImage(GdcBasePrimitives *pDC);
	bool WriteImage(GdcBasePrimitives *pDC);
};

class GdcIco : public GdcFileFormat {

	GdcSprite **pFrame;

public:
	GdcIco();
	~GdcIco();

	bool Seek(int Image) { return FALSE; }
	int GetIndex() { return Current; }
	int GetImages() { return Images; }
	bool SetImages(int Images) { return FALSE; }
	bool Read(bool All) { return FALSE; }
	bool Write(bool All) { return FALSE; }

};

#define GDC_RLE_COLOUR				0x0001
#define GDC_RLE_MONO				0x0002
#define GDC_RLE_READONLY			0x0004
#define GDC_RLE_KEY				0x0008		// set if valid

class GdcRleDC : public GdcMemDC {
protected:
	COLOUR	Key;

	int	Flags;
	int	Length;
	int	Alloc;
	uchar	*Data;
	uchar	**ScanLine;

	bool SetLength(int Len);
	bool FindScanLines();
	void Empty();

public:
	GdcRleDC();
	virtual ~GdcRleDC();

	void Move(GdcRleDC *pDC);

	bool Create(int x, int y, int Bits, int LineLen = 0);
	bool CreateInfo(int x, int y, int Bits);
	void ReadOnly(bool Read);
	bool ReadOnly();
	void Mono(bool Mono);
	bool Mono();

	void Update(int Flags);
	void Draw(GdcBasePrimitives *Dest, int x, int y);

	int SizeOf();
	bool Read(VFile &F);
	bool Write(VFile &F);
};

class GdcMbm : public GdcFileFormat {

	GdcDCBase	*pBmp;
	GdcRleDC	*pRle;

	bool IsRead() { return 0; }
	bool IsWrite() { return 0; }

	bool ReadAny();
	bool WriteAny();

public:
	GdcMbm();
	~GdcMbm();

	bool Open(char *Name, int Attrib);
	bool Close();

	bool ReadImage(GdcBasePrimitives *pDC);
	bool WriteImage(GdcBasePrimitives *pDC);
	bool ReadRle(GdcRleDC *pDC);
	bool WriteRle(GdcRleDC *pDC);
};

class GdcPcx : public GdcFileFormat {
public:
	bool ReadImage(GdcBasePrimitives *pDC);
	bool WriteImage(GdcBasePrimitives *pDC);
};

#define MAX_CODES   				4095

class GdcGif : public GdcFileFormat {

	GdcBasePrimitives *pDC;

	// Old GIF coder stuff
	short linewidth;
	int lines;
	int pass;
	short curr_size;			/* The current code size */
	short clearc;				/* Value for a clear code */
	short ending;				/* Value for a ending code */
	short newcodes;				/* First available code */
	short top_slot;				/* Highest code for current size */
	short slot;				/* Last read code */
	short navail_bytes;			/* # bytes left in block */
	short nbits_left;			/* # bits left in current byte */
	uchar b1;				/* Current byte */
	uchar byte_buff[257];			/* Current block */
	uchar *pbytes;				/* Pointer to next byte in block */
	uchar stack[MAX_CODES+1];		/* Stack for storing pixels */
	uchar suffix[MAX_CODES+1];		/* Suffix table */
	ushort prefix[MAX_CODES+1];		/* Prefix linked list */
	int bad_code_count;

	int get_byte();
	int out_line(uchar *pixels, int linewidth, int interlaced, int BitDepth);
	short init_exp(short size);
	short get_next_code();
	short decoder(int BitDepth, uchar interlaced);

public:
	bool ReadImage(GdcBasePrimitives *pDC);
	bool WriteImage(GdcBasePrimitives *pDC);
};

class GdcPng : public GdcFileFormat {

	static bool CrcTableComputed;
	static ulong CrcTable[256];

	void MakeCrcTable();
	ulong UpdateCrc(ulong crc, uchar *buf, int len);
	ulong Crc(uchar *buf, int len);

	struct IHDR_Obj {

		uchar Raw[13];

		int x;
		int y;
		uchar Bits;
		uchar ColourType;
		uchar Compression;
		uchar Filter;
		uchar Interlace;

		int TotalBits;

		void Read();
		void Write();

	} Header;

	int Pos;
	uchar *PrevScanLine;
	GdcBasePrimitives *pDC;
	BytePipe DataPipe;

	bool ProcessData();
	uchar PaethPredictor(uchar a, uchar b, uchar c);
	bool WriteObject(ulong ChunkId, uchar *Ptr, int Len);

public:
	GdcPng() { Pos = 0; PrevScanLine = 0; }
	~GdcPng() { DeleteArray(PrevScanLine); }

	bool ReadImage(GdcBasePrimitives *pDC);
	bool WriteImage(GdcBasePrimitives *pDC);
};

// JPEG
class GdcJpeg : public GdcFileFormat {
public:
	bool ReadImage(GdcBasePrimitives *pDC);
	bool WriteImage(GdcBasePrimitives *pDC);
};

// TIFF
#define TIFF_USE_LZW

class IFD {

	void *Data;

public:
	ushort	Tag;
	ushort	Type;
	ulong	Count;
	ulong	Offset;

	IFD();
	~IFD();
	void *GetData();
	char *Str();
	int ElementSize();
	int Value();
	int ArrayValue(int i);
	bool Read(VFile &f);
	bool Write(VFile &f);
};

class GdcTiff : public GdcFileFormat {

	List<IFD> Blocks;
	IFD *FindTag(ushort n);
	int ScanLength;

	#ifdef TIFF_USE_LZW
	// Codes
	int CodeSize;
	int ClearCode;
	int EoiCode;
	int NextCode;
	uchar *NextBuf;
	int CodeLength[4096];
	uchar *CodePtr[4096];
	uchar *CodeBuf;

	// Input buffering
	int BitPos;
	int BitsLeft;
	uchar *BufPos;
	int BufSize;
	uchar *Buffer;

	// Output buffering
	GdcBasePrimitives *pOutDC;
	uchar *Out;
	uchar *OutEnd;
	int CurrentScanline;

	void InitializeTable();
	int GetNextCode();
	void WriteString(uchar *d, int Size);
	#endif

	bool ProcessRead(GdcBasePrimitives *pDC);
	bool ProcessWrite(GdcBasePrimitives *pDC);

public:
	GWindow *Parent;

	GdcTiff();
	~GdcTiff();

	bool ReadImage(GdcBasePrimitives *pDC);
	bool WriteImage(GdcBasePrimitives *pDC);
};

#endif