/*hdr
**	FILE:			gui.h
**	AUTHOR:			Matthew Allen
**	DATE:			19/12/97
**	DESCRIPTION:	Graphical user interface
**
**
**	Copyright (C) 1997-8, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/////////////////////////////////////////////////////////////////////////////////////
// Includes
#ifndef __GUI_H
#define __GUI_H

#include <string.h>
#include <InterfaceKit.h>

/////////////////////////////////////////////////////////////////////////////////////
// Class names
class VKey;
class VMouse;
class GObject;
class GTarget;
class GComponent;
class GEvent;
class GId;
class GApp;
class GAppWindow;
class GWindow;
class GView;
class GFileType;
class GFileSelect;
class GFindReplace;
class GSubMenu;
class GMenuItem;
class GMenu;
class GToolBar;
class GToolButton;
class GSplitter;
class GStatusPane;
class GStatusBar;
class GToolColour;
class GScrollBar;
class GControl;
class GImageList;
class GFont {};
class Progress;

/////////////////////////////////////////////////////////////////////////////////////
typedef BMessage GMessage;
#define Message(m) m->what
extern GMessage CreateMsg(int m, int a, int b);

/////////////////////////////////////////////////////////////////////////////////////
// Defines
#define hAppInst()				GWin32App::Instance()

#define SetFlag(i, f)			(i) |= (f)
#define ClearFlag(i, f)			(i) &= ~(f)
#define TestFlag(i, f)			((i) & (f))

#define K_CHAR					0x0

#define MK_LEFT					B_PRIMARY_MOUSE_BUTTON
#define MK_RIGHT				B_SECONDARY_MOUSE_BUTTON
#define MK_MIDDLE				B_TERTIARY_MOUSE_BUTTON
#define MK_CTRL					0x08
#define MK_ALT					0x10
#define MK_SHIFT				0x20

// Edge types
#define SUNKEN					1
#define RAISED					2
#define CHISEL					3
#define FLAT					4

// Colours
#define TOOL_VLOW				Rgb24(128, 128, 128)
#define TOOL_LOW				Rgb24(164, 164, 164)
#define TOOL_MED				Rgb24(216, 216, 216)
#define TOOL_HIGH				Rgb24(236, 236, 236)
#define TOOL_VHIGH				Rgb24(255, 255, 255)
#define C_WORKSPACE				Rgb24(255, 255, 255)

// Messages
#define WM_USER					(0x1000) // how strange, this isn't defined in BeOS?? :P

// wParam = bool Inside; // is the mouse inside the client area?
// lParam = MAKELONG(x, y); // mouse location
#define WM_MOUSEENTER			(WM_USER+0x100)
#define WM_MOUSEEXIT			(WM_USER+0x101)

// wParam = (GWindow*) Wnd;
// lParam = (int) Flags;
#define WM_CHANGE				(WM_USER+0x102)

// wParam = (GWindow*) Wnd;
// lParam = (char*) Text; // description from window
#define WM_DESCRIBE				(WM_USER+0x103)

// return (bool)
#define WM_WANT_DIALOG_PROC		(WM_USER+0x104)

#define WM_MENU					(WM_USER+0x105)

// Dialog stuff
#define IDOK					1
#define IDCANCEL				2

// Keys
#define VK_DELETE				B_DELETE
#define VK_F3					B_F3_KEY
#define VK_SHIFT				B_SHIFT_KEY
#define VK_ESCAPE				B_ESCAPE
#define VK_RETURN				B_ENTER
#define VK_BACK					B_BACKSPACE
#define VK_RIGHT				B_RIGHT_ARROW
#define VK_LEFT					B_LEFT_ARROW
#define VK_UP					B_UP_ARROW
#define VK_DOWN					B_DOWN_ARROW
#define VK_PRIOR				B_PAGE_UP
#define VK_NEXT					B_PAGE_DOWN
#define VK_HOME					B_HOME
#define VK_END					B_END
#define VK_INSERT				B_INSERT


/////////////////////////////////////////////////////////////////////////////////////
// Externs
extern char *NewStr(char *s);
extern BMessage NewMsg(int Msg, int a, int b);

extern void DrawBox(GdcBasePrimitives *pDC, GdcRegion &r, bool Sunken, bool Fill);
extern void DrawMoveDeepEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type);
extern void DrawMoveShallowEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type);
extern void DrawMoveFill(GdcBasePrimitives *pDC, GdcRegion &r, int Width = -1);
extern long MouseWatcher(void *Ptr);

extern char *LgiLoadString(int Id);
extern bool LgiExecute(char *File, char *Args, char *Dir);
extern bool LgiGetCursorPos(GdcPt2 *Pt);

// standard message box
// Type:
#define MB_OK					1
#define MB_YESNO				2
#define MB_YESNOCANCEL			3

// return:
#define IDOK					1
#define IDCANCEL				2
#define IDYES					3
#define IDNO					4

extern int LgiMsg(GWindow *Parent, char *Str, char *Title, int Type, ...);

/////////////////////////////////////////////////////////////////////////////////////
// Classes
class VKey {
public:
	uchar c;
	int Flags;
	int Data;
	bool Down;
	bool IsChar;

	bool Ctrl() { return Flags & MK_CTRL; }
	bool Alt() { return Flags & MK_ALT; }
	bool Shift() { return Flags & MK_SHIFT; }
};

class VMouse {
public:
	int x;
	int y;
	int Flags;
	bool Down;
	bool Double;

	bool Left() { return Flags & MK_LEFT; }
	bool Middle() { return Flags & MK_MIDDLE; }
	bool Right() { return Flags & MK_RIGHT; }
	bool Ctrl() { return Flags & MK_CTRL; }
	bool Alt() { return Flags & MK_ALT; }
	bool Shift() { return Flags & MK_SHIFT; }
};

class Progress {
protected:
	char *Description;
	int Start;
	int Val;
	int Low, High;
	char *Type;
	double Scale;

public:
	Progress();
	Progress(char *desc, int l, int h, char *type = NULL, double scale = 1.0);
	virtual ~Progress();

	virtual char *GetDescription() { return Description; }
	virtual void SetDescription(char *d);
	virtual void GetLimits(int *l, int *h);
	virtual void SetLimits(int l, int h);
	virtual int Value() { return Val; }
	virtual void Value(int v) { Val = v; }
	virtual double GetScale() { return Scale; }
	virtual void SetScale(double s) { Scale = s; }
	virtual char *GetType() { return Type; }
	virtual void SetType(char *t) { Type = t; }

	Progress &operator =(Progress &p);
};

class GObject {

	char *ObjName;

public:
	GObject()
	{
		ObjName = 0;
	}
	
	virtual ~GObject()
	{
		DeleteArray(ObjName);
	}

	virtual char *Name() { return ObjName; }
	virtual bool Name(char *n)
	{
		if (n == ObjName) return TRUE;
		DeleteObj(ObjName);
		if (n)
		{
			ObjName = NEW(char[strlen(n)+1]);
			if (ObjName)
			{
				strcpy(ObjName, n);
				return TRUE;
			}
		}
		return FALSE;
	}

	virtual int Sizeof() { return 0; }
	virtual bool Serialize(VFile &f, bool Write) { return FALSE; }
};

VFile &operator >>(VFile &f, GObject &o);
VFile &operator <<(VFile &f, GObject &o);

class GTarget {
public:
	virtual int OnEvent(BMessage *Msg) { return 0; }
};

class GComponent {
public:
	// Status
	virtual bool Enabled() { return FALSE; }
	virtual void Enabled(bool e) {}
	virtual bool Visible() { return FALSE; }
	virtual void Visible(bool v) {}
	virtual bool Focus() { return FALSE; }
	virtual void Focus(bool f) {}
};

class GEvent {
public:
	// Mouse related
	virtual void OnMouseClick(VMouse &m) {}
	virtual void OnMouseEnter(VMouse &m) {}
	virtual void OnMouseExit(VMouse &m) {}
	virtual void OnMouseMove(VMouse &m) {}

	// Keyboard related
	virtual void OnKey(VKey &k) {}

	// Other
	virtual void OnPaint(GdcBasePrimitives *pDC) {}
	virtual void OnPosChange() {}
};

class GId {
protected:
	int Id;

public:
	GId() { Id = 0; }

	virtual int GetId() { return Id; }
	virtual void SetId(int i) { Id = i; }
	virtual int Value() { return 0; }
	virtual void Value(int i) {}
};

// redirector for the BWindow object back to the GWindow owner
class GWnd : public BWindow {

	GWindow *Notify;

public:
	GWnd(GWindow *notify);
	GWnd(GWindow *notify, BRect frame, char *title, window_type type, uint32 flags, uint32 workspaces = B_CURRENT_WORKSPACE);
	bool QuitRequested();
};

/////////////////////////////////////////////////////////////////////////////////
#define SysFont				(GApp::ObjInstance()->SystemFont)
#define ExitApp()			(GApp::ObjInstance()->Exit())

class GApp : public BApplication {

	static char *CmdLine;
	static GApp *TheApp;

public:
	GdcBeTtf *SystemFont;
	GAppWindow *AppWnd;

	GApp(char *MimeSig, char *lpCmdLine);
	virtual ~GApp();

	static GApp *ObjInstance() { return TheApp; }

	virtual void OnCmdLine(char *Cmd);
	virtual void OnArgs(int Args, char **Arg);
	virtual void Exit(int Code = 0);
};

typedef List<GWindow> WindowList;

#define GWF_ENABLE					0x0001
#define GWF_VISIBLE					0x0002
#define GWF_FOCUS					0x0004
#define GWF_OVER					0x0008
#define GWF_DROP_TARGET				0x0010
#define GWF_DIALOG_PROCESSING		0x0020

/*
#define GRedirector()

	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool focused = true) { OnFocus(focused); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }
*/

class GWindow : public GObject, public GTarget, public GComponent, public GEvent {

	friend		class GAppWindow;
	friend		class GView;
	friend		class GDialog;
	friend		class GControl;
	friend		class GMenu;
	friend		class GSubMenu;
	friend		long MouseThread(void *Ptr);
	friend		class GWnd;

protected:
	// OS
	BView		*Sys_View;

	// General
	int			WndFlags;
	GdcRegion	Pos;
	GWindow		*Parent;
	GWindow		*Notify;
	GWindow		*WantPulses;

	void Sys_KeyDown(const char *bytes, int32 numBytes);
	void Sys_KeyUp(const char *bytes, int32 numBytes);
	void Sys_MouseDown(BPoint point);
	void Sys_MouseMoved(BPoint point, uint32 transit, const BMessage *message);
	virtual bool QuitRequested() {}

public:
	bool		Mark;
	WindowList	Children;

	GWindow(BView *wnd);
	virtual ~GWindow();

	virtual BWindow *GetBWindow() { return (Sys_View) ? Sys_View->Window() : 0; }
	virtual BView *GetBView() { return Sys_View; }
	virtual bool IsDropTarget() { return TestFlag(WndFlags, GWF_DROP_TARGET); }
	virtual void IsDropTarget(bool t) { if (t) SetFlag(WndFlags, GWF_DROP_TARGET); else ClearFlag(WndFlags, GWF_DROP_TARGET); }

	// Create a virtual window
	bool Create(char *name, GdcRegion &Pos);

	// Target
	int OnEvent(BMessage *Msg);
	virtual int OnNotify(GControl *Ctrl, int Flags) { return 0; }

	// Flags
	bool Enabled();
	void Enabled(bool e);
	bool Visible();
	void Visible(bool v);
	bool Focus();
	void Focus(bool f);

	// Events
	// void OnMouseClick(VMouse &m);
	// void OnMouseEnter(VMouse &m);
	// void OnMouseExit(VMouse &m);
	// void OnMouseMove(VMouse &m);
	// void OnKey(VKey &k);
	// void OnPaint(GdcBasePrimitives *pDC);
	virtual void OnFocus(bool f) {}
	virtual void OnPulse() {}
	virtual void OnCreate() {}
	virtual bool OnRequestClose() { return true; }
	void OnPosChange();

	// Other
	bool Name(char *n);
	char *Name();
	virtual bool SetPos(GdcRegion &p, bool Repaint = FALSE);

	virtual bool Invalidate(GdcRegion *r = NULL, bool Repaint = FALSE);
	virtual bool Capture(bool c);
	virtual bool IsOver(VMouse &m);
	virtual GWindow *WindowFromPoint(int x, int y);
	virtual void SetPulse(int Length = -1);

	// Children
	virtual bool Attach(GWindow *p);
	virtual bool Detach();
	virtual void OnChildrenChanged(GWindow *Wnd, bool Attaching) {}

	// Stubs
	virtual GWindow *GetParent() { return Parent; }
	virtual void SetParent(GWindow *p) { Parent = p; }
	virtual GControl *IsControl() { return 0; }
	virtual GWindow *GetNotify() { return Notify; }
	virtual void SetNotify(GWindow *n) { Notify = n; }
	virtual int X() { return Pos.X(); }
	virtual int Y() { return Pos.Y(); }
	virtual GdcRegion &GetPos() { return Pos; }
	virtual GControl *FindControl(int Id) { return 0; }
};

// Dialog stuff
#include "guidlg.h"

////////////////////////////////////////////////////////////////////////////////////////////////
class GScrollBar : public GTarget, public GComponent {

	BScrollBar *Bar;

public:
	GScrollBar(BScrollBar *bar);
	~GScrollBar();

	// data
	bool Vertical();
	int Pos();
	void SetPos(int p);
	void Limits(int &Low, int &High);
	void SetLimits(int Low, int High);
	int Page();
	void SetPage(int p);
	
	// events
	int OnEvent(BMessage *Msg);

	// overridables
	virtual void OnChange(int Pos) {}
};

///////////////////////////////////////////////////////////////////////////////////////////////
#define GV_EDGE_TOP				0x0001
#define GV_EDGE_RIGHT			0x0002
#define GV_EDGE_BOTTOM			0x0004
#define GV_EDGE_LEFT			0x0008

class GView : public BView, public GWindow {
protected:
	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool focused = true) { OnFocus(focused); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }
	
	int				ExtraFlags;
	BScrollView		*ScrollView;
	GScrollBar		*VScroll;
	GScrollBar		*HScroll;

	GdcRegion *FindLargest(GdcComplexRegion &r);
	GdcRegion *FindSmallestFit(GdcComplexRegion &r, int Sx, int Sy);
	GdcRegion *FindLargestEdge(GdcComplexRegion &r, int Edge);
	bool SetScrollBars(bool x, bool y);

public:
	GView();
	~GView();

	virtual void GetScrollPos(int &x, int &y);
	virtual void SetScrollPos(int x, int y);
	int OnEvent(BMessage *Msg);

	virtual bool Pour(GdcComplexRegion &r) { return FALSE; }
};

///////////////////////////////////////////////////////////////////////////////////////////////
// Menu wrappers
class GSubMenu : public BMenu, public GObject, public GTarget, public GComponent {

	friend class GMenuItem;
	friend class GMenu;

protected:
	GWindow *Window;

public:
	GSubMenu(char *name = "");
	virtual ~GSubMenu();

	GMenuItem *AppendItem(char *Str, int Id, bool Enabled);
	GMenuItem *AppendSeparator();
	GSubMenu *AppendSub(char *Str);

	int Float(GWindow *Parent, int x, int y);
};

class GMenuItem : public BMenuItem, public GObject, public GTarget, public GComponent {

	friend class GSubMenu;
	friend class GMenu;

private:
protected:
	BMessage	*Msg;

public:
	GMenuItem();
	GMenuItem(GSubMenu *p);
	virtual ~GMenuItem();

	void Id(int i);
	void Checked(bool c);
	bool Name(char *n);
	void Enabled(bool e);
	void Focus(bool f);

	int Id();
	bool Checked() { return 0; }
	bool Enabled() { return 0; }
	bool Visible() { return TRUE; }
	bool Focus() { return 0; }
};

class GMenu : public BMenuBar, public GObject, public GTarget, public GComponent {

	friend class GSubMenu;
	friend class GMenuItem;

protected:
	GWindow *Window;

public:
	GMenu();
	virtual ~GMenu();

	GMenuItem *AppendItem(char *Str, int Id, bool Enabled);
	GMenuItem *AppendSeparator();
	GSubMenu *AppendSub(char *Str);

	bool Attach(GWindow *p);
	bool Detach();
	bool Load(GWindow *p, char *Res);
};

///////////////////////////////////////////////////////////////////////////////////////////////////
typedef List<GView> ViewList;

class GAppWindow : public BView, public GWindow {

	bool QuitRequested();

protected:
	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool focused = true) { OnFocus(focused); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }

	BWindow		*Wnd;
	GMenu		*Menu;
	ViewList	Views;

	void		Pour();

public:
	GAppWindow();
	~GAppWindow();

	virtual BWindow *GetBWindow() { return Wnd; }
	GMenu *GetMenu() { return Menu; }
	bool Visible();
	void Visible(bool v);
	bool Name(char *n);
	char *Name();
	GdcRegion &GetPos();
	bool SetPos(GdcRegion &p, bool Repaint = FALSE);

	void OnChildrenChanged(GWindow *Wnd, bool Attaching);

	void OnCreate();
	void OnPosChange();
	int OnEvent(BMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	virtual int OnCommand(int Cmd, int Event, void* hWnd) { return 0; }
	virtual void OnReceiveFiles(List<char> &Files) {}
};

///////////////////////////////////////////////////////////////////////////////////////////////////
// File select dialog
class GFileType : public GObject {

	char *Ext;

public:
	GFileType()
	{
		Ext = 0;
	}

	~GFileType()
	{
		DeleteObj(Ext);
	}

	char *Extension() { return Ext; }
	bool Extension(char *e)
	{
		return (Ext = NewStr(e)) != 0;
	}

	char *Description() { return Name(); }
	bool Description(char *d) { return Name(d); }
};

typedef List<GFileType> FileTypeList;

class GFileSelect : public GObject, public BHandler {

	BFilePanel	*Panel;
	BMessenger	*Messenger;

	int			Size;
	char		*FileStr;
	FileTypeList TypeList;
	GWindow		*ParentWnd;
	bool		MultiSel;
	char		*InitDir;
	char		*DefExt;
	char		*TitleStr;
	bool		WaitForMessage;

	void Wait();
	void MessageReceived(BMessage *message);
	bool DoDialog(file_panel_mode Mode);

public:
	GFileSelect();
	~GFileSelect();

	char *Name();
	bool Name(char *n);
	char *operator [](int i);
	int Names();
	char *File();
	char *Extension();

	int Types();
	GFileType *Type(int n);
	bool Type(char *Description, char *Extension);

	GWindow *Parent();
	void Parent(GWindow *Window);

	bool MultiSelect();
	void MultiSelect(bool Multi);

	char *InitialDir();
	void InitialDir(char *InitDir);

	char *Title();
	void Title(char *Title);

	char *DefaultExtension();
	void DefaultExtension(char *DefExt);

	bool Open();
	bool Save();
};

class GFindReplace {

	int FindMsgID;
	bool FindOp;

public:
	char Find[256];
	char Replace[256];

	GFindReplace(GWindow *hParent = NULL);
	virtual ~GFindReplace();

	void SetParent(GWindow *hParent);

	bool OpenFind();
	bool OpenReplace();
	int OnEvent(BMessage *Msg);
	virtual bool OnFind();
	virtual bool OnReplace();

	bool CaseSensitive();
	void CaseSensitive(bool i);
	bool WholeWord();
	void WholeWord(bool i);
	bool BtnFindNext();
	bool BtnReplace();
	bool BtnReplaceAll();
};

/////////////////////////////////////////////////////////////////////////////////////////////
#define IMGLST_SELECTED				0x0001
#define IMGLST_DISABLED				0x0002

class GImageList : public GdcMemDC {

	int Sx, Sy;
	GdcBasePrimitives *Mask;

public:
	GImageList(int x, int y, GdcBasePrimitives *pDC = NULL);
	~GImageList();

	bool Create(int x, int y, int Bits, int LineLen = 0, bool KeepData = FALSE);
	void Update(int Flags);
	void Draw(GdcBasePrimitives *pDest, int x, int y, int Image, int Flags = 0);
};

///////////////////////////////////////////////////////////////////////////////////////////////
// Toolbar stuff

// Tool button types
#define TBT_PUSH				0
#define TBT_RADIO				1
#define TBT_TOGGLE				2

// Default tool button sizes
#define TBS_X					19
#define TBS_Y					18

// Commands
#define IDM_NONE				0
#define IDM_SEPARATOR			-1
#define IDM_BREAK				-2

// Dimensions
#define BORDER_SHADE			1
#define BORDER_SPACER			1
#define BORDER_SEPARATOR		4
#define BORDER_BUTTON			1

class GToolButton : public BView, public GControl {

	friend class GToolBar;

protected:
	void AttachedToWindow() { OnCreate(); }
	void Draw(BRect UpdateRect) { GdcViewDC DC(Sys_View); OnPaint(&DC); }
	void FrameMoved(BPoint Point) { Pos.Offset(Point.x-Pos.x1, Point.y-Pos.y1); OnPosChange(); }
	void FrameResized(float width, float height) { Pos.Dimension(width+1, height+1); OnPosChange(); }
	void Pulse() { OnPulse(); }
	void MessageReceived(BMessage *message) { OnEvent(message); }
	void MakeFocus(bool focused = true) { OnFocus(focused); }
	void KeyDown(const char *bytes, int32 numBytes) { Sys_KeyDown(bytes, numBytes); }
	void KeyUp(const char *bytes, int32 numBytes) { Sys_KeyUp(bytes, numBytes); }
	void MouseDown(BPoint point) { Sys_MouseDown(point); }
	void MouseMoved(BPoint point, uint32 transit, const BMessage *message) { Sys_MouseMoved(point, transit, message); }
	
	int		Type;		// Button Type

	bool		Clicked;
	bool		Down;
	int		ImgIndex;
	
	virtual	void SetDown(bool d);

public:
	GToolButton(int Bx, int By);
	~GToolButton();

	int Value() { return Down; }
	void Value(bool i);

	void OnPaint(GdcBasePrimitives *pDC);

	void OnMouseClick(VMouse &m);
	void OnMouseEnter(VMouse &m);
	void OnMouseExit(VMouse &m);

	virtual void OnCommand(VMouse &m);
	virtual void SendNotify(int Flags);
	virtual bool GetDimension(int &x, int &y) { return FALSE; }
};

class GToolBar : public GView {

	friend class GToolButton;

protected:
	int Bx, By;
	int Sx, Sy;
	bool Vertical;
	int LastIndex;
	GImageList *ImgList;

	virtual int PostCommand(int Cmd);
	virtual int PostDescription(GWindow *Ctrl, char *Text);

public:
	GToolBar();
	~GToolBar();

	int X();
	int Y();

	int OnEvent(BMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);

	bool IsVertical() { return Vertical; }
	void IsVertical(bool v) { Vertical = v; }
	bool SetBitmap(char *File, int Bx, int By);
	bool SetDC(GdcBasePrimitives *pDC, int Bx, int By);

	GToolButton *AppendButton(char *Tip, int Id, int Type = TBT_PUSH, int Enabled = TRUE, int IconId = -1);
	bool AppendControl(GControl *Ctrl);
	bool AppendSeparator();
	bool AppendBreak();

	void Empty();
	bool Pour(GdcComplexRegion &r);
};

////////////////////////////////////////////////////////////////////////////////////////////////
class GSplitter : public GView {

	bool Vertical;
	bool Moving;
	int SplitPos;
	GdcRegion Bar;
	int Offset;

	GdcRegion PosA;
	GdcRegion PosB;
	GWindow *ViewA;
	GWindow *ViewB;

	void CalcRegions(GdcRegion Rect);

public:
	GSplitter();
	~GSplitter();

	bool IsVertical() { return Vertical; }
	void IsVertical(bool v) { Vertical = v; }
	int Split() { return SplitPos; }
	GWindow *GetViewA() { return ViewA; }
	GWindow *GetViewB() { return ViewB; }
	void SetViewA(GWindow *a);
	void SetViewB(GWindow *b);
	void Split(int s);

	bool Pour(GdcComplexRegion &r);
	void OnPaint(GdcBasePrimitives *pDC);

	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
};

////////////////////////////////////////////////////////////////////////////////////////////////
#define STATUSBAR_SEPARATOR			4
#define GSP_SUNKEN				0x0001

typedef GStatusPane *GStatusPanePtr;

class GStatusBar : public GView, private PtrArray {

	friend class GStatusPane;

protected:
	GdcBeTtf *Font;
	void RePour();

public:
	GStatusBar();
	~GStatusBar();

	bool Pour(GdcComplexRegion &r);
	void OnPaint(GdcBasePrimitives *pDC);

	GStatusPane *AppendPane(char *Text, int Width);
	bool AppendPane(GStatusPane *Pane);
};

class GStatusPane : public BView, public GWindow {

	friend class GStatusBar;

protected:
	int		Flags;
	int		Width;
	GdcBasePrimitives *pDC;

public:
	GStatusPane();
	~GStatusPane();

	char *Name() { return GObject::Name(); }
	bool Name(char *n);
	void OnPaint(GdcBasePrimitives *pDC);

	int X();
	void X(int x);
	bool Sunken();
	void Sunken(bool s);
	GdcBasePrimitives *Bitmap();
	void Bitmap(GdcBasePrimitives *pdc);
};

/////////////////////////////////////////////////////////////////////////////////////////////
class GCommand : public GObject, public GComponent {

	int		Flags;

public:
	int		Id;
	GToolButton	*ToolButton;
	GMenuItem	*MenuItem;
	VKey		*Accelerator;
	char		*TipHelp;
	
	GCommand();
	~GCommand();

	bool Enabled();
	void Enabled(bool e);
	bool Value();
	void Value(bool v);
};

/////////////////////////////////////////////////////////////////////////////////////////////
class GClipBoard {

	GWindow *Owner;
	bool Open;

public:
	GClipBoard(GWindow *o);
	~GClipBoard();

	bool Empty();

	// Text
	bool Text(char *Str);
	char *Text();

	// Bitmap
	bool Bitmap(GdcBasePrimitives *pDC);
	GdcBasePrimitives *Bitmap();
	
};

/////////////////////////////////////////////////////////////////////////////////////////
class ProgressList : public List<Progress> {

	bool InUse;

public:
	ProgressList();

	bool Lock();
	void Unlock();
};

class GProgressPane : public Progress, public GView {

	friend class GProgressDlg;

	Progress *Ref;

protected:
	bool Wait;
	bool Canceled;

	GText *Desc;
	GText *ValText;
	GText *Rate;
	GProgress *Bar;
	GButton *But;

public:
	GProgressPane();
	~GProgressPane();

	char *GetDescription();
	void SetDescription(char *d);
	void GetLimits(int *l, int *h) { Progress::GetLimits(l, h); }
	void SetLimits(int l, int h);
	int Value() { return Progress::Value(); }
	void Value(int v);
	bool IsCanceled() { return Canceled; }

	void OnCreate();
	int OnNotify(GControl *Ctrl, int Flags);
	void OnPaint(GdcBasePrimitives *pDC);
	bool Pour(GdcComplexRegion &r);
	GFont GetFont();
};

class GProgressDlg : public GDialog, public Progress {
protected:
	bool Wait;
	List<GProgressPane> Progri;

public:
	GProgressDlg(GWindow *parent = 0, bool wait = FALSE);
	~GProgressDlg();

	GProgressPane *Push();
	void Pop(GProgressPane *p = 0);
	char *GetDescription();
	void SetDescription(char *d);
	void GetLimits(int *l, int *h);
	void SetLimits(int l, int h);
	int Value();
	void Value(int v);
	bool IsCanceled();
	char *GetType();
	void SetType(char *t);

	void OnSync(ProgressList *Prg);

	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	void OnCreate();
};

//////////////////////////////////////////////////////////////////////////
class GThread {

	thread_id ThreadId;
	static long EntryPoint(void *Ptr);
	int ExitCodeVal;
	bool Exited;

public:
	GThread(bool Sleep = FALSE, int StackSize = 8 << 10);
	virtual ~GThread();

	thread_id Handle() { return ThreadId; }
	int ExitCode();
	bool IsExited();
	void Run(bool b);
	void Terminate();
	virtual int Main();
};

/////////////////////////////////////////////////////////////////////////
class GTrayIcon : public GObject, public GComponent {
protected:
	static int Id;

	int Val;		// which icon visible
	int Icons;		// size of array
	int MyId;
	int Msg;
	GWindow *Notify;

public:
	GTrayIcon(int Icons = 1);
	~GTrayIcon();

	bool LoadIcon(char *Str, int Index = 0);
	bool Visible();
	void Visible(bool v);

	int Value();
	void Value(int v);
	int NotifyMsg() { return Msg; }
	void NotifyMsg(int i) { Msg = i; }
	GWindow *GetNotify() { return Notify; }
	void SetNotify(GWindow *n) { Notify = n; }
};

#endif