/*hdr
**      FILE:           GuiCommon.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Common windows and controls
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"

#include "gdc2.h"
#include "gui.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
GFileSelect::GFileSelect() : BHandler("GFileSelect")
{
	Size = 4096;
	FileStr = NEW(char[Size]);
	if (FileStr)
	{
		*FileStr = 0;
	}
	else
	{
		Size = 0;
	}
	
	MultiSel = FALSE;
	InitDir = 0;
	DefExt = 0;
	TitleStr = 0;
	Panel = 0;
}

GFileSelect::~GFileSelect()
{
	DeleteObj(Panel);
	DeleteArray(FileStr);
	DeleteArray(InitDir);
	DeleteArray(DefExt);
	DeleteArray(TitleStr);

	for (GFileType *Type = TypeList.First(); Type; Type = TypeList.Next())
	{
		DeleteObj(Type);
	}
}

char *GFileSelect::Name()
{
	return FileStr;
}

bool GFileSelect::Name(char *n)
{
	bool Status = FALSE;

	if (n AND FileStr)
	{
		strncpy(FileStr, n, Size);
	}

	return Status;
}

char *GFileSelect::operator [](int i)
{
	return NULL;
}

int GFileSelect::Names()
{
	return 1;
}

char *GFileSelect::File()
{
	char *Ret = 0;
	char *Last = FileStr;
	char *Dir = strchr(Last, '/');
	while (Dir)
	{
		Last = Dir + 1;
		Dir = strchr(Last, '/');
	}
	return Last;
}

char *GFileSelect::Extension()
{
	char *Ret = File();
	char *Dot = strchr(Ret, '.');
	if (Dot)
	{
		return Dot + 1;
	}
	return NULL;
}

int GFileSelect::Types()
{
	return TypeList.GetItems();
}

GFileType *GFileSelect::Type(int n)
{
	return (GFileType*) TypeList[n];
}

bool GFileSelect::Type(char *Description, char *Extension)
{
	GFileType *Type = NEW(GFileType);
	if (Type)
	{
		Type->Description(Description);
		Type->Extension(Extension);
		TypeList.Insert(Type);
	}

	return Type != 0;
}

GWindow *GFileSelect::Parent()
{
	return ParentWnd;
}

void GFileSelect::Parent(GWindow *Window)
{
	ParentWnd = Window;
}

bool GFileSelect::MultiSelect()
{
	return MultiSel;
}

void GFileSelect::MultiSelect(bool Multi)
{
	MultiSel = Multi;
}

char *GFileSelect::InitialDir()
{
	return InitDir;
}

void GFileSelect::InitialDir(char *Dir)
{
	DeleteArray(InitDir);
	InitDir = NewStr(Dir);
}

char *GFileSelect::Title()
{
	return TitleStr;
}

void GFileSelect::Title(char *title)
{
	DeleteArray(TitleStr);
	TitleStr = NewStr(title);
}

char *GFileSelect::DefaultExtension()
{
	return DefExt;
}

void GFileSelect::DefaultExtension(char *DefExt)
{
	DeleteArray(DefExt);
	DefExt = NewStr(DefExt);
}

void GFileSelect::Wait()
{
	thread_id	this_tid = find_thread(NULL);
	BLooper		*pLoop;
	BWindow		*pWin = 0;

	WaitForMessage = true;
	pLoop = BLooper::LooperForThread(this_tid);
	if (pLoop)
	{
		pWin = dynamic_cast<BWindow*>(pLoop);
	}

	if (pWin)
	{
		do
		{
			// update the window periodically
			pWin->UpdateIfNeeded();
			snooze(50);
		} while (Panel->IsShowing() AND WaitForMessage);
	}
	else
	{
		do
		{
			// just wait for exit
			snooze(50);
		} while (Panel->IsShowing() AND WaitForMessage);
	}
}

void GFileSelect::MessageReceived(BMessage *message)
{
	WaitForMessage = false;

	entry_ref ref;
	if (message->FindRef("refs", &ref) == B_OK)
	{
		BEntry Entry(&ref, true);
		BPath Path;
		Entry.GetPath(&Path);
		strcpy(FileStr, Path.Path());
	}
}

bool GFileSelect::DoDialog(file_panel_mode Mode)
{
	BEntry Dir(InitDir);
	entry_ref DirRef;
	Dir.GetRef(&DirRef);

	FileStr[0] = 0;
	Messenger = 0;
	Panel = NEW(BFilePanel(	Mode,
							NULL,
							&DirRef,
							B_FILE_NODE,
							MultiSel,
							NEW(BMessage(WM_CHANGE)),
							NULL,
							TRUE, // modal
							FALSE)); // hide when done
	if (Panel)
	{
		BWindow *Wnd = Panel->Window();
		if (Wnd->Lock())
		{
			Wnd->AddHandler(this);
			Wnd->Unlock();
		}
		
		Messenger = NEW(BMessenger(this));
		Panel->SetTarget(*Messenger);
		Panel->Show();
		Wait();
	}

	return strlen(FileStr) > 0;
}

bool GFileSelect::Open()
{
	return DoDialog(B_OPEN_PANEL);
}

bool GFileSelect::Save()
{
	return DoDialog(B_SAVE_PANEL);
}

///////////////////////////////////////////////////////////////////////////////////////////////
GFindReplace::GFindReplace(GWindow *Parent)
{
}

GFindReplace::~GFindReplace()
{
}

void GFindReplace::SetParent(GWindow *Parent)
{
}

bool GFindReplace::OpenFind()
{
	return FALSE;
}

bool GFindReplace::OpenReplace()
{
	return FALSE;
}

int GFindReplace::OnEvent(GMessage *Msg)
{
	int Status = FALSE;
	return Status;
}

bool GFindReplace::OnFind()
{
	return TRUE;
}

bool GFindReplace::OnReplace()
{
	return TRUE;
}

bool GFindReplace::CaseSensitive()
{
	return FALSE;
}

void GFindReplace::CaseSensitive(bool i)
{
}

bool GFindReplace::WholeWord()
{
	return 0;
}

void GFindReplace::WholeWord(bool i)
{
}

bool GFindReplace::BtnFindNext()
{
	return 0;
}

bool GFindReplace::BtnReplace()
{
	return 0;
}

bool GFindReplace::BtnReplaceAll()
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GClipBoard::GClipBoard(GWindow *o)
{
	Owner = o;
	Open = FALSE;
}

GClipBoard::~GClipBoard()
{
	if (Open)
	{
		Open = FALSE;
	}
}

bool GClipBoard::Empty()
{
	return FALSE;
}

bool GClipBoard::Text(char *Str)
{
	bool Status = FALSE;

	if (Str)
	{
	}

	return Status;
}

char *GClipBoard::Text()
{
	char *Str = NULL;

	return Str;
}

bool GClipBoard::Bitmap(GdcBasePrimitives *pDC)
{
	bool Status = FALSE;
	if (pDC)
	{
	}
	return Status;
}

GdcBasePrimitives *GClipBoard::Bitmap()
{
	GdcBasePrimitives *pDC = NULL;
	return pDC;
}
