/*hdr
**      FILE:           GuiDlg.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998 Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

#define TICKS_PER_SECOND					1000000
#define DOUBLE_CLICK_TIMEOUT				(TICKS_PER_SECOND/3)

///////////////////////////////////////////////////////////////////////////////////////////
extern rgb_color BeRGB(uchar r, uchar g, uchar b);

rgb_color BeRGB(uchar r, uchar g, uchar b)
{
	rgb_color c;
	c.red = r;
	c.green = g;
	c.blue = b;
	return c;
}

///////////////////////////////////////////////////////////////////////////////////////////
GDialog::GDialog() 
	: BView(BRect(0, 0, 1, 1), "Dialog", B_FOLLOW_ALL, B_WILL_DRAW),
	GWindow(this)
{
	ModalSem = 0;
	GWindow::Name("Dialog");
	IsResizable = FALSE;
	SetViewColor(216, 216, 216, 0);

	Wnd = NEW(GWnd(this, BRect(200, 200, 500, 500), "", B_MODAL_WINDOW, B_NOT_RESIZABLE));
	if (Wnd)
	{
		Wnd->SetLook(B_TITLED_WINDOW_LOOK);
		
		BRect r = Wnd->Bounds();
		ResizeTo(r.Width(), r.Height());
		Wnd->AddChild(this);
	}
}

GDialog::~GDialog()
{
}

bool GDialog::QuitRequested()
{
	if (ModalSem)
	{
		EndModal();
	}
	
	return TRUE;
}

GControl *GDialog::FindControl(int Id)
{
	Iterator<GWindow> List(&Children);
	for (GWindow *c = List.First(); c; c = List.Next())
	{
		GControl *Ctrl = dynamic_cast<GControl*>(c);
		if (Ctrl AND Ctrl->GetId() == Id)
		{
			return Ctrl;
		}
	}

	return FALSE;
}

void GDialog::MoveOnScreen()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);

	if (Pos.x2 >= Screen.X())
		Pos.Offset(Screen.X() - Pos.x2, 0);
	if (Pos.y2 >= Screen.Y())
		Pos.Offset(0, Screen.Y() - Pos.y2);
	if (Pos.x1 < 0)
		Pos.Offset(-Pos.x1, 0);
	if (Pos.y1 < 0)
		Pos.Offset(0, -Pos.y1);
}

void GDialog::MoveToCenter()
{
	GdcRegion Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	Pos.Offset(-Pos.x1, -Pos.y1);
	Pos.Offset((Screen.X() - Pos.X()) / 2, (Screen.Y() - Pos.Y()) / 2);
}

void GDialog::MoveToMouse()
{
	BPoint Pt;
	ulong Buttons;
	
	GetMouse(&Pt, &Buttons);
	Pos.Offset(-Pos.x1, -Pos.y1);
	Pos.Offset(Pt.x-(Pos.X()/2), Pt.y-(Pos.Y()/2));
}

status_t WaitForDelete(sem_id blocker)
{
	status_t	result;
	thread_id	this_tid = find_thread(NULL);
	BLooper		*pLoop;
	BWindow		*pWin = 0;

	pLoop = BLooper::LooperForThread(this_tid);
	if (pLoop)
		pWin = dynamic_cast<BWindow*>(pLoop);

	// block until semaphore is deleted (modal is finished)
	if (pWin)
	{
		do
		{
			// update the window periodically
			pWin->UpdateIfNeeded();
			result = acquire_sem_etc(blocker, 1, B_TIMEOUT, 50000);
		} while (result != B_BAD_SEM_ID);
	}
	else
	{
		do
		{
			// just wait for exit
			result = acquire_sem(blocker);
		} while (result != B_BAD_SEM_ID);
	}
	
	return result;
}

int GDialog::DoModal()
{
	ModalSem = create_sem(0, "ModalSem");

	// MFC NOTE: for further compatibility, we could put an
	// InitDialog() hook here
		
	if (Wnd->Lock())
	{
		Wnd->ResizeTo(Pos.X(), Pos.Y());
		Wnd->MoveTo(Pos.x1, Pos.y1);
		Wnd->SetTitle(GObject::Name());
	
		for (GWindow *w = Children.First(); w; w = Children.Next())
		{
			w->Attach(this);
		}
	
		Wnd->Show();
		Wnd->Unlock();
	}
	
	// MFC NOTE: WaitForDelete is different from CWnd::RunModalLoop
	// because it doesn't process messages.
	status_t Ret = WaitForDelete(ModalSem);

	if (Wnd->Lock())
	{
		Wnd->Quit();
	}
	
	return Ret;
}

int GDialog::DoModeless()
{
	if (Wnd->Lock())
	{
		Wnd->ResizeTo(Pos.X(), Pos.Y());
		Wnd->MoveTo(Pos.x1, Pos.y1);
		Wnd->SetTitle(GObject::Name());
	
		for (GWindow *w = Children.First(); w; w = Children.Next())
		{
			w->Attach(this);
		}
	
		Wnd->Show();
		Wnd->Unlock();
	}
	return FALSE;
}

int GDialog::OnEvent(BMessage *Msg)
{
	int Status = 0;

	// printf("GDialog::OnEvent(%i)\n", Msg->what);

	return Status;
}

void GDialog::EndModal(int Code)
{
	RemoveSelf();
	if (Wnd->Lock())
	{
		Wnd->Hide();
		Wnd->Unlock();
	}
	delete_sem(ModalSem); // causes DoModal to unblock
}

void GDialog::EndModeless(int Code)
{
	RemoveSelf();
	if (Wnd->Lock())
	{
		Wnd->Hide();
		Wnd->Quit();
	}
}

#include "resource.h"

#define DIALOG_X			1.56
#define DIALOG_Y			1.76
#define CTRL_X				1.50
#define CTRL_Y				1.64
#define Mx(x)				((x)*CTRL_X)
#define My(y)				((y)*CTRL_Y)

#define ES_AUTOHSCROLL	0x01
#define ES_MULTILINE	0x02
#define ES_AUTOVSCROLL	0x04
#define ES_WANTRETURN	0x08
#define WS_VSCROLL		0x10
#define IDC_STATIC -1

#define LTEXT(Str, Id, x, y, cx, cy)		Children.Insert(NEW(GText(Id, Mx(x), My(y), Mx(cx), My(cy), Str)))
#define PUSHBUTTON(Str, Id, x, y, cx, cy)	Children.Insert(NEW(GButton(Id, Mx(x), My(y), Mx(cx), My(cy), Str)))
#define EDITTEXT(Id, x, y, cx, cy, Flags)	Children.Insert(_EDITTEXT(Id, x, y, cx, cy, 0))
GEdit *_EDITTEXT(int Id, int x, int y, int cx, int cy, int Flags)
{
	GEdit *Edit = NEW(GEdit(Id, Mx(x), My(y), Mx(cx), My(cy), 0));
	if (Edit)
	{
		if (Flags & ES_MULTILINE)
		{
			Edit->MultiLine(TRUE);
		}
		
		if (Flags & ES_WANTRETURN)
		{
			Edit->WantReturn(TRUE);
		}
	}
}

bool GDialog::LoadFromResource(char *Resource)
{
	if (NOT StrICmp(Resource, "Contact"))
	{
		Pos.Dimension(270.0 * DIALOG_X, 266.0 * DIALOG_Y);
		GWindow::Name("Contact");

		EDITTEXT(IDC_FIRST,54,7,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_LAST,54,21,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_EMAIL,54,35,200,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_NICK,54,49,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_SPOUSE,174,49,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_STREET,54,78,200,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_SUBURB,54,92,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_POSTCODE,174,92,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_STATE,54,106,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_COUNTRY,174,106,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_WORK,54,142,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_MOBILE,174,142,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_HOME,54,156,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_FAX,174,156,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_ICQ,54,170,80,12,ES_AUTOHSCROLL);
		EDITTEXT(IDC_WEBPAGE,54,184,200,12,ES_AUTOHSCROLL);
		PUSHBUTTON("OK",IDOK,211,7,43,12);
		PUSHBUTTON("Cancel",IDCANCEL,211,21,43,11);
		LTEXT("First:",IDC_STATIC,35,9,16,8);
		LTEXT("Last:",IDC_STATIC,35,23,16,8);
		LTEXT("Email:",IDC_STATIC,31,37,20,8);
		// GROUPBOX("Address:",IDC_STATIC,7,65,255,62);
		LTEXT("Street:",IDC_STATIC,29,81,22,8);
		LTEXT("Suburb:",IDC_STATIC,25,94,26,8);
		LTEXT("PostCode:",IDC_STATIC,139,94,34,8);
		LTEXT("State:",IDC_STATIC,31,108,20,8);
		LTEXT("Country:",IDC_STATIC,146,109,27,8);
		// GROUPBOX("Contact:",IDC_STATIC,7,130,256,74);
		LTEXT("Work:",IDC_STATIC,30,142,20,8);
		LTEXT("Home:",IDC_STATIC,28,156,22,8);
		LTEXT("Mobile:",IDC_STATIC,148,142,24,8);
		LTEXT("Fax:",IDC_STATIC,157,156,14,8);
		LTEXT("ICQ:",IDC_STATIC,34,172,15,8);
		LTEXT("Webpage:",IDC_STATIC,15,185,34,8);
		LTEXT("Nickname:",IDC_STATIC,16,51,35,8);
		LTEXT("Spouse:",IDC_STATIC,144,51,27,8);
		LTEXT("Note:",IDC_STATIC,32,208,18,8);
		EDITTEXT(IDC_NOTE,54,207,200,52,ES_MULTILINE | ES_AUTOVSCROLL | ES_WANTRETURN | WS_VSCROLL);

		return TRUE;
	}
	
	return FALSE;
}

///////////////////////////////////////////////////////////////////////////////////////////
GControl::GControl(BView *view) : GWindow(view)
{
	Pos.ZOff(10, 10);
	SetId(0);
	Sunken(TRUE);
	Sys_LastClick = 0;
}

GControl::~GControl()
{
}

bool GControl::Attach(GWindow *parent)
{
	Parent = parent;
	if (Parent AND Sys_View)
	{
		Parent->Sys_View->AddChild(Sys_View);
		Sys_View->ResizeTo(Pos.X(), Pos.Y());
		Sys_View->MoveTo(Pos.x1, Pos.y1);
		return TRUE;
	}
	return 0;
}

int GControl::OnEvent(BMessage *Msg)
{
	return 0;
}

GdcPt2 GControl::SizeOfStr(char *Str)
{
	GdcPt2 Pt(0, 0);
	if (Str)
	{
		BFont Font;
		Sys_View->GetFont(&Font);
		
		font_height Ht;
		Font.GetHeight(&Ht);
		
		Pt.y = Ht.ascent + Ht.leading;
		Pt.x = Font.StringWidth(Str);
	}

	return Pt;
}

void GControl::MouseClickEvent(bool Down)
{
	VMouse m;
	BPoint Pt;
	uint32 Buttons;
	if (GetBView()->LockLooper())
	{
		bigtime_t Now = system_time();

		GetBView()->GetMouse(&Pt, &Buttons, FALSE);
		m.x = Pt.x;
		m.y = Pt.y;
		m.Flags = Buttons;

		if (Down)
		{
			uint32 *INow = (uint32*) &Now;
			// printf("Now: %08.8X %08.8X\n", INow[0], INow[1]);
			uint32 *ISys = (uint32*) &Sys_LastClick;
			// printf("Sys: %08.8X %08.8X\n", ISys[0], ISys[1]);
			uint32 Diff = INow[0]-ISys[0]; // urrrrrrhhhh!
			// printf("Diff: %08.8X (%u)\n", Diff, Diff);

			m.Double = Diff < DOUBLE_CLICK_TIMEOUT;
			Sys_LastClick = Now;

			// printf("MouseClickEvent() Down(%i) Double(%i)\n", Down, m.Double);
		}
		else
		{
			m.Double = false;
		}
		
		m.Down = Down;
		OnMouseClick(m);
		GetBView()->UnlockLooper();
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
// Button
GButton::GButton(int id, int x, int y, int cx, int cy, char *name)
	: GControl(this),
	BButton(BRect(0, 0, 1, 1), "Button", "", NEW(BMessage(WM_CHANGE)))
{
	GWindow::Name(name);
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	((BControl*) Sys_View)->SetLabel(GObject::Name());
}

GButton::~GButton()
{
}

status_t GButton::Invoke(BMessage *message)
{
	status_t Ret = BButton::Invoke(message);

	GWindow *n = (GWindow::Notify) ? GWindow::Notify : GWindow::Parent;
	if (n)
	{
		n->OnNotify(this, 0);
	}
	
	return Ret;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Edit
GEdit::GEdit(int id, int x, int y, int cx, int cy, char *name)
	: GControl(this),
	BTextControl(BRect(0, 0, 1, 1), "Edit", "", (name)?name:"", NEW(BMessage(WM_CHANGE)))
{
	GdcPt2 Size = SizeOfStr((name)?name:"A");
	if (cx < 0) cx = Size.x + 6;
	if (cy < 0) cy = Size.y + 4;

	MultiLineEdit = FALSE;
	PasswordEdit = FALSE;
	SetId(id);
	if (name)
	{
		GWindow::Name(name);
	}
	
	Pos.Set(x, y, x+max(cx, 10), y+max(cy, 10));
}

GEdit::~GEdit()
{
}

void GEdit::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	GWindow::Name(Str);
}

int GEdit::Value()
{
	char *n = GWindow::Name();
	return (n) ? atoi(n) : 0;
}

char *GEdit::Name()
{
	return Text();
}

bool GEdit::Name(char *s)
{
	SetText(s);
	return true;
}

int GEdit::OnEvent(BMessage *Msg)
{
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Check box
GCheckBox::GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState)
	: GControl(this),
	BCheckBox(BRect(0,0,1,1), "Check", (name)?name:"", NEW(BMessage(WM_CHANGE))) 
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	Val = InitState;
	Three = FALSE;
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
}

GCheckBox::~GCheckBox()
{
}

int GCheckBox::Value()
{
	Val = Value();
	return Val;
}

void GCheckBox::Value(int b)
{
	Val = b;
	SetValue(Val);
}

int GCheckBox::OnEvent(BMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
// Text
GText::GText(int id, int x, int y, int cx, int cy, char *name)
	: GControl(this),
	BStringView(BRect(0, 0, 1, 1), "Text", "")
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x;
	if (cy < 0) cy = Size.y;

	if (name)
	{
		GWindow::Name(name);
		SetText(name);
	}
	
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
}

GText::~GText()
{
}

void GText::Draw(BRect updateRect)
{
	int y = 1;
	char *c = Text();
	if (c)
	{
		char *Start = c;
		font_height FntH;
	
		Sys_View->GetFontHeight(&FntH);
		int Ht = FntH.ascent + FntH.descent + FntH.leading + 1;
	
		while (*c)
		{
			if (*c == '\n')
			{
				int32 Len = (int32) c - (int32) Start;
	
				Sys_View->MovePenTo(0, y * Ht);
				Sys_View->DrawString(Start, Len);
				Start = c + 1;
				y++;
			}
			
			c++;
		}
	
		int32 Len = (int32) c - (int32) Start;
		Sys_View->MovePenTo(0, y * Ht);
		Sys_View->DrawString(Start, Len);
	}
}

int GText::OnEvent(BMessage *Msg)
{
	return 0;
}


///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
int GRadioGroup::NextId = 20000;

GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init)
	: GControl(this)
	, BBox(BRect(0, 0, 1, 1), "Group")
{
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	InitialValue = Init;

	if (name)
	{
		GWindow::Name(name);
		SetLabel(name);
	}
}

GRadioGroup::~GRadioGroup()
{
/*
	for (	GRadioButton *But = Buttons.First();
		But;
		But = Buttons.Next())
	{
		DeleteObj(But);
	}
*/
}

bool GRadioGroup::Attach(GWindow *parent)
{
	bool Status = GControl::Attach(parent);
	if (Status)
	{
		for (	GRadioButton *But = Buttons.First();
			But;
			But = Buttons.Next())
		{
			But->Attach(this);
		}
		Value(InitialValue);
	}
	return Status;
}

int GRadioGroup::Value()
{
	int i=0;
	for (	GRadioButton *But = Buttons.First();
		But;
		But = Buttons.Next(), i++)
	{
		if (But->Value()) return i;
	}
	return 0;
}

void GRadioGroup::Value(int i)
{
	GRadioButton *But = Buttons[i];
	if (But)
	{
		But->Value(TRUE);
	}
}

int GRadioGroup::OnEvent(BMessage *Msg)
{
	return 0;
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	GRadioButton *But = NEW(GRadioButton(NextId++, x, y, -1, -1, name, Buttons.GetItems() == 0));
	if (But)
	{
		Buttons.Insert(But);
	}
	return But;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first)
	: GControl(this)
	, BRadioButton(BRect(0,0,1,1), "RadioBtn", (name)?name:"", NEW(BMessage(WM_CHANGE)))
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	First = first;
	Pos.Set(x, y, x+cx, y+cy);
	SetId(id);
	if (name)
	{
		GWindow::Name(name);
	}
}

GRadioButton::~GRadioButton()
{
}

int GRadioButton::Value()
{
	return BRadioButton::Value();
}

void GRadioButton::Value(int i)
{
	BRadioButton::SetValue(i);
}

int GRadioButton::OnEvent(BMessage *Msg)
{
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////
// Tab Control
GTabControl::GTabControl(int id, int x, int y, int cx, int cy, char *name, int Init)
	: GControl(this)
	, BTabView(BRect(0, 0, 1, 1), "Tab")
{
	SetId(id);
	Pos.ZOff(cx, cy);
	Pos.Offset(x, y);
	GWindow::Name(name);
	InitialValue = Init;
	Current = 0;
	NextTab = 0;
	SetViewColor(216,216,216,0);
}

GTabControl::~GTabControl()
{
}

GControl *GTabControl::FindControl(int Id)
{
	GControl *Ctrl = 0;

	if (GetId() == Id)
	{
		return this;
	}

	for (	GTabPage *Page = Tabs.First();
		Page AND NOT Ctrl;
		Page = Tabs.Next())
	{
		Iterator<GControl> Ctrls(&Page->Children);
		for (GControl *c = Ctrls.First(); c AND NOT Ctrl; c = Ctrls.Next())
		{
			Ctrl = c->FindControl(Id);
		}
	}
	return Ctrl;
}

void GTabControl::Select(int32 tab)
{
	BTabView::Select(tab);

	BRect r = Bounds();
	r.bottom -= TabHeight();
	
	BTab *Tab = TabAt(Selection());
	if (Tab)
	{
		BView *View = Tab->View();
		if (View)
		{
			View->MoveTo(r.left, r.top);
			View->ResizeTo(r.Width(), r.Height());
		}
	}
}

bool GTabControl::Attach(GWindow *parent)
{
	bool Status = GControl::Attach(parent);
	if (Status)
	{
		int i=0;
		Iterator<GTabPage> TabList(&Tabs);
		for (	GTabPage *Page = TabList.First();
				Page;
				Page = TabList.Next(), i++)
		{
			Page->Attach(this);
		}
		
		Value(0);
	}

	return Status;
}

int GTabControl::Value()
{
	Current = Selection();
	return Current;
}

void GTabControl::Value(int i)
{
	Select(i);
	Current = Selection();
}

int GTabControl::OnEvent(BMessage *Msg)
{
	return 0;
}

GTabPage *GTabControl::Append(char *name)
{
	GTabPage *Page = NEW(GTabPage(name));
	if (Page)
	{
		Tabs.Insert(Page);
	}
	return Page;
}

//////////////////////////////////////////////////////////////////////////////////
GTabPage::GTabPage(char *name) :
	GWindow(this),
	BView(BRect(0, 0, 1, 1), "TabPage", B_FOLLOW_ALL_SIDES, 0)
{
	Parent = 0;
	Pos.ZOff(0, 0);
	SetViewColor(216, 216, 216, 0);
	Name(name);
	Tab = 0;
}

GTabPage::~GTabPage()
{
}



bool GTabPage::Attach(GWindow *parent)
{
	bool Status = FALSE;
	Parent = dynamic_cast<GTabControl*>(parent);
	if (Parent)
	{
		BRect r = Parent->Bounds();
		r.bottom -= Parent->TabHeight();
		MoveTo(r.left, r.top);
		ResizeTo(r.Width(), r.Height());
		
		Tab = NEW(BTab);
		if (Tab)
		{
			Parent->AddTab(this, Tab);
			Tab->SetLabel(Name());

			for (GWindow *w = Children.First(); w; w = Children.Next())
			{
				w->Attach(this);
			}

			Status = TRUE;
		}
	}

	return Status;
}

char *GTabPage::Name()
{
	return GObject::Name();
}

bool GTabPage::Name(char *name)
{
	bool Status = GObject::Name(name);
	if (name)
	{
		SetName(name);
	}
	return Status;
}

int GTabPage::OnEvent(BMessage *Msg)
{
	return 0;
}

bool GTabPage::Enabled()
{
	return Tab->IsEnabled();
}

void GTabPage::Enabled(bool e)
{
	Tab->SetEnabled(e);
}

bool GTabPage::Visible()
{
	return TRUE;
}

void GTabPage::Visible(bool v)
{
	for (	GWindow *w = Children.First();
			w;
			w = Children.Next())
	{
		w->Visible(v);
	}
}

void GTabPage::Append(GWindow *Wnd)
{
	Children.Insert(Wnd);
}

bool GTabPage::AppendDialogFromResource(char *Name)
{
	return FALSE;
}

//////////////////////////////////////////////////////////////////////////////////
// Slider control
GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert) :
	GControl(this),
	BScrollBar(BRect(0, 0, 1, 1), "Scroll", NULL, 0, 1, (vert)?B_VERTICAL:B_HORIZONTAL)
{
	SetId(id);
	if (vert)
	{
		Pos.Set(x, y, x+13, y+cy);
	}
	else
	{
		Pos.Set(x, y, x+cx, y+13);
	}
	
	GWindow::Name(name);
	Vertical = vert;
	Min = Max = 0;
	Val = 0;
}

GSlider::~GSlider()
{
}

void GSlider::Value(int i)
{
	Val = i;
	SetValue(i);
}

int GSlider::Value()
{
	Val = Value();
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	float a, b;
		
	GetRange(&a, &b);
	min = Min = a;
	max = Max = b;
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
	SetRange(Min, Max);
}

int GSlider::OnEvent(BMessage *Msg)
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////////////
GCombo::GCombo(int id, int x, int y, int cx, int cy, char *name) :
	GControl(this),
	BMenuField(BRect(0, 0, 1, 1), "Combo", "Blah", new BMenu(""))
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	SortItems = FALSE;
}

GCombo::~GCombo()
{
}

char *GCombo::Name()
{
	return MenuItem()->Label();
}

bool GCombo::Name(char *n)
{
	MenuItem()->SetLabel(n);
	return TRUE;
}

void GCombo::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GCombo::Value()
{
	return atoi(Name());
}

int GCombo::OnEvent(BMessage *Msg)
{
	if (Msg->what == WM_CHANGE)
	{
		BMenuItem *Item = 0;
		Msg->FindPointer("source", (void**) &Item);
		if (Item)
		{
			MenuItem()->SetLabel(Item->Label());
		}
	}
	return 0;
}

bool GCombo::Delete()
{
	int32 Index = (Cur) ? FindIndex(Cur->Ptr) : -1;
	bool Status = DLinkList::Delete();

	if (Status AND Index >= 0)
	{
		Menu()->RemoveItem(Index);
	}

	return Status;
}

bool GCombo::Delete(int i)
{
	bool Status = DLinkList::Delete(i);

	if (Status)
	{
		Menu()->RemoveItem((int32) i);
	}

	return Status;
}

bool GCombo::Delete(char *p)
{
	int32 Index = FindIndex(p);
	bool Status = DLinkList::Delete((void*)p);

	if (Status AND Index >= 0)
	{
		Menu()->RemoveItem(Index);
	}

	return Status;
}

void GCombo::OnCreate()
{
	BMenuField::AttachedToWindow();
	BMenu *m = Menu();
	if (m)
	{
		for (int32 i=0; i<m->CountItems(); i++)
		{
			BMenuItem *Item = m->ItemAt(i);
			Item->SetTarget(this);
		}
	}

	MenuItem()->SetLabel(Menu()->ItemAt(0)->Label());
}

bool GCombo::Insert(char *p, int Index)
{
	if (NOT p) p = "";
	bool Status = DLinkList::Insert((void*)p, Index);
	BMessage *Msg = 0;
	BMenuItem *Item = NEW(BMenuItem(p, Msg = NEW(BMessage(WM_CHANGE))));
	if (Item)
	{
		Menu()->AddItem(Item, (Index<0)?Menu()->CountItems():Index);
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////
GBitmap::GBitmap(int id, int x, int y, char *FileName)
	: GControl(this),
	BControl(	BRect(0, 0, 1, 1),
				NULL, "",
				NULL,
				B_FOLLOW_NONE,
				B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE)
{
	SetId(id);
	Pos.ZOff(1, 1);
	Pos.Offset(x, y);
	pDC = 0;
	SetViewColor(216, 216, 216, 0);

	GdcBmp F;
	if (F.Open(FileName, O_READ))
	{
		pDC = NEW(GdcMemDC);
		if (pDC AND F.ReadImage(pDC))
		{
			Pos.Dimension(pDC->X()+4, pDC->Y()+4);
		}
		else
		{
			DeleteObj(pDC);
		}
	}
}

GBitmap::~GBitmap()
{
	DeleteObj(pDC);
}

int GBitmap::OnEvent(BMessage *Msg)
{
	return 0;
}

void GBitmap::OnPaint(GdcBasePrimitives *pScreen)
{
	GdcRegion a(0, 0, Pos.X()-1, Pos.Y()-1);
	DrawMoveDeepEdge(pScreen, a, SUNKEN);
	if (pDC)
	{
		pScreen->Blt(2, 2, pDC);
	}
	else
	{
		pScreen->Colour(CBit(pScreen->GetBits(), Rgb24(255, 255, 255)));
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(VMouse &m)
{
	if (NOT m.Down AND GWindow::Parent)
	{
		GDialog *p = (GDialog*) GWindow::Parent;
		p->OnNotify(this, 0);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
// List control
GListItem::GListItem() :
	GItem(15.0)
{
	Parent = 0;
	Cols = 0;
	Str = 0;
}

GListItem::~GListItem()
{
	if (Str)
	{
		for (int i=0; i<Cols; i++)
		{
			DeleteArray(Str[i]);
		}
		DeleteArray(Str);
	}
}

bool GListItem::GetBitmapInfo(BBitmap **bmp, BRect *rect)
{
	int Image = GetImage(IsSelected());
	if (Image >= 0 AND bmp AND rect)
	{
		if (Parent AND Parent->ImageList)
		{
			*bmp = Parent->ImageList->Handle();
			rect->left = Image * 16;
			rect->top = 0;
			rect->right = rect->left + 15;
			rect->bottom = 15;
			return true;
		}
	}
	return false;
}

void GListItem::Update()
{
	if (Parent)
	{
		int32 i = Parent->IndexOf(this);
		Parent->InvalidateItem(i);
	}
}

bool GListItem::Selected()
{
	return IsSelected();
}

void GListItem::Selected(bool b)
{
	if (b)
	{
		Select();
	}
	else
	{
		Deselect();
	}
}

bool GListItem::SetText(char *s, int i)
{
	SetColumnContent(i, s, false); 
}

char *GListItem::GetText(int i)
{
	return GetColumnContentText(i);
}

////////////////////////////////////////////////////////////////////////////////////////////
GList::GList(int id, int x, int y, int cx, int cy, char *name) :
	GControl(this),
	ColumnListView(	BRect(x, y, x+cx-14, y+cy-14),
					&ScrollView,
					"List",
					B_FOLLOW_LEFT | B_FOLLOW_TOP, 
					B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE, 
					B_SINGLE_SELECTION_LIST,
					false,
					true,
					true,
					true,
					B_NO_BORDER)
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) GWindow::Name(name);
	Val = -1;
	EditLabels = 0;
}

GList::~GList()
{
	Empty();
}

bool GList::SetPos(GdcRegion &p, bool Repaint)
{
	if (LockLooper())
	{
		ScrollView->MoveTo(p.x1, p.y1);
		ScrollView->ResizeTo(p.X()-1, p.Y()-1);
		// BListView::ResizeTo(p.X()-17, p.Y()-17);

		Pos = p;
		OnPosChange();
		UnlockLooper();
	}
	
	
	return TRUE;
}

bool GList::Attach(GWindow *parent)
{
	GWindow::Parent = parent;
	if (GWindow::Parent)
	{
		GWindow::Parent->GetBView()->AddChild(ScrollView);

		if (LockLooper())
		{
			if (AskImage())
			{
				ColumnListView::AddColumn(new CLVColumn("", 18, CLV_NOT_MOVABLE), 0);
			}
			UnlockLooper();
		}
	
		ScrollView->MoveTo(Pos.x1, Pos.y1);
		ScrollView->ResizeTo(Pos.X(), Pos.Y());
		return TRUE;
	}
	return FALSE;
}

void GList::AddColumn(char *Name, int Width, int Align)
{
	bool Unlock = LockLooper();
	ColumnListView::AddColumn(new CLVColumn((Name)?Name:"", Width, CLV_SORT_KEYABLE));
	if (Unlock) UnlockLooper();
}

void GList::EmptyColumns()
{
	bool Unlock = LockLooper();
	int Count = CountColumns() - (AskImage() ? 1 : 0);
	if (Count > 0)
	{
		CLVColumn *Col = ColumnAt(AskImage() ? 1 : 0);
		RemoveColumns(Col, Count);
	}
	if (Unlock) UnlockLooper();
}

void GList::ShowColumnHeader(bool Show)
{
	ColumnHeader = Show;
}

bool GList::ShowColumnHeader()
{
	return ColumnHeader;
}

void GList::SelectionChanged()
{
	// printf("GList::SelectionChanged()\n");

	int32 Index = CurrentSelection();
	if (Index >= 0)
	{
		GListItem *Item = (GListItem*) ItemAt(Index);
		if (Item)
		{
			OnItemSelect(Item);
		}
	}
}

void GList::MouseDown(BPoint point)
{
	MouseClickEvent(true);
	BListView::MouseDown(point);
	MouseClickEvent(false);
}

void GList::OnMouseClick(VMouse &m)
{
	// printf("GList::OnMouseClick(...) X(%i) Y(%i) Flags(%i) Down(%i)\n", m.x, m.y, m.Flags, m.Down);

	if (LockLooper())
	{
		int32 Index = CurrentSelection();
		if (Index >= 0)
		{
			GListItem *Item = (GListItem*) ItemAt(Index);
			if (Item)
			{
				OnItemClick(Item, m);
			}
		}
		
		UnlockLooper();
	}
}

int GList::OnEvent(BMessage *Msg)
{
	int Status = FALSE;
	return Status;
}

bool GList::Remove(GListItem *Obj)
{
	bool Status = FALSE;

	if (LockLooper())
	{
		if (Obj AND HasItem(Obj))
		{
			RemoveItem(Obj);
			Status = true;
		}

		UnlockLooper();
	}

	return Status;
}

bool GList::Delete()
{
	bool Status = FALSE;
	return Status;
}

bool GList::Delete(int i)
{
	bool Status = false;

	if (LockLooper())
	{
		BListItem *Item = RemoveItem(i);
		if (Item)
		{
			DeleteObj(Item);
			Status = true;
		}

		UnlockLooper();
	}
	
	return Status;
}

bool GList::Delete(GListItem *Item)
{
	bool Status = false;

	if (Item AND HasItem(Item))
	{
		BListView::RemoveItem(Item);
		DeleteObj(Item);
		Status = true;
	}
	
	return Status;
}

bool GList::Insert(GListItem *p, int Index)
{
	bool Status = FALSE;
	if (p)
	{
		bool Unlock = LockLooper();
	
		p->Parent = this;
		if (Index < 0)
		{
			AddItem(p);
		}
		else
		{
			AddItem(p, Index);
		}
		Status = true;
		
		if (Unlock) UnlockLooper();
	}	
	return Status;
}

GListItem *GList::First()
{
	return (GListItem*) FirstItem();
}

GListItem *GList::Last()
{
	return (GListItem*) LastItem();
}

GListItem *GList::Next()
{
	return 0;
}

GListItem *GList::Prev()
{
	return 0;
}

void GList::Empty()
{
	if (LockLooper())
	{
		GListItem *Item = First();
		while (Item)
		{
			RemoveItem(Item);
			DeleteObj(Item);
			Item = First();
		}

		UnlockLooper();
	}
}

void GList::RemoveAll()
{
	if (LockLooper())
	{
		RemoveItems(0, CountItems());
		UnlockLooper();
	}
}

int GList::Value()
{
	return Val;
}

void GList::Value(int n)
{
	Val = n;
}

GListItem *GList::Current()
{
	int32 Index = (GListItem*) CurrentSelection();
	BListItem *i = (Index >= 0) ? ItemAt(Index) : 0;
	return (GListItem*) i;
}

void GList::OnItemClick(GListItem *Item, VMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GList::OnItemBeginDrag(GListItem *Item, int Flags)
{
	if (Item)
	{
		VMouse m;
		ZeroObj(m);
		m.Flags = Flags;
		Item->OnBeginDrag(m);
	}
}

bool GList::Select(GListItem *Obj)
{
	if (Obj)
	{
		BListView::Select(IndexOf(Obj));
		return true;
	}

	return false;
}

bool GList::GetAll(List<GListItem> &l)
{
	for (int i=0; i<CountItems(); i++)
	{
		l.Insert((GListItem*) ItemAt(i));
	}

	return true;
}

bool GList::GetSelection(List<GListItem> &l)
{
	int Index = 0;
	for (int i=0; (Index = CurrentSelection(i)) >= 0; i++)
	{
		l.Insert((GListItem*)ItemAt(Index));
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////
GTreeItem::GTreeItem() :
	GItem(15.0)
{
	Str = 0;
	Tree = 0;
	Parent = 0;
}

GTreeItem::~GTreeItem()
{
	DeleteArray(Str);
}

GTreeItem *GTreeItem::GetChild()
{
	if (Tree)
	{
		BListItem *Bi = Tree->ItemUnderAt(this, TRUE, 0);
		GTreeItem *Ti = dynamic_cast<GTreeItem*>(Bi);
		return Ti;
	}
	return 0;
}

GTreeItem *GTreeItem::GetNext()
{
	if (Tree)
	{
		int MyIndex = Tree->IndexOf(this);
		BListItem *Super = Tree->Superitem(this);
		BListItem *Bi = 0;
		if (Super)
		{
			Bi = Tree->ItemUnderAt(Super, TRUE, MyIndex + 1);
		}
		else
		{
			Tree->ItemAt(MyIndex + 1);
		}
		GTreeItem *Ti = dynamic_cast<GTreeItem*>(Bi);
		return Ti;
	}
	return 0;
}

char *GTreeItem::GetText(int i)
{
	return Str;
}

bool GTreeItem::SetText(char *s, int i)
{
	DeleteArray(Str);
	Str = NewStr(s);
	return Str != 0;
}

int GTreeCompareFunc(const BListItem *a, const BListItem *b)
{
	return ((GTreeItem*)a)->Index - ((GTreeItem*)b)->Index;
}

GTreeItem *GTreeItem::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = (Obj) ? Obj : NEW(GTreeItem);
	if (NewObj AND Tree)
	{
		if (Tree->LockLooper())
		{
			Obj->Index = Tree->CountItemsUnder(this, true);
			Obj->Tree = Tree;
			Tree->AddUnder(Obj, this);
			Tree->SortItemsUnder(this, true, GTreeCompareFunc);
			Tree->UnlockLooper();
		}
	}
	return NewObj;
}

void GTreeItem::Update()
{
}

bool GTreeItem::Selected()
{
	return IsSelected();
}

void GTreeItem::Selected(bool b)
{
	if (Tree)
	{
		Tree->BOutlineListView::Select(Tree->FullListIndexOf(this));
	}
}

bool GTreeItem::Expanded()
{
	return IsExpanded();
}

void GTreeItem::Expanded(bool b)
{
	SetExpanded(b);
}

void GTreeItem::DrawItem(BView *owner, BRect frame, bool complete)
{
	rgb_color color;
	if (IsSelected())
	{
		color.red = 192;
		color.green = 192;
		color.blue = 192;
	}
	else
	{
		color = owner->ViewColor();
	}

	owner->SetHighColor(color);
	owner->SetLowColor(color);
	owner->FillRect(frame);
	
	int Image = -1;
	int x = 4;
	if (Tree->ImageList)
	{
		Image = GetImage(Selected());
		x += 16;
	}

	if (Str)
	{
		owner->MovePenTo(frame.left+x, frame.bottom-2);
		owner->SetHighColor(BeRGB(0, 0, 0));
		owner->DrawString(Str);
	}

	if (Image >= 0)
	{
		int x = Image * 16;
		BBitmap *Bmp = Tree->ImageList->Handle();
		BRect S(x, 0, x+15, 15);
		BRect D(frame.left, frame.top, frame.left+15, frame.top+15);
		drawing_mode Mode = owner->DrawingMode();
		owner->SetDrawingMode(B_OP_OVER);
		owner->DrawBitmap(Bmp, S, D);
		owner->SetDrawingMode(Mode);
	}
}

//////////////////////////////////////////////////////////////////////////////
GTree::GTree(int id, int x, int y, int cx, int cy, char *name) :
	GControl(this),
	BOutlineListView(	BRect(x, y, x+cx-14, y+cy-14),
						"List",
						B_SINGLE_SELECTION_LIST)
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	if (name) GControl::Name(name);

	Lines = TRUE;
	Buttons = TRUE;
	LinesAtRoot = TRUE;
	EditLabels = FALSE;
	ScrollView = 0;

}

GTree::~GTree()
{
}

bool GTree::SetPos(GdcRegion &p, bool Repaint)
{
	ScrollView->MoveTo(p.x1, p.y1);
	ScrollView->ResizeTo(p.X()-1, p.Y()-1);
	ResizeTo(p.X()-16, p.Y()-16);
	Pos = p;
	return TRUE;
}

bool GTree::Attach(GWindow *parent)
{
	GWindow::Parent = parent;
	if (GWindow::Parent)
	{
		ScrollView = NEW(BScrollView(	"scroll_cities",
										this,
										B_FOLLOW_LEFT | B_FOLLOW_TOP,
										0,
										TRUE,
										TRUE,
										B_NO_BORDER));
		if (ScrollView)
		{
			GWindow::Parent->GetBView()->AddChild(ScrollView);
			ScrollView->MoveTo(Pos.x1, Pos.y1);
			ScrollView->ResizeTo(Pos.X() + 3, Pos.Y() + 3);

			// standard white background where we don't draw
			ScrollView->SetViewColor(255, 255, 255, 0);

			return TRUE;
		}
	}
	return FALSE;
}

void GTree::MouseDown(BPoint point)
{	
	MouseClickEvent(true);
	BOutlineListView::MouseDown(point);
	MouseClickEvent(false);
}

void GTree::SelectionChanged()
{
	// printf("GTree::SelectionChanged()\n");

	int32 Index = FullListCurrentSelection();
	if (Index >= 0)
	{
		GTreeItem *Item = (GTreeItem*) FullListItemAt(Index);
		if (Item)
		{
			OnItemSelect(Item);
		}
	}
}

void GTree::OnMouseClick(VMouse &m)
{
	// printf("GTree::OnMouseClick(...) X(%i) Y(%i) Flags(%i) Down(%i)\n", m.x, m.y, m.Flags, m.Down);

	if (LockLooper())
	{
		int32 Index = FullListCurrentSelection();
		if (Index >= 0)
		{
			GTreeItem *Item = (GTreeItem*) FullListItemAt(Index);
			if (Item)
			{
				OnItemClick(Item, m);
			}
		}
		
		UnlockLooper();
	}
}

int GTree::OnEvent(BMessage *Msg)
{
	return FALSE;
}

GTreeItem *GTree::Insert(GTreeItem *Obj, int Pos)
{
	GTreeItem *NewObj = (Obj) ? Obj : NEW(GTreeItem);
	if (NewObj)
	{
		NewObj->Tree = this;
		if (LockLooper())
		{
			AddItem(NewObj, (Pos < 0) ? 100000000 : Pos);
			UnlockLooper();
		}
	}
	return NewObj;
}

bool GTree::Remove(GTreeItem *Obj)
{
	if (Obj)
	{
		RemoveItem(Obj);
		return TRUE;
	}
	return FALSE;
}

void GTree::RemoveAll()
{
	MakeEmpty();
}

bool GTree::Delete(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj)
	{
		RemoveItem(Obj);
	}
	return Status;
}

bool GTree::Select(GTreeItem *Obj)
{
	bool Status = FALSE;
	if (Obj)
	{
		int32 Index = FullListIndexOf(Obj);
		BListView::Select(Index);
	}
	return Status;
}

void GTree::OnItemClick(GTreeItem *Item, VMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GTree::OnItemBeginDrag(GTreeItem *Item, int Flags)
{
	if (Item)
	{
		VMouse m;
		ZeroObj(m);
		m.Flags = Flags;
		Item->OnBeginDrag(m);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
GItem::GItem(float minheight) :
	CLVEasyItem(0, false, false, minheight)
{
}

GItemContainer::GItemContainer()
{
	Flags = 0;
	ImageList = 0;
}

GItemContainer::~GItemContainer()
{
	if (OwnList())
	{
		DeleteObj(ImageList);
	}
	else
	{
		ImageList = 0;
	}
}

bool GItemContainer::SetImageList(GImageList *list, bool Own)
{
	ImageList = list;
	OwnList(Own);
	AskImage(ImageList != NULL);
	return ImageList != NULL;
}

bool GItemContainer::Load(char *File)
{
	bool Status = FALSE;
	GdcBmp f;
	if (f.Open(File, O_READ))
	{
		GdcMemDC *pDC = NEW(GdcMemDC);
		if (pDC)
		{
			Status = f.ReadImage(pDC);
		}
	}
	return Status;
}

///////////////////////////////////////////////////////////////////////////////////////////
GProgress::GProgress(int id, int x, int y, int cx, int cy, char *name) :
	GControl(this),
	BStatusBar(BRect(0, 0, 1, 1), "Progress", "")
{
	SetId(id);
	Pos.Set(x, y, x+cx, y+cy);
	//if (name) GControl::Name(name);
}

GProgress::~GProgress()
{
}

void GProgress::AttachedToWindow(void)
{
	BStatusBar::AttachedToWindow();
	SetBarHeight(10);
	SetMaxValue(High - Low);
	Update(Val - Low);
}

void GProgress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
	SetMaxValue(h - l);
}

void GProgress::GetLimits(int &l, int &h)
{
	l = Low;
	h = High;
}

void GProgress::Value(int v)
{
	Val = v;
	if (BView::Parent())
	{
		Update((Val - Low) - CurrentValue());
	}
}

int GProgress::Value()
{
	if (BView::Parent())
	{
		Val = Low + CurrentValue();
	}
	return Val;
}

int GProgress::OnEvent(BMessage *Msg)
{
	return 0;
}


