/*hdr
**      FILE:           GuiMenu.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           18/7/98
**      DESCRIPTION:    Gui menu system
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

///////////////////////////////////////////////////////////////////////////////////////////////
extern char *NewStrLessAnd(char *s);
char *NewStrLessAnd(char *s)
{
	char *ns = NewStr(s);
	if (ns)
	{
		char *In = ns;
		char *Out = ns;
		while (*In)
		{
			if (*In != '&')
			{
				*Out++ = *In;
			}
			In++;
		}
		*Out = 0;
	}
	return ns;
}

///////////////////////////////////////////////////////////////////////////////////////////////
char *Temp = 0;

GSubMenu::GSubMenu(char *name) : BMenu(Temp = NewStrLessAnd(name))
{
	DeleteArray(Temp);
	Window = 0;
}

GSubMenu::~GSubMenu()
{
}

GMenuItem *GSubMenu::AppendItem(char *Str, int Id, bool Enabled)
{
	GMenuItem *Item = NEW(GMenuItem());
	if (Item)
	{
		if (Window AND Window->Sys_View)
		{
			Item->SetTarget(Window->Sys_View);
		}
		AddItem(Item);
		
		char *s = NewStrLessAnd(Str);
		if (s)
		{
			Item->Name(s);
			DeleteArray(s);
		}

		Item->Id(Id);
		Item->Enabled(Enabled);
	}
	return Item;
}

GMenuItem *GSubMenu::AppendSeparator()
{
	AddSeparatorItem();
	return NULL;
}

GSubMenu *GSubMenu::AppendSub(char *Str)
{
	GSubMenu *Sub = NEW(GSubMenu(Str));
	if (Sub)
	{
		Sub->Window = Window;
		AddItem(Sub);
	}

	return Sub;
}

int GSubMenu::Float(GWindow *Parent, int x, int y)
{
	BPopUpMenu *Popup = NEW(BPopUpMenu("PopUpMenu"));
	if (Popup)
	{
		for (int n=0; n<CountItems(); n++)
		{
			GMenuItem *i = (GMenuItem*) ItemAt(n);
			if (i)
			{
				BSeparatorItem *Sep = dynamic_cast<BSeparatorItem*>(i);
				if (Sep)
				{
					Popup->AddSeparatorItem();
				}
				else
				{
					GMenuItem *Item = NEW(GMenuItem);
					if (Item)
					{
						Item->Id(i->Id());
						Item->Name(i->Label());
						Popup->AddItem(Item);
					}
				}
			}
		}	
	
		BPoint Pt(x-20, y-20);
		GMenuItem *Item = dynamic_cast<GMenuItem*>(Popup->Go(Pt));
		if (Item)
		{
			return Item->Id();
		}
	}
	
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GMenuItem::GMenuItem() : BMenuItem("<error>", Msg = NEW(BMessage(WM_MENU)))
{
}

GMenuItem::GMenuItem(GSubMenu *p) : BMenuItem(p)
{
	Msg = 0;
}

GMenuItem::~GMenuItem()
{
}

void GMenuItem::Id(int i)
{
	if (Msg)
	{
		Msg->AddInt32("Cmd", i);
	}
}

int GMenuItem::Id()
{
	if (Msg)
	{
		int32 i;
		if (Msg->FindInt32("Cmd", &i) == B_OK)
		{
			return i;
		}
	}
	return 0;
}

void GMenuItem::Checked(bool c)
{
	SetMarked(c);
}

bool GMenuItem::Name(char *n)
{
	bool Status = GObject::Name(n);
	SetLabel(n);
	return Status;
}

void GMenuItem::Enabled(bool e)
{
	SetEnabled(e);
}

void GMenuItem::Focus(bool f)
{
	Highlight(f);
}

///////////////////////////////////////////////////////////////////////////////////////////////
GMenu::GMenu() : BMenuBar(BRect(0, 0, 1, 1), "MenuBar")
{
	Window = NULL;
}

GMenu::~GMenu()
{
	Detach();
}

bool GMenu::Attach(GWindow *p)
{
	bool Status = FALSE;
	Window = p;
	if (Window)
	{
		Window->Sys_View->AddChild(this);
		Status = TRUE;
	}
	return Status;
}

bool GMenu::Detach()
{
	bool Status = FALSE;

	if (Window)
	{
		Window->Sys_View->RemoveChild(this);
		Window = 0;
	}

	return Status;
}

bool GMenu::Load(GWindow *p, char *Res)
{
	bool Status = FALSE;

	if (p AND Res)
	{
		Window = p;
	}

	return Status;
}

GMenuItem *GMenu::AppendItem(char *Str, int Id, bool Enabled)
{
	GMenuItem *Item = NEW(GMenuItem());
	if (Item)
	{
		Item->SetTarget(Window->Sys_View);
		AddItem(Item);

		char *s = NewStrLessAnd(Str);
		if (s)
		{
			Item->Name(Str);
			DeleteArray(s);
		}
		
		Item->Id(Id);
		Item->Enabled(Enabled);
	}
	return Item;
}

GMenuItem *GMenu::AppendSeparator()
{
	AddSeparatorItem();
	return NULL;
}

GSubMenu *GMenu::AppendSub(char *Str)
{
	GSubMenu *Sub = NEW(GSubMenu(Str));
	if (Sub)
	{
		Sub->Window = Window;
		AddItem(Sub);
	}

	return Sub;
}


///////////////////////////////////////////////////////////////////////////////////////////
GCommand::GCommand()
{
	Flags = GWF_ENABLE | GWF_VISIBLE;
	Id = 0;
	ToolButton = 0;
	MenuItem = 0;
	Accelerator = 0;
	TipHelp = 0;
}

GCommand::~GCommand()
{
	DeleteArray(Accelerator);
	DeleteArray(TipHelp);
}

bool GCommand::Enabled()
{
	return TestFlag(Flags, GWF_ENABLE);
}

void GCommand::Enabled(bool e)
{
	if (Enabled() != e)
	{
		if (e)
		{
			SetFlag(Flags, GWF_ENABLE);
		}
		else
		{
			ClearFlag(Flags, GWF_ENABLE);
		}

		if (ToolButton)
		{
			ToolButton->Enabled(e);
		}
		if (MenuItem)
		{
			MenuItem->Enabled(e);
		}
	}
}

bool GCommand::Value()
{
	bool v = FALSE;
	if (ToolButton)
	{
		v |= ToolButton->Value();
	}
	if (MenuItem)
	{
		v |= MenuItem->Checked();
	}
	return v;
}

void GCommand::Value(bool v)
{
	if (ToolButton)
	{
		ToolButton->Value(v);
	}
	if (MenuItem)
	{
		MenuItem->Checked(v);
	}
}
