/*
**	FILE:		GuiViews.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Standard Views
**
**	Copyright (C) 1998, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

//////////////////////////////////////////////////////////////////////////////
GView::GView()
	: BView(BRect(0, 0, 1, 1), NULL, B_FOLLOW_NONE, B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE),
	GWindow(this)
{
	ExtraFlags = 0;
	ScrollView = 0;
	VScroll = 0;
	HScroll = 0;

	SetViewColor(B_TRANSPARENT_32_BIT);	
}

GView::~GView()
{
	DeleteObj(ScrollView);
	DeleteObj(VScroll);
	DeleteObj(HScroll);
}

void GView::GetScrollPos(int &x, int &y)
{
	if (HScroll)
	{
		x = HScroll->Pos();
	}
	else
	{
		x = 0;
	}

	if (VScroll)
	{
		y = VScroll->Pos();
	}
	else
	{
		y = 0;
	}
}

void GView::SetScrollPos(int x, int y)
{
	if (HScroll)
	{
		HScroll->SetPos(x);
	}

	if (VScroll)
	{
		VScroll->SetPos(y);
	}
}

bool GView::SetScrollBars(bool x, bool y)
{
	bool Status = FALSE;

	DeleteObj(ScrollView);
	ScrollView = NEW(BScrollView("ScrollView", this, ResizingMode(), Flags(), x, y, B_NO_BORDER));

	DeleteObj(VScroll);
	DeleteObj(HScroll);
	if (ScrollView)
	{
		if (x)
		{
			HScroll = NEW(GScrollBar(ScrollView->ScrollBar(B_HORIZONTAL)));
		}
		
		if (y)
		{
			VScroll = NEW(GScrollBar(ScrollView->ScrollBar(B_VERTICAL)));
		}
	}

	return Status;
}

int GView::OnEvent(BMessage *Msg)
{
	if (VScroll) VScroll->OnEvent(Msg);
	if (HScroll) HScroll->OnEvent(Msg);
	return GWindow::OnEvent(Msg);
}

#define ADJ_LEFT			1
#define ADJ_RIGHT			2
#define ADJ_UP				3
#define ADJ_DOWN			4

extern int IsAdjacent(GdcRegion &a, GdcRegion &b);
int IsAdjacent(GdcRegion &a, GdcRegion &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

extern GdcRegion JoinAdjacent(GdcRegion &a, GdcRegion &b, int Adj);
GdcRegion JoinAdjacent(GdcRegion &a, GdcRegion &b, int Adj)
{
	GdcRegion t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

GdcRegion *GView::FindLargest(GdcComplexRegion &r)
{
	int Pixels = 0;
	GdcRegion *Best = 0;
	static GdcRegion Final;

	// do initial run through the list to find largest single region
	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		// search for adjoining rectangles that maybe we can add
		for (GdcRegion *i = r.First(); i; i = r.Next())
		{
			if (i != Best)
			{
				int Adj = IsAdjacent(*Best, *i);
				if (Adj)
				{
					GdcRegion t = JoinAdjacent(*Best, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > Pixels)
					{
						Pixels = Pix;
						Final = t;
					}
				}
			}
		}
	}

	return &Final;
}

GdcRegion *GView::FindSmallestFit(GdcComplexRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GdcRegion *Best = 0;

	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		if (	(i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GdcRegion *GView::FindLargestEdge(GdcComplexRegion &r, int Edge)
{
	GdcRegion *Best = 0;

	for (GdcRegion *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if (	(Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (	((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (	((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

//////////////////////////////////////////////////////////////////////////////////////
GScrollBar::GScrollBar(BScrollBar *bar)
{
	Bar = bar;
}

GScrollBar::~GScrollBar()
{
	Bar = 0;
}

bool GScrollBar::Vertical()
{
	return Bar->Orientation() == B_VERTICAL;
}

int GScrollBar::Pos()
{
	return Bar->Value();
}

void GScrollBar::SetPos(int p)
{
	Bar->SetValue(p);
}

void GScrollBar::Limits(int &Low, int &High)
{
	float l, h;
	Bar->GetRange(&l, &h);
	Low = l;
	High = h;
}

void GScrollBar::SetLimits(int Low, int High)
{
	Bar->SetRange(Low, High);
}

int GScrollBar::Page()
{
	float small, large;
	Bar->GetSteps(&small, &large);
	return large;
}

void GScrollBar::SetPage(int p)
{
	Bar->SetSteps(1.0, p);
}

int GScrollBar::OnEvent(BMessage *Msg)
{
	/*
	if (	(Msg == WM_HSCROLL AND Type == SB_HORZ) OR
		(Msg == WM_VSCROLL AND Type == SB_VERT))
	{
		int OldPos = Info.nPos;
		int NewPos = Info.nPos;
		int ScrollCode = (int) LOWORD(a);
		int nPos = (short) HIWORD(a);

		switch (ScrollCode)
		{
			case SB_BOTTOM:
			case SB_LINEDOWN:
			{
				NewPos = limit(NewPos+1, Info.nMin, Info.nMax);
				break;
			}
			case SB_TOP:
			case SB_LINEUP:
			{
				NewPos = limit(NewPos-1, Info.nMin, Info.nMax);
				break;
			}
			case SB_PAGEDOWN:
			{
				NewPos = limit(NewPos+(int)Info.nPage, Info.nMin, Info.nMax);
				break;
			}
			case SB_PAGEUP:
			{
				NewPos = limit(NewPos-(int)Info.nPage, Info.nMin, Info.nMax);
				break;
			}
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK:
			{
				NewPos = limit(nPos, Info.nMin, Info.nMax);
				break;
			}
		}

		if (OldPos != NewPos)
		{
			Info.nPos = NewPos;
			Update();
			OnChange(nPos);

			if (Parent)
			{
				SendMessage(Parent->Handle(), WM_CHANGE, nPos, (LPARAM) this);
			}
		}
	}
*/

	return FALSE;
}

//////////////////////////////////////////////////////////////////////////////////////
GStatusBar::GStatusBar()
{
	GWindow::Name("LGI_StatusBar");

	Font = NEW(GdcBeTtf);
	if (Font)
	{
		if (NOT Font->CreateFont(	-8,
						0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0,
						"Ms Sans Serif"))
		{
			DeleteObj(Font);
		}
		else
		{
			Font->Transparent(TRUE);
		}
	}
}

GStatusBar::~GStatusBar()
{
	for (int i=0; i<GetSize(); i++)
	{
		DeleteObj((*this)[0]);
	}

	DeleteObj(Font);
}

bool GStatusBar::Pour(GdcComplexRegion &r)
{
	GdcRegion *Best = FindLargestEdge(r, GV_EDGE_BOTTOM);
	if (Best)
	{
		GdcRegion Take = *Best;
		if (Take.Y() > 21)
		{
			Take.y1 = max((Take.y2 - 21), Take.y1);
			SetPos(Take);
			RePour();
			return TRUE;
		}
	}
	return FALSE;
}

void GStatusBar::OnPaint(GdcBasePrimitives *pDC)
{
	int Xm1 = pDC->X()-1;
	int Ym1 = pDC->Y()-1;

	pDC->Colour(CBit(pDC->GetBits(), TOOL_LOW));
	pDC->Line(Xm1, Ym1, 0, Ym1);
	pDC->Line(Xm1, Ym1, Xm1, 0);

	pDC->Colour(CBit(pDC->GetBits(), TOOL_HIGH));
	pDC->Line(0, 0, Xm1, 0);
	pDC->Line(0, 0, 0, Ym1);

	pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
	pDC->Rectangle(1, 1, Xm1-1, Ym1-1);

	for (	GStatusPane *Pane = (GStatusPane*) Children.First();
		Pane;
		Pane = (GStatusPane*) Children.Next())
	{
		pDC->SetOrigin(-Pane->Pos.x1, -Pane->Pos.y1);
		Pane->OnPaint(pDC);
	}
}

void GStatusBar::RePour()
{
	int X = Pos.X() - 5;
	int i = Children.GetItems()-1;

	for (	GStatusPane *Pane = (GStatusPane*) Children.Last();
		Pane;
		Pane = (GStatusPane*) Children.Prev(), i--)
	{
		if (i)
		{
			// not first
			X -= Pane->Width; 
			Pane->Pos.ZOff(Pane->Width, Pos.Y()-7);
			Pane->Pos.Offset(X, 3);
			X -= STATUSBAR_SEPARATOR;
		}
		else
		{
			// first
			Pane->Pos.ZOff(X-2, Pos.Y()-4);
			Pane->Pos.Offset(2, 2);
			X = 0;
		}
	}
}

GStatusPane *GStatusBar::AppendPane(char *Text, int Width)
{
	GStatusPane *Pane = 0;
	if (Text)
	{
		Pane = NEW(GStatusPane);
		if (Pane)
		{
			Pane->GWindow::Parent = this;
			Pane->GWindow::Name(Text);
			Pane->Width = Width;
			Children.Insert(Pane);
		}
	}
	return Pane;
}

bool GStatusBar::AppendPane(GStatusPane *Pane)
{
	if (Pane)
	{
		Pane->GWindow::Parent = this;
		Children.Insert(Pane);
	}
	return TRUE;
}

// Pane
GStatusPane::GStatusPane()
	: BView(BRect(0, 0, 1, 1), NULL, B_FOLLOW_NONE, B_WILL_DRAW | B_FRAME_EVENTS | B_FULL_UPDATE_ON_RESIZE),
	GWindow(this)
{
	GWindow::Parent = 0;
	Flags = 0;
	Width = 32;
	Pos.ZOff(Width-1, 20);
	pDC = 0;
}

GStatusPane::~GStatusPane()
{
	DeleteObj(pDC)
}

bool GStatusPane::Name(char *n)
{
	bool Status = GObject::Name(n);
	GdcRegion p(0, 0, Pos.X()-1, Pos.Y()-1);
	p.Size(1, 1);
	GWindow::Invalidate(&p);
	return Status;
}

void GStatusPane::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
	int Off = 1;

	if (Sunken())
	{
		Off = 0;
		DrawMoveShallowEdge(pDC, r, SUNKEN);
	}

	GStatusBar *Par = (GStatusBar*) GWindow::Parent;
	if (Par AND Par->Font)
	{
		Par->Font->Text(pDC, 2+Off, 1+Off, Name(), -1, &r);
	}
}

int GStatusPane::X()
{
	return Width;
}

void GStatusPane::X(int x)
{
	Width = x;
	if (GWindow::Parent)
	{
		((GStatusBar*) GWindow::Parent)->RePour();
	}
}

bool GStatusPane::Sunken()
{
	return TestFlag(Flags, GSP_SUNKEN);
}

void GStatusPane::Sunken(bool s)
{
	bool Old = Sunken();
	if (s) SetFlag(Flags, GSP_SUNKEN);
	else ClearFlag(Flags, GSP_SUNKEN);

	if (Old != Sunken() AND GWindow::Parent)
	{
		GWindow::Parent->GWindow::Invalidate(&Pos);
	}
}

GdcBasePrimitives *GStatusPane::Bitmap()
{
	return pDC;
}

void GStatusPane::Bitmap(GdcBasePrimitives *pNewDC)
{
	DeleteObj(pDC);
	if (pNewDC)
	{
		pDC = NEW(GdcMemDC);
		if (pDC)
		{
			if (pDC->Create(pNewDC->X(), pNewDC->Y(), pNewDC->GetBits()))
			{
				pDC->Blt(0, 0, pNewDC);
			}
			else
			{
				DeleteObj(pDC);
			}
		}
	}

	if (GWindow::Parent)
	{
		GWindow::Parent->Invalidate(&Pos);
	}
}

//////////////////////////////////////////////////////////////////////////////
#define SPLITER_BORDER			4
#define SPLITER_BAR			5
#define SPLITER_INSET			2
#define SPLITER_MIN_X			(2*SPLITER_INSET)
#define SPLITER_MIN_Y			(2*SPLITER_INSET)

#define SPLITER_FULL_DRAG

GSplitter::GSplitter()
{
	GWindow::Name("LGI_Spliter");
	Vertical = TRUE;
	SplitPos = -1;
	Moving = FALSE;

	PosA.ZOff(0, 0);
	PosB.ZOff(0, 0);
	ViewA = 0;
	ViewB = 0;
	Split(0);
}

GSplitter::~GSplitter()
{
	DeleteObj(ViewA);
	DeleteObj(ViewB);
}

void GSplitter::SetViewA(GWindow *a)
{
	ViewA = a;
	if (a)
	{
		a->Attach(this);
	}
}

void GSplitter::SetViewB(GWindow *b)
{
	ViewB = b;
	if (b)
	{
		b->Attach(this);
	}
}

void GSplitter::Split(int s)
{
	int NewPos = max(s, 4); // upper limit: Pos.Y()-18
	if (NewPos != SplitPos)
	{
		SplitPos = NewPos;
		CalcRegions(Pos);
		if (Visible())
		{
			GWindow::Invalidate();
		}
	}
}

void GSplitter::CalcRegions(GdcRegion Rect)
{
	PosA = Rect;
	PosA.Offset(-PosA.x1, -PosA.y1);
	PosA.Size(SPLITER_BORDER, SPLITER_BORDER);
	PosB = PosA;

	if (Vertical)
	{
		PosA.x2 = PosA.x1 + SplitPos;
		PosB.x1 = PosA.x2 + SPLITER_BAR;
	}
	else
	{
		PosA.y2 = PosA.y1 + SplitPos;
		PosB.y1 = PosA.y2 + SPLITER_BAR;
	}

	if (ViewA)
	{
		GdcRegion r = PosA;
		r.Size(SPLITER_INSET, SPLITER_INSET);
		ViewA->SetPos(r);
	}
	if (ViewB)
	{
		GdcRegion r = PosB;
		r.Size(SPLITER_INSET, SPLITER_INSET);
		ViewB->SetPos(r);
	}
}

bool GSplitter::Pour(GdcComplexRegion &r)
{
	GdcRegion *Best = FindLargest(r);
	if (Best)
	{
		CalcRegions(*Best);
		SetPos(*Best);
		return TRUE;
	}
	return FALSE;
}

void GSplitter::OnPaint(GdcBasePrimitives *pDC)
{
	GdcRegion r;
	r.ZOff(pDC->X()-1, pDC->Y()-1);

	DrawMoveShallowEdge(pDC, r, RAISED);
	pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
	DrawMoveFill(pDC, r, SPLITER_BORDER);

	// bar
	if (Vertical)
	{
		pDC->Rectangle(PosA.x2, PosA.y1, PosB.x1, PosA.y2);
	}
	else
	{
		pDC->Rectangle(PosA.x1, PosA.y2, PosA.x2, PosB.y1);
	}

	// PosA region
	r = PosA;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		DrawMoveDeepEdge(pDC, r, SUNKEN);

		if (ViewA)
		{
		}
		else
		{
			pDC->Colour(0xFFFFFF);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}

	r = PosB;
	if (r.X() > SPLITER_MIN_X AND r.Y() > SPLITER_MIN_Y)
	{
		DrawMoveDeepEdge(pDC, r, SUNKEN);

		if (ViewB)
		{
		}
		else
		{
			pDC->Colour(0xFFFFFF);
			pDC->Rectangle(&r);
		}
	}
	else
	{
		// too small
	}
}

extern void HashRectangle(GWindow *Wnd, GdcRegion *r);
void HashRectangle(GWindow *Wnd, GdcRegion *r)
{
	/*
	if (r AND Wnd AND Wnd->Handle())
	{
		HDC hdc = GetDC(Wnd->Handle());
		if (hdc)
		{
			BITMAPINFO *bmp = (BITMAPINFO*) NEW(char[sizeof(BITMAPINFO)+sizeof(RGBQUAD)+32]);
			int Pat[8] = {0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555, 0xAAAAAAAA, 0x55555555};
			if (bmp)
			{
				bmp->bmiHeader.biSize = sizeof(bmp->bmiHeader);
				bmp->bmiHeader.biWidth = 8;
				bmp->bmiHeader.biHeight = 8;
				bmp->bmiHeader.biPlanes = 1;
				bmp->bmiHeader.biBitCount = 1;
				bmp->bmiHeader.biCompression = BI_RGB;
				bmp->bmiHeader.biSizeImage = 32;
				bmp->bmiHeader.biXPelsPerMeter = 3000;
				bmp->bmiHeader.biYPelsPerMeter = 3000;
				bmp->bmiHeader.biClrUsed = 0;
				bmp->bmiHeader.biClrImportant = 0;

				bmp->bmiColors[0].rgbRed = 255;
				bmp->bmiColors[0].rgbGreen = 255;
				bmp->bmiColors[0].rgbBlue = 255;
				bmp->bmiColors[1].rgbRed = 0;
				bmp->bmiColors[1].rgbGreen = 0;
				bmp->bmiColors[1].rgbBlue = 0;

				memcpy(bmp->bmiColors + 2, Pat, 32);
				
				LOGBRUSH brush;
				brush.lbStyle = BS_DIBPATTERNPT;
				brush.lbColor = DIB_RGB_COLORS;
				brush.lbHatch = (LONG) bmp;

				HBRUSH hbrush = CreateBrushIndirect(&brush);
				if (hbrush)
				{
					HPEN hpen = GetStockObject(NULL_PEN);
					hpen = SelectObject(hdc, hpen);
					hbrush = SelectObject(hdc, hbrush);

					SetTextColor(hdc, RGB(255, 255, 255));
					SetBkColor(hdc, RGB(0, 0, 0));
					SetROP2(hdc, R2_XORPEN);
					SetBkMode(hdc, TRANSPARENT);
					::Rectangle(hdc, r->x1, r->y1, r->x2+1, r->y2+1);

					hbrush = SelectObject(hdc, hbrush);
					hpen = SelectObject(hdc, hpen);
					DeleteObject(hbrush);
				}

				ReleaseDC(Wnd->Handle(), hdc);
				DeleteArray(bmp);
			}
		}
	}
*/
}

void GSplitter::OnMouseClick(VMouse &m)
{

	if (m.Down)
	{
		if (Vertical)
		{
			if (m.x > PosA.x2 AND m.x < PosB.x1 AND m.y > PosA.y1 AND m.y < PosA.y2)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x2;
				Bar.x2 = PosB.x1;
				Bar.y1 = PosA.y1;
				Bar.y2 = PosA.y2;
				Offset = m.x - Bar.x1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
		else
		{
			if (m.x > PosA.x1 AND m.x < PosA.x2 AND m.y > PosA.y2 AND m.y < PosB.y1)
			{
				Capture(TRUE);
				Moving = TRUE;

				Bar.x1 = PosA.x1;
				Bar.x2 = PosA.x2;
				Bar.y1 = PosA.y2;
				Bar.y2 = PosB.y1;
				Offset = m.y - Bar.y1;

				#ifndef SPLITER_FULL_DRAG
				HashRectangle(this, &Bar);
				#endif
			}
		}
	}
	else
	{
		#ifdef SPLITER_FULL_DRAG
		Capture(FALSE);
		Moving = FALSE;
		#else
		if (Moving)
		{
			HashRectangle(this, &Bar);
			Capture(FALSE);
			Moving = FALSE;

			int NewPos;
			if (Vertical)
			{
				NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
			}
			else
			{
				NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
			}

			if (NewPos != SplitPos)
			{
				SplitPos = NewPos;
				CalcRegions(Pos);
				GWindow::Invalidate();
			}
		}
		#endif
	}
}

void GSplitter::OnMouseMove(VMouse &m)
{
	if (Moving)
	{
		#ifdef SPLITER_FULL_DRAG
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}

		int NewPos;
		if (Vertical)
		{
			NewPos = limit(Bar.x1-4, 4, Pos.X()-18);
		}
		else
		{
			NewPos = limit(Bar.y1-4, 4, Pos.Y()-18);
		}

		if (NewPos != SplitPos)
		{
			SplitPos = NewPos;
			CalcRegions(Pos);
			GWindow::Invalidate();
		}
		#else
		HashRectangle(this, &Bar);
		if (Vertical)
		{
			Bar.Offset((m.x - Offset) - Bar.x1, 0);
		}
		else
		{
			Bar.Offset(0, (m.y - Offset) - Bar.y1);
		}
		HashRectangle(this, &Bar);
		#endif
	}
}
