/*hdr
**      FILE:           Contain.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           27/7/95
**      DESCRIPTION:    Continer classes
**
**      Copyright (C) 1996, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>

#include "defs.h"
#include "memdev.h"
#include "contain.h"

PtrArray::PtrArray()
{
	Used = 0;
	Size = 0;
	BankSize = DEFAULT_ARRAY_BANK;
	Array = 0;
}

PtrArray::~PtrArray()
{
	if (Array) delete [] Array;
}

bool PtrArray::SetSize(int NewSize)
{
	bool Status = FALSE;

	if (NewSize > Size OR abs(NewSize-Size) > BankSize)
	{
		int BankMod = NewSize % BankSize;
		
		if (BankMod)
		{
			NewSize += BankSize - BankMod;
		}
		
		void **a = NEW(void*[NewSize]);
		if (a)
		{
			int Common = min(Used, NewSize);
			memcpy(a, Array, Common);
			if (Common < NewSize) memset(a+Common, 0, sizeof(*a)*(NewSize-Common));

			DeleteArray(Array);
			Array = a;			

			Used = Common;			
			Size = NewSize;
			Status = TRUE;
		}
	}
	else
	{
		Used = max(NewSize, 0);
		Status = TRUE;
	}

	return Status;
}

bool PtrArray::Add(void *Ptr, int Where)
{
	if (Used >= Size - 1)
	{
		if (NOT SetSize(Size + BankSize))
		{
			return FALSE;
		}
	}

	Array[Used] = Ptr;
	Used++;

	return TRUE;
}

bool PtrArray::Remove(int Where)
{
	bool Status = FALSE;

	if (Array AND Where >= 0 AND Where < Used)
	{
		memmove(&Array[Where], &Array[Where+1], Used-Where);
		Used--;
		Status = TRUE;
	}
	
	return Status;
}

bool PtrArray::Empty()
{
	Used = 0;
	return TRUE;
}

VoidPtr &PtrArray::operator [](int i)
{
	void *Ptr = 0;
	return (Array AND i >= 0 AND i < Used) ? Array[i] : Ptr;
}

template <class StackType>
Stack<StackType>::Stack()
{
	Top = 0;
}

template <class StackType>
Stack<StackType>::~Stack()
{
	Empty();
}

template <class StackType>
void Stack<StackType>::Push(StackType *Data)
{
	StackType *Temp = NEW(StackNode);

	if (Temp)
	{
		Temp->Next = Top;
		Top = Temp;
		Top->Data = Data;
	}
}

template <class StackType>
StackType *Stack<StackType>::Pop()
{
	if (Top)
	{
		StackType *Data = Top->Data;
		StackNode *Temp = Top;
		Top = Top->Next;

		delete Temp;
		return Data;
	}

	return 0;
}

template <class StackType>
StackType *Stack<StackType>::Peek()
{
	return (Top) ? Top->Data : 0;
}

template <class StackType>
int Stack<StackType>::IsEmpty()
{
	return (Top) ? TRUE : FALSE;
}

template <class StackType>
void Stack<StackType>::Empty()
{
	while (!IsEmpty())
	{
		Pop();
	}
}

template <class QueType>
Que<QueType>::Que()
{
	First = Last = 0;
}

template <class QueType>
Que<QueType>::~Que()
{
	Empty();
}

template <class QueType>
void Que<QueType>::Insert(QueType *Data)
{
	if (Last)
	{
		Last->Next = NEW(QueNode);
		Last = Last->Next;
	}
	else
	{
		First = Last = NEW(QueNode);
	}

	Last->Next = 0;
	Last->Data = Data;
}

template <class QueType>
QueType *Que<QueType>::Delete()
{
	if (First)
	{
		QueNode *Temp = First;
		QueData *Data = First->Data;
		First = First->Next;
		if (!First) Last = 0;		

		delete Temp;
		return Data;
	}

	return 0;
}

template <class QueType>
QueType *Que<QueType>::Peek()
{
	return (First) ? First->Data : 0;
}

template <class QueType>
int Que<QueType>::IsEmpty()
{
	return (First) ? FALSE : TRUE;
}

template <class QueType>
void Que<QueType>::Empty()
{
	while (!IsEmpty())
	{
		Delete();
	}
}

template <class PQueType>
PriorityQue<PQueType>::PriorityQue()
{
	Nodes = 0;
	Top = 0;
	Size = 0;
	SetSize(DEFAULT_PQUE_SIZE);
}

template <class PQueType>
PriorityQue<PQueType>::~PriorityQue()
{
	Empty();
}

template <class PQueType>
void PriorityQue<PQueType>::SetSize(long NewSize)
{
	if (NewSize > 0)
	{
		PQueNode *Temp = NEW(PQueNode[NewSize]);
		
		if (Temp)
		{
			memset(Temp,0,sizeof(PQueNode)*NewSize);
			
			if (Top)
			{
				memcpy(Temp,Top,sizeof(PQueNode)*Nodes);
				delete [] Top;
			}
			
			Top = Temp;
			Size = NewSize;
		}
	}
	else
	{
		Empty();
	}
}

template <class PQueType>
void PriorityQue<PQueType>::BubbleUp(long Start)
{
	long Parent = (Start - 1) / 2;
	
	while (Start)
	{
		if (Top[Parent].Key > Top[Start].Key)
		{
			PQueNode Temp = Top[Parent];
			Top[Parent] = Top[Start];
			Top[Start] = Temp;			

			Start = Parent;
			Parent = (Start - 1) / 2;
		}
		else
		{
			Start = 0;
		}
	}
}

template <class PQueType>
void PriorityQue<PQueType>::BubbleDown()
{
	long Start = 0;
	long Child = 1;
	
	while (Child < Nodes)
	{
		long Swap = (Top[Child].Key < Top[Child+1].Key)
				? Child : Child + 1;
		Swap = min(Swap, Nodes-1);

		if (Top[Swap].Key < Top[Start].Key)
		{
			PQueNode Temp = Top[Swap];
			Top[Swap] = Top[Start];
			Top[Start] = Temp;			
			Start = Swap;
			Child = (Start * 2) + 1;
		}
		else
		{
			Child = Nodes;
		}
	}
}

template <class PQueType>
void PriorityQue<PQueType>::Insert(PQueType *Data,long Key)
{
	if (!Top || Nodes >= Size)
	{
		SetSize((Size) ? Size * 2 : DEFAULT_PQUE_SIZE);
	}

	if (Top && Nodes < Size)
	{
		Top[Nodes].Data = Data;
		Top[Nodes].Key = Key;
		Nodes++;

   		BubbleUp(Nodes-1);
	}
}

template <class PQueType>
PQueType *PriorityQue<PQueType>::DeleteMin()
{
	if (Top && Nodes > 0)
	{
		PQueType *Temp = Top->Data;
		
		Nodes--;

		if (Nodes >= 1)
		{
			Top[0] = Top[Nodes];			
			BubbleDown();
		}
		
		return Temp;
	}
	
	return 0;
}

template <class PQueType>
PQueType *PriorityQue<PQueType>::Peek()
{
	return (Top) ? Top->Data : 0;
}

template <class PQueType>
int PriorityQue<PQueType>::IsEmpty()
{
	return (Top) ? FALSE : TRUE;
}

template <class PQueType>
void PriorityQue<PQueType>::Empty()
{
	delete [] Top;
	Top = 0;
	Size = 0;
	Nodes = 0;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////

DLinkList::DLinkList()
{
	Items = 0;
	Cur = FirstObj = LastObj = 0;
}

DLinkList::~DLinkList()
{
	Empty();
}

void DLinkList::Empty()
{
	Item *i = FirstObj;
	while (i)
	{
		Item *Ni = i->Next;
		delete i;
		i = Ni;
	}

	Cur = FirstObj = LastObj = 0;
	Items = 0;
}

void *DLinkList::operator [](int Index)
{
	if (Index >= 0 AND Index < Items)
	{
		Item *i = FirstObj;
		while (i AND Index)
		{
			i = i->Next;
			Index--;
		}

		if (Index == 0 AND i)
		{
			Cur = i;
			return i->Ptr;
		}
	}

	return 0;
}

int DLinkList::FindIndex(void *p)
{
	int i=0;
	Item *item = FirstObj;
	while (item)
	{
		if (item->Ptr == p)
		{
			return i;
		}
		
		i++;
		item = item->Next;

	}

	return -1;
}

bool DLinkList::Delete()
{
	if (Cur)
	{
		Item *i = Cur->Next;

		if (Cur->Prev)
		{
			Cur->Prev->Next = Cur->Next;
		}
		else
		{
			FirstObj = Cur->Next;
		}

		if (Cur->Next)
		{
			Cur->Next->Prev = Cur->Prev;
		}
		else
		{
			LastObj = Cur->Prev;
		}

		DeleteObj(Cur);
		Cur = i;
		Items--;

		return TRUE;
	}

	return FALSE;
}

bool DLinkList::Delete(int Index)
{
	Item *i = FirstObj;
	while (i AND Index)
	{
		i = i->Next;
		Index--;
	}

	if (Index == 0 AND i)
	{
		void *Ptr = (void*) i->Ptr;

		if (FirstObj == i)
		{
			FirstObj = i->Next;
		}

		if (LastObj == i)
		{
			LastObj = i->Prev;
		}

		if (i->Next) i->Next->Prev = i->Prev;
		if (i->Prev) i->Prev->Next = i->Next;

		if (i == Cur)
		{
			Cur = 0;
			Index = -1;
		}

		DeleteObj(i);
		Items--;

		return TRUE;
	}

	return FALSE;
}

bool DLinkList::Delete(void *p)
{
	Item *i = FirstObj;
	while (i AND i->Ptr != p)
	{
		i = i->Next;
	}

	if (i AND i->Ptr == p)
	{
		void *Ptr = (void*) i->Ptr;

		if (FirstObj == i)
		{
			FirstObj = i->Next;
		}

		if (LastObj == i)
		{
			LastObj = i->Prev;
		}

		if (i->Next) i->Next->Prev = i->Prev;
		if (i->Prev) i->Prev->Next = i->Next;

		if (i == Cur)
		{
			Cur = 0;
		}

		DeleteObj(i);
		Items--;

		return TRUE;
	}

	return FALSE;
}

bool DLinkList::Insert(void *p, int Index)
{
	Item *n = NEW(Item);
	if (n)
	{
		n->Ptr = p;
		if (NOT FirstObj)
		{
			FirstObj = LastObj = n;
			n->Prev = n->Next = 0;
			Items = 1;

			return TRUE;
		}
		else
		{
			if (Index < 0)
			{
				Item *i = LastObj;
				LastObj = i->Next = n;
				n->Prev = i;
				n->Next = 0;
				Items++;

				return TRUE;
			}
			else
			{
				Item *i = FirstObj;
				if (Index > 0)
				{
					while (Index AND i->Next)
					{
						i = i->Next;
						Index--;
					}
				}

				if (i == FirstObj)
				{
					FirstObj = n;
				}

				n->Prev = i->Prev;
				n->Next = i;
				if (n->Prev) n->Prev->Next = n;
				if (n->Next) n->Next->Prev = n;
				Items++;

				return TRUE;
			}
		}
	}

	return FALSE;
}

void *DLinkList::First()
{
	Cur = FirstObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkList::Last()
{
	Cur = LastObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkList::Next()
{
	if (Cur)
	{
		Cur = Cur->Next;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkList::Prev()
{
	if (Cur)
	{
		Cur = Cur->Prev;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkList::Find(void *p)
{
	Cur = FirstObj;
	while (Cur AND Cur->Ptr != p)
	{
		Cur = Cur->Next;
	}
	return (Cur) ? Cur->Ptr : 0;
}

////////////////////////////////////////////////////////////////////////////////////////////
DLinkIterator::DLinkIterator(DLinkList *list)
{
	List = list;
	Cur = 0;
}

DLinkIterator::~DLinkIterator()
{
	List = 0;
	Cur = 0;
}

void *DLinkIterator::First()
{
	Cur = List->FirstObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkIterator::Last()
{
	Cur = List->LastObj;
	return (Cur) ? Cur->Ptr : 0;
}

void *DLinkIterator::Next()
{
	if (Cur)
	{
		Cur = Cur->Next;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkIterator::Prev()
{
	if (Cur)
	{
		Cur = Cur->Prev;
		return (Cur) ? Cur->Ptr : 0;
	}
	return 0;
}

void *DLinkIterator::Find(void *p)
{
	Cur = List->FirstObj;
	while (Cur AND Cur->Ptr != p)
	{
		Cur = Cur->Next;
	}
	return (Cur) ? Cur->Ptr : 0;
}

////////////////////////////////////////////////////////////////////////////////////////
BytePipe::BytePipe()
{
}

BytePipe::~BytePipe()
{
	Empty();
}

void BytePipe::Empty()
{
	for (Block *b = Mem.First(); b; b = Mem.Next())
	{
		DeleteObj(b);
	}
}

int BytePipe::Sizeof()
{
	int Size = 0;
	for (Block *b = Mem.First(); b; b = Mem.Next())
	{
		if (b->Ptr)
		{
			Size += b->Size - b->Next;
		}
	}
	return Size;
}

bool BytePipe::Pop(uchar *Ptr, int Size)
{
	bool Status = FALSE;

	if (Ptr AND Size <= Sizeof())
	{
		for (Block *b = Mem.First(); b AND Size > 0; b = Mem.Next())
		{
			int Copy = min(Size, b->Size - b->Next);
			if (Copy > 0)
			{
				memcpy(Ptr, b->Ptr + b->Next, Copy);
				Ptr += Copy;
				Size -= Copy;
				b->Next += Copy;
			}
		}

		for (b = Mem.First(); b AND b->Next >= b->Size; b = Mem.First())
		{
			Mem.Delete(b);
			DeleteObj(b);
		}

		Status = (Size == 0);
	}

	return Status;
}

bool BytePipe::Push(uchar *Ptr, int Size)
{
	bool Status = FALSE;

	if (Ptr AND Size > 0)
	{
		Block *b = NEW(Block);
		if (b)
		{
			b->Ptr = NEW(uchar[Size]);
			if (b->Ptr)
			{
				memcpy(b->Ptr, Ptr, Size);
				b->Size = Size;
				b->Next = 0;
				Mem.Insert(b);
				Status = TRUE;
			}
			else
			{
				DeleteObj(b);
			}
		}
	}

	return Status;
}

