/*
**	FILE:			DragAndDrop.h
**	AUTHOR:			Matthew Allen
**	DATE:			27/3/99
**	DESCRIPTION:	Drag and drop classes
**
**	Copyright (C) 1999, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#ifndef __DRAG_AND_DROP
#define __DRAG_AND_DROP

#ifdef WIN32
#include <shlobj.h>
#define CF_GDragDropSource					"LGI::GDragDropSource"
#else if BEOS
#define S_OK								0
#endif

//////////////////////////////////////////////////////////////////////////////
#ifdef BEOS
#define DROPEFFECT_NONE						0
#define DROPEFFECT_COPY						1
#define DROPEFFECT_MOVE						2
#define DROPEFFECT_LINK						3
#endif

#ifdef WIN32
class GDragDropSource : public IDataObject, public IDropSource, public IEnumFORMATETC {
#else if BEOS
class GDragDropSource {
#endif
private:
	#ifdef WIN32
	int Index;
	UINT DdFormat;

	// IUnknown
	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ppv);
	
	// IEnumFORMATETC
	HRESULT STDMETHODCALLTYPE Next(ULONG celt, FORMATETC *rgelt, ULONG *pceltFetched);
	HRESULT STDMETHODCALLTYPE Skip(ULONG celt);
	HRESULT STDMETHODCALLTYPE Reset();
	HRESULT STDMETHODCALLTYPE Clone(IEnumFORMATETC **ppenum);
	HRESULT STDMETHODCALLTYPE GetCanonicalFormatEtc(FORMATETC * pFormatetcIn, FORMATETC * pFormatetcOut);
	HRESULT STDMETHODCALLTYPE SetData(FORMATETC *pFormatetc, STGMEDIUM *pmedium, BOOL fRelease);
	HRESULT STDMETHODCALLTYPE DAdvise(FORMATETC *pFormatetc, DWORD advf, IAdviseSink *pAdvSink, DWORD *pdwConnection);
	HRESULT STDMETHODCALLTYPE DUnadvise(DWORD dwConnection);
	HRESULT STDMETHODCALLTYPE EnumDAdvise(IEnumSTATDATA **ppenumAdvise);

	// IDropSource
	HRESULT STDMETHODCALLTYPE QueryContinueDrag(BOOL fEscapePressed, DWORD InputState);
	HRESULT STDMETHODCALLTYPE GiveFeedback(DWORD dwEffect);

	// IDataObject
	HRESULT STDMETHODCALLTYPE GetData(FORMATETC *pFormatEtc, STGMEDIUM *PMedium);
	HRESULT STDMETHODCALLTYPE GetDataHere(FORMATETC *pFormatEtc, STGMEDIUM *PMedium);
	HRESULT STDMETHODCALLTYPE QueryGetData(FORMATETC *pFormatEtc);
	HRESULT STDMETHODCALLTYPE EnumFormatEtc(DWORD dwDirection, IEnumFORMATETC **ppFormatEtc);
	#else if BEOS
	#endif

public:
	GDragDropSource();
	~GDragDropSource();

	// Start a drag operation
	int Drag(int Effect);

	// Override these
	virtual void OnDragInit(bool Success) {}
	virtual char *TypeOf() { return NULL; }							// string identifing the type of data
	virtual bool GetData(uchar **Data, int *Size) { return false; }	// get the data
};

#ifdef WIN32
class GDragDropTarget : public IDropTarget {
#else if BEOS
class GDragDropTarget {
#endif
private:
	GWindow *To;
	GDragDropSource *DragDropObj;
	uchar *DragDropData;
	int DragDropLength;

	#ifdef WIN32
	// IUnknown
	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID iid, void **ppv);

	// IDropTarget
	HRESULT STDMETHODCALLTYPE DragEnter(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
	HRESULT STDMETHODCALLTYPE DragOver(DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);
	HRESULT STDMETHODCALLTYPE DragLeave(void);
	HRESULT STDMETHODCALLTYPE Drop(IDataObject *pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect);

	// Misc
	HGLOBAL GetDataHandle(IDataObject *pDataObject);
	bool UnwrapData(HGLOBAL hMem);
	#else if BEOS
	#endif

public:
	GDragDropTarget();
	~GDragDropTarget();

	// call this when you have a HWND
	void SetWindow(GWindow *To);

	// Override these
	virtual void OnDragInit(bool Success) {}
	virtual int WillAccept(GDragDropSource *Source, GdcPt2 Pt, int KeyState) { return DROPEFFECT_NONE; }
	virtual int OnDrop(uchar *Data, int Length, GdcPt2 Pt, int KeyState) { return DROPEFFECT_NONE; }
	virtual void OnDragEnter() {}
	virtual void OnDragExit() {}
};

#endif