/*hdr
**      FILE:           Inet.h
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Internet sockets classes
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __INET_H
#define __INET_H

#ifdef WIN32
#include <winsock.h>
#else if BEOS
#include <socket.h>
#include <netdb.h>
#endif

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

#define DEFAULT_TIMEOUT			30

class ISockets;
class IConnection;

class ISockets {

	bool SocketsOpen;

public:
	ISockets();
	virtual ~ISockets();

	bool IsOpen() { return SocketsOpen; }
	bool Open();
	void Close();
	IConnection *CreateConnection();
};

class IConnection : public GObject {

	friend class ISockets;

protected:
	ISockets *Parent;
	bool IsOK() { return (Parent) ? Parent->IsOpen() : FALSE; }

	#ifdef WIN32
	SOCKET		Socket;
	#else if BEOS
	int			Socket;
	#endif
	ulong		IpAddress;
	int			Port;
	hostent		*Host;
	sockaddr_in	RemoteAddr;

	bool		Connected;
	int			LastError;
	int			TimeOutSec;

public:
	int			BytesWritten;
	int			BytesRead;

	IConnection();
	virtual ~IConnection();

	bool Open(char *HostAddr, int port);
	void Close();
	bool Write(char *Data, int Len, int Flags);
	int Read(char *Data, int Len, int Flags);
	int ReadLen(char *Data, int Len, int Flags);
	int Error();

	int TimeOut() { return TimeOutSec; }
	void TimeOut(int t) { TimeOutSec = t; }

	virtual void OnRead(char *Data, int Len) {}
	virtual void OnWrite(char *Data, int Len) {}
	virtual void OnError(int ErrorCode, char *ErrorDescription) {}
	virtual void OnInfomation(char *Str) {}
	virtual void SetParameter(int Which, int What) {}
};

#endif