/*hdr
**	FILE:		TedView.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		23/9/98
**	DESCRIPTION:	Text editor view pane
**
**	Copyright (C) 1998, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>
#include <string.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

#include "GTextView.h"

// Defines
#define IDM_CUT					100
#define IDM_COPY				101
#define IDM_PASTE				102

// #define toupper(c) (((c)>='a'AND(c)<='z')?(c)-'a'+'A':(c))

// Strings
char *Untitled = "<Untitled>";
char *FindMiss = "Couldn't find text.";

///////////////////////////////////////////////////////////////////////////////////////////////
bool AlphaOrDigit(char c)
{
	return IsDigit(c) OR IsAlpha(c);
}

int CountLines(char *Str, int Len)
{
	int i = 1;
	for (; Str AND *Str AND Len>=0; Len--, Str++)
	{
		if (*Str == '\n') i++;
	}
	return i;
}

void StrCpyToCrLf(char *d, char *s, int Len)
{
	char *Start = d;

	if (s AND d)
	{
		while (*s AND Len--)
		{
			if (*s == 10)
			{
				*d++ = 13;
				*d++ = 10;
			}
			else if (*s != 13)
			{
				*d++ = *s;
			}
			s++;
		}
		*d = 0;
	}
}

void StrCpyToLf(char *d, char *s, int Len)
{
	if (s AND d)
	{
		while (*s AND Len--)
		{
			if (*s == 10)
			{
				*d++ = 10;
			}
			else if (*s != 13)
			{
				*d++ = *s;
			}
			s++;
		}
		*d = 0;
	}
}

/*
char *strnstr(char *s, char *find, int Len)
{
	if (s AND find)
	{
		Len -= max(0, strlen(find)-1);

		if (Len > 0)
		{
			while (*s AND Len--)
			{
				char *Str = s;
				char *Found = find;

				while (*Str AND *Found AND *Str == *Found)
				{
					Str++;
					Found++;
				}

				if (NOT *Found)
				{
					// found the string
					return s;
				}

				s++;
			}
		}
	}

	return NULL;
}

char *strinstr(char *s, char *find, int Len)
{
	if (s AND find)
	{
		Len = max(0, Len - (strlen(find) - 1));

		if (Len > 0)
		{
			while (*s AND Len--)
			{
				char *Str = s;
				char *Found = find;

				while (*Str AND *Found AND toupper(*Str) == toupper(*Found))
				{
					Str++;
					Found++;
				}

				if (NOT *Found)
				{
					// found the string
					return s;
				}

				s++;
			}
		}
	}

	return NULL;
}
*/

///////////////////////////////////////////////////////////////////////////////////////////////
#ifdef WIN32
GWin32Class GTextView::WndClass;
GTextView::GTextView(GWindow *notify, char *appname, LOGFONT *LogFont)
#else if BEOS
GTextView::GTextView(GWindow *notify, char *appname, BFont *LogFont) :
	GControl(this),
	BControl(	BRect(0, 0, 1, 1),
				NULL, "Text Editor", NULL,
				B_FOLLOW_NONE,
				B_WILL_DRAW | B_NAVIGABLE)
#endif
{
	#ifdef WIN32
	WndClass.Create(Class());
	#endif

	Notify = notify;
	AppName = appname;
	TopMargin = 4;
	LeftMargin = 3;
	Spacer = 2;
	LineHt = 20;
	Border = FLAT;
	ExtraFlags = 0;
	VScroll = 0;
	HScroll = 0;
	WrapAtCol = 80;
	UrlDetect = FALSE;
	SetScrollBars(FALSE, TRUE);

	MouseSelect = FALSE;
	bool FontCreated = FALSE;

	if (LogFont)
	{
#ifdef WIN32
		FontCreated = Font.CreateFont(LogFont);
#else if BEOS
		((BFont)Font) = *LogFont;
		FontCreated = true;
#endif
	}
	else
	{
#ifdef WIN32
		FontCreated = Font.CreateFont(	-12,
										0, 0, 0, 0, 0, 0, 0,
										ANSI_CHARSET,
										OUT_TT_PRECIS,
										CLIP_DEFAULT_PRECIS,
										DEFAULT_QUALITY,
										DEFAULT_PITCH,
										"Courier New");
#else if BEOS
		FontCreated = Font.CreateFont(	12,
										0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
										"Courier10 BT");
										
		SetViewColor(B_TRANSPARENT_COLOR);
#endif
	}

	if (FontCreated)
	{
		Font.Transparent(FALSE);
		LineHt = Font.Y("Ap");
		
		TextView::Flags |= TVF_DIRTY_BORDER | TVF_DIRTY_ALL;

		char Spaces[33];
		TabSize = 8;
		memset(Spaces, ' ', TabSize);
		Spaces[TabSize] = 0;
		TabStops = NEW(int[TABSTOPS]);
		if (TabStops)
		{
			int n = Font.X(Spaces, TabSize);
			for (int i=0; i<TABSTOPS; i++)
			{
				TabStops[i] = i * n;
			}
		}
	}

	Dirty(0);
	Doc.CursorCreate(&User, 0, 0);
	Old = User;
}

GTextView::~GTextView()
{
	DeleteArray(TabStops);
}

#ifdef WIN32
char *GTextView::Class()
{
	return LGI_TEXT;
}

DWORD GTextView::Style()
{
	return WS_CHILD | WS_TABSTOP | ExtraFlags;
}
#endif

void GTextView::Title(char *s)
{
	if (Notify)
	{
		Notify->OnEvent(&CreateMsg(WM_TITLE, 0, (int) s));
	}
}

void GTextView::SetStatus(int i, char *s)
{
	if (Notify)
	{
		Notify->OnEvent(&CreateMsg(WM_STATUS_UPDATE, i, (int) s));
	}
}

void GTextView::Dirty(int Type)
{
	TextView::Dirty(Type);

	if (TextView::Flags & TVF_DIRTY_BORDER)
	{
		if (Doc.GetName())
		{
			Title(Doc.GetName());
		}
		else
		{
			Title(Untitled);
		}
		TextView::Flags &= ~TVF_DIRTY_BORDER;
	}

	if (TextView::Flags & (TVF_DIRTY_SELECTION | TVF_DIRTY_ALL))
	{
		GWindow::Invalidate(NULL);
		return;
	}

	if (TextView::Flags & (TVF_DIRTY_TO_EOL | TVF_DIRTY_TO_EOP))
	{
		int x, y;
		CalcCursorPos(User, x, y);
		GdcRegion r(0, y, X(), y+Spacer+LineHt);
		GWindow::Invalidate(&r);

		if (TextView::Flags & TVF_DIRTY_TO_EOP)
		{
			r.y1 = r.y2 + 1;
			r.y2 = Y();
			GWindow::Invalidate(&r);
		}
		return;
	}

	if (TextView::Flags & TVF_DIRTY_CURSOR)
	{
		int x, y, Ox = 0, Oy = 0;

		/*
		switch (Border)
		{
			case SUNKEN:
			case RAISED:
			{
				Ox = 2;
				Oy = 2;
				break;
			}
		}
		*/

		CalcCursorPos(Old, x, y);
		GdcRegion OldR(x+Ox, y+Oy, x+Ox, y+Spacer+LineHt+Oy+1);
		GWindow::Invalidate(&OldR);

		CalcCursorPos(User, x, y);
		GdcRegion NewR(x+Ox, y+Oy, x+Ox, y+Spacer+LineHt+Oy+1);
		GWindow::Invalidate(&NewR);
	}
}

bool GTextView::ClearDirty(bool Ask)
{
	if (Doc.IsDirty())
	{
		int Result = IDYES;

		if (Ask)
		{
			char Str[256];
			char *Name = (Doc.GetName()) ? Doc.GetName() : Untitled;
			sprintf(Str, "Do you want to save '%s'?", Name);
			Result = LgiMsg(this, Str, "Save?", MB_YESNOCANCEL);
		}

		if (Result == IDYES)
		{
			// save the file
			if (Doc.GetName())
			{
				Save(Doc.GetName());
			}
			else
			{
				GFileSelect Select;

				Select.Parent(this);

				if (Doc.GetName())
				{
					Select.Name(Doc.GetName());
				}

				Select.Type("Code", "*.cpp;*.h;*.c");
				Select.Type("Text files", "*.txt;*.html;*.htm");
				Select.Type("All files", "*.*");

				if (Select.Save())
				{
					Save(Select.Name());
				}
			}
		}
		else if (Result == IDNO)
		{
			// do nothing
		}
		else
		{
			return FALSE;
		}
	}
	
	return TRUE;
}

void GTextView::CalcCursorPos(Cursor c, int &x, int &y)
{
	int Sy = Spacer + LineHt;
	char *Str = c.GetStr();
	if (Str)
	{
		x = LeftMargin + Font.X(Str, c.X()) - 1;
	}
	else
	{
		x = 0;
	}

	y = TopMargin + (Sy * (c.Y() - HiddenLines));
}

void GTextView::CalcMousePos(Cursor &c, VMouse m)
{
	int x = m.x;
	int y = m.y;

	#ifdef BEOS
	switch (Border)
	{
		case RAISED:
		case SUNKEN:
		{
			x -= 2;
			y -= 2;
		}
	}
	#endif

	int Sy = Spacer + LineHt;
	int Line = HiddenLines + (y / Sy);

	Doc.CursorCreate(&c, 0, limit(Line, 0, Doc.GetLines()-1));
	int CharX = Font.CharAt(x-LeftMargin, c.GetStr(), c.LineLength());

	if (Line < 0)
	{
		CharX = 0;
	}

	if (Line >= Doc.GetLines())
	{
		CharX = c.LineLength();
	}

	c.SetX(CharX);
}

bool GTextView::OnMoveCursor(int Dx, int Dy, bool NoSelect)
{
	Old = User;
	int Status = TextView::OnMoveCursor(Dx, Dy, NoSelect);
	if (Old.X() != User.X() OR Old.Y() != User.Y())
	{
		char Msg[64];
		sprintf(Msg, "(%i, %i)", User.X(), User.Y());
		SetStatus(1, Msg);
	}
	return Status;
}

void GTextView::OnSetCursor(int X, int Y)
{
	Old = User;
	TextView::OnSetCursor(X, Y);
	if (Old.X() != User.X() OR Old.Y() != User.Y())
	{
		Dirty(TVF_DIRTY_CURSOR);
		
		char Msg[64];
		sprintf(Msg, "(%i, %i)", User.X(), User.Y());
		SetStatus(1, Msg);
	}
}

void GTextView::DrawCursor(GdcBasePrimitives *pDC, Cursor &c, COLOUR col)
{
	if (pDC)
	{
		int x, y;
		CalcCursorPos(c, x, y);
		pDC->Colour(col);
		pDC->Line(x, y, x, y + Spacer + LineHt - 2);
	}
}

bool GTextView::ClipText(char *Str, int Len)
{
	bool Status = FALSE;
	if (Str)
	{
		if (Len < 0)
		{
			Len = strlen(Str);
		}

		int Lines = CountLines(Str, Len);
		char *Text = NEW(char[Len+Lines+1]);
		if (Text)
		{
			StrCpyToCrLf(Text, Str, Len);

			GClipBoard Clip(this);
			Clip.Empty();
			Clip.Text(Text);

			DeleteObj(Text);
		}
	}
	return Status;
}

char *GTextView::ClipText()
{
	GClipBoard Clip(this);

	DeleteObj(ClipData);
	char *Str = Clip.Text();
	if (Str)
	{
		ClipData = NEW(char[strlen(Str)+1]);
		if (ClipData)
		{
			StrCpyToLf(ClipData, Str, strlen(Str));
		}

		DeleteObj(Str);
	}

	return ClipData;
}

bool GTextView::Cut()
{
	bool Status = TextView::Cut();
	UpdateScrollBars();
	return Status;
}

bool GTextView::Paste()
{
	bool Status = TextView::Paste();
	UpdateScrollBars();
	return Status;
}

bool GTextView::Copy()
{
	return TextView::Copy();
}

bool GTextView::UpdateHiddenCheck()
{
	bool Status = TextView::UpdateHiddenCheck();
	if (Status)
	{
		UpdateScrollBars();
	}
	return Status;
}

void GTextView::OnSetHidden(int Hidden)
{
	TextView::OnSetHidden(Hidden);
	UpdateScrollBars();
}

void GTextView::UpdateScrollBars()
{
	int Sy = Spacer + LineHt;

	#ifdef WIN32
	RECT Client;
	if (hWnd AND GetClientRect(hWnd, &Client))
	{
		DisplayLines = ((Client.bottom - Client.top) / Sy);
	}

	// SetScrollBars(FALSE, Doc.GetLines() > DisplayLines);
	if (VScroll)
	{
		VScroll->SetLimits(0, Doc.GetLines()-1);
		VScroll->SetPage(DisplayLines);
		VScroll->SetPos(HiddenLines);
	}
	#else if BEOS
	
	BView *View = GetBView();
	if (View AND View->LockLooper())
	{
		BRect frame = GetBView()->Frame();
		DisplayLines = (frame.bottom-frame.top) / Sy;
		View->UnlockLooper();
	}
	
	#endif
}

void GTextView::OnPosChange()
{
	UpdateScrollBars();
	Dirty(TVF_DIRTY_ALL);
}

void GTextView::OnCreate()
{
	UpdateScrollBars();
}

bool GTextView::OnMultiLineTab(bool In)
{
	if ((TextView::Flags & TVF_SELECTION) AND Start.Y() != End.Y())
	{
		Cursor *a, *b;

		if (Start > End)
		{
			a = &End;
			b = &Start;
		}
		else
		{
			a = &Start;
			b = &End;
		}

		a->SetX(0);
		Cursor c = *a;

		for (; c.Y()<=b->Y(); c.MoveY(1))
		{
			if (c.Y() == b->Y() AND b->X() == 0)
			{
				break;
			}

			if (In)
			{
				if (c.LineLength() > 0)
				{
					char Tab[2] = { 9, 0 };
					Doc.Insert(&c, Tab, 1);
				}
			}
			else
			{
				char *Str = c;
				int Len = 0;

				if (*Str == 9)
				{
					Len = 1;
				}
				else
				{
					for (; *Str == ' ' AND Len <= 8; Str++)
					{
						Len++;
					}
				}

				if (Len)
				{
					Doc.Delete(&c, Len);
				}
			}

			if (c.Y() >= Doc.GetLines() - 1)
			{
				break;
			}
		}

		Dirty(TVF_DIRTY_ALL);

		bool MoveLastToEnd = (b->X() > 0);


		Doc.CursorCreate(a, 0, a->Y());
		a->SetX(0);
		Doc.CursorCreate(b, 0, b->Y());
		if (MoveLastToEnd) b->SetX(b->LineLength());
		Doc.CursorCreate(&User, User.X(), User.Y());

		return TRUE;
	}

	return FALSE;
}

void GTextView::AfterInsertText(char *Text, int Len)
{
	if (Text)
	{
		if (Len < 0)
		{
			Len = strlen(Text);
		}

		if (WrapType == TEXTED_WRAP_NONREFLOW AND User.X() >= WrapAtCol AND Len == 1)
		{
			Cursor Ms = User;

			OnMoveCursor(-1, 0, TRUE);
			
 			char c = *User;
			bool White = strchr(WhiteSpace, c) != 0;
			bool Delimit = strchr(Delimiters, c) != 0;
			bool Alpha = IsDigit(c) OR IsAlpha(c);

			if (NOT White)
			{
				while (	(NOT strchr(WhiteSpace, *User)) AND
					OnMoveCursor(-1, 0, TRUE))
				{
				}
			}

			int OldY = User.Y();
			OnMoveCursor(1, 0, TRUE);
			Doc.Insert(&User, "\n", 1);
			Doc.CursorCreate(&User, 10000, OldY+1);
			OnMoveCursor(-1, 0, TRUE);

			Dirty(TVF_DIRTY_ALL);
		}
	}
}

bool GTextView::OnFind()
{
	if (BtnFindNext() AND strlen(Find))
	{
		bool Match = DoFind();
		if (Match)
		{
			return GFindReplace::OnFind();
		}
		else
		{
			LgiMsg(this, FindMiss, (AppName)?AppName:"Find", MB_OK);
		}
	}

	return 0;
}

bool GTextView::OnReplace()
{
	int FindLen = strlen(Find);
	int ReplaceLen = strlen(Replace);

	if (FindLen AND ReplaceLen)
	{
		if (BtnFindNext())
		{
			bool Match = DoFind();
			if (NOT Match)
			{
				LgiMsg(this, FindMiss, (AppName)?AppName:"Replace", MB_OK);
			}
		}
		else if (BtnReplace())
		{
			Cursor StartPos = User;

			if (TextView::Flags & TVF_SELECTION)
			{
				if (Start < End)
				{
					User = Start;
				}
				else
				{
					User = End;
				}

				OnEndSelection();
			}

			if (DoFind())
			{
				if (StartPos == User)
				{
					// The current selection was ready to be
					// replaced
					OnDeleteSelection(FALSE);
					Doc.Insert(&User, Replace);
					Doc.CursorCreate(&User, User.X(), User.Y());
					User.MoveX(ReplaceLen);

					if (NOT DoFind())
					{
						return GFindReplace::OnReplace();
					}
				}
			}
			else
			{
				LgiMsg(this, FindMiss, (AppName)?AppName:"Replace", MB_OK);
			}
		}
		else if (BtnReplaceAll())
		{
			if (TextView::Flags & TVF_SELECTION)
			{
				if (Start < End)
				{
					User = Start;
				}
				else
				{
					User = End;
				}

				OnEndSelection();
			}

			int Items = 0;
			while (DoFind())
			{
				OnDeleteSelection(FALSE);
				Doc.Insert(&User, Replace);
				Doc.CursorCreate(&User, User.X(), User.Y());
				User.MoveX(ReplaceLen);

				Items++;
			}

			if (NOT Items)
			{
				LgiMsg(this, FindMiss, (AppName)?AppName:"Replace", MB_OK);
			}
			else
			{
				return GFindReplace::OnReplace();
			}
		}
	}

	return 0;
}

bool GTextView::DoFind()
{
	bool Match = FALSE;

	int Length = strlen(Find);
	if (Length > 0)
	{
		int XOff = User.X();

		// Find
		for (int Line=User.Y(); NOT Match AND Line<Doc.GetLines(); Line += 32)
		{
			int EndLine = min(Line+31, Doc.GetLines());
			TextLock Lock;
			if (Doc.Lock(&Lock, Line, EndLine))
			{
				for (int l=Line; NOT Match AND l<=EndLine; l++)
				{
					char *c = Lock[l];
					if (c)
					{
						char *Str = 0;
						int LineLength = StrLen(c);
						if (XOff < LineLength)
						{
							if (CaseSensitive())
							{
								Str = strnstr(c+XOff, Find, LineLength-XOff);
							}
							else
							{
								Str = strnistr(c+XOff, Find, LineLength-XOff);
							}

							if (Str)
							{
								Match = TRUE;
								int X = (int) (Str - c);

								Cursor Found;
								Doc.CursorCreate(&Found, X, l);

								if (WholeWord())
								{
									if (X > 0)
									{
										char Before = c[X-1];

										if (	NOT strchr(WhiteSpace, Before) AND
											NOT strchr(Delimiters, Before))
										{
											Match = FALSE;
										}
									}

									if (X + Length < LineLength - 1)
									{
										char After = c[X+Length];

										if (	NOT strchr(WhiteSpace, After) AND
											NOT strchr(Delimiters, After))
										{
											Match = FALSE;
										}
									}
								}

								if (Match)
								{
									if (TextView::Flags & TVF_SELECTION)
									{
										OnEndSelection();
									}

									OnStartSelection(&Found);
									End.MoveX(Length);
									User = End;
									UpdateHiddenCheck();
								}
							}
						}
					}

					XOff = 0;
				}

				Doc.UnLock(&Lock);
			}
		}
	}

	return Match;
}

// Selects a word or URL
void GTextView::SelectWord(VMouse &m)
{
	Cursor Ms;

	CalcMousePos(Ms, m);
	OnSetCursor(Ms.X(), Ms.Y());
	char c = *User;
	bool White = strchr(WhiteSpace, c) != 0;
	bool Delimit = strchr(Delimiters, c) != 0;
	bool Alpha = IsDigit(c) OR IsAlpha(c);

	// Url detect algorithm gets first crack at it
	if (UrlDetect AND NOT White)
	{
		bool UrlFound = FALSE;
		char *UrlDelim = "~/:%+-?@&$#.";

		// seek back to the start of a possible URL
		while (	(strchr(UrlDelim, *User) OR 
			AlphaOrDigit(*User)) AND
			OnMoveCursor(-1, 0))
		{
		}
		OnMoveCursor(1, 0);
		OnStartSelection(&User);
		TextView::Flags |= TVF_SHIFT;

		// find the end of the string
		while (	(strchr(UrlDelim, *User) OR 
			AlphaOrDigit(*User)) AND
			OnMoveCursor(1, 0))
		{
		}

		Cursor S;
		int Len;

		if (Start < End)
		{
			S = Start;
			Len = End - Start;
		}
		else
		{
			S = End;
			Len = Start - End;
		}
		
		char *Url = NEW(char[Len+1]);
		if (Url)
		{
			memcpy(Url, (char*) S, Len);
			Url[Len] = 0;

			// see if it's got the valid prefix
			char *Match[] = {	"http://",
						"ftp://",
						"mailto:",
						"www.",
						"ftp.",
						0};
			char *Prefix[] = {	"", "", "",
						"http://",
						"ftp://",
						0};
			for (int i=0; Match[i]; i++)
			{
				if (strnicmp(Url, Match[i], strlen(Match[i])) == 0)
				{
					// Ok so trigger the event

					char *Temp = NEW(char[strlen(Prefix[i])+strlen(Url)+1]);
					if (Temp)
					{
						*Temp = 0;
						strcat(Temp, Prefix[i]);
						strcat(Temp, Url);
						OnUrl(Url);
						DeleteArray(Temp);
					}
					UrlFound = TRUE;
					break;
				}
			}

			if (NOT UrlFound AND strchr(Url, '@'))
			{
				// possibly a email address without the "mailto:"
				// some better detection may be in order here.
				OnUrl(Url);
				UrlFound = TRUE;
			}

			DeleteArray(Url);
		}

		if (UrlFound)
		{
			// no need to hang around
			return;
		}

		// put the cursor back to it's original location for the normal word find
		OnSetCursor(Ms.X(), Ms.Y());
	}

	// Normal select word algorithm
	while (	((White AND strchr(WhiteSpace, *User)) OR 
			(Delimit AND strchr(Delimiters, *User)) OR 
			(Alpha	AND NOT strchr(WhiteSpace, *User)
			AND NOT strchr(Delimiters, *User))) AND
			OnMoveCursor(-1, 0))
	{
	}
	OnMoveCursor(1, 0);
	OnStartSelection(&User);
	TextView::Flags |= TVF_SHIFT;

	while (	((White AND strchr(WhiteSpace, *User)) OR 
			(Delimit AND strchr(Delimiters, *User)) OR 
			(Alpha	AND NOT strchr(WhiteSpace, *User)
			AND NOT strchr(Delimiters, *User))) AND
			OnMoveCursor(1, 0))
	{
	}
}

void GTextView::OnKey(VKey &K)
{
	printf("GTextView::OnKey(%c, %i, %X, %X, %i, %i)\n", K.c, K.c, K.Flags, K.Data, K.Down, K.IsChar);

	if (NOT K.IsChar AND K.Down)
	{
		switch (K.c)
		{
			case 'x':
			case 'X':
			{
				if (K.Flags & MK_CTRL)
				{
					Cut();
				}
				return;
			}
			case 'c':
			case 'C':
			{
				if (K.Flags & MK_CTRL)
				{
					Copy();
				}					
				return;
			}
			case 'v':
			case 'V':
			{
				if (K.Flags & MK_CTRL)
				{
					Paste();
				}
				return;
			}
			case 's':
			case 'S':
			{
				if (K.Flags & MK_CTRL)
				{
					OnSave();
				}
				return;
			}
			case VK_F3:
			{
				if (NOT DoFind())
				{
					LgiMsg(this, FindMiss, (AppName)?AppName:"Find", MB_OK);
				}
				return;
			}
			case 'f':
			case 'F':
			{
				if (K.Flags & MK_CTRL)
				{
					OpenFind();
				}
				break;
			}
			case 'h':
			case 'H':
			{
				if (K.Flags & MK_CTRL)
				{
					OpenReplace();
				}
				break;
			}
		}
	}

	ProcessKey(K);
}

void GTextView::OnPulse()
{
	if (MouseSelect)
	{
		static GdcPt2 Last(-1, -1);
		GdcPt2 Pt(0, 0);
		#ifdef WIN32
		LgiGetCursorPos(&Pt);
		POINT p;
		p.x = Pt.x;
		p.y = Pt.y;
		ScreenToClient(hWnd, &p);
		Pt.x = p.x;
		Pt.y = p.y;
		#else if BEOS
		BPoint Cur;
		uint32 Button;
		GetBView()->GetMouse(&Cur, &Button);
		Pt.x = Cur.x;
		Pt.y = Cur.y;
		#endif

		if (Last.x != Pt.x OR Last.y != Pt.y)
		{
			VMouse m;
			ZeroObj(m);
			m.x = Pt.x;
			m.y = Pt.y;

			Cursor Ms;
			CalcMousePos(Ms, m);
			OnSetCursor(Ms.X(), Ms.Y());
		}
		
		Last = Pt;
	}
}

void GTextView::OnMouseClick(VMouse &m)
{
	#ifdef WIN32
	Focus(TRUE);
	#else if BEOS
	BControl::MakeFocus(true);
	#endif

	if (m.Down)
	{
		if (m.Flags & MK_LEFT)
		{
			if (m.Double)
			{
				SelectWord(m);
			}
			else
			{
				// start selection with mouse
				Cursor Ms;
				CalcMousePos(Ms, m);
				OnSetCursor(Ms.X(), Ms.Y());
				OnStartSelection(&User);

				Capture(TRUE);
				MouseSelect = TRUE;
				SetPulse(100);
			}
		}
		
		if (m.Flags & MK_RIGHT)
		{
			// context menu
			GdcPt2 Pt;
			if (LgiGetCursorPos(&Pt))
			{
				GSubMenu *RClick = NEW(GSubMenu);
				if (RClick)
				{
					RClick->AppendItem("Cut", IDM_CUT, TRUE);
					RClick->AppendItem("Copy", IDM_COPY, TRUE);
					RClick->AppendItem("Paste", IDM_PASTE, TRUE);
				}

				// SendMessage(hWnd, WM_COMMAND, (WPARAM) MAKELONG(Id, 0), (LPARAM) hWnd);

				int Id = RClick->Float(this, Pt.x, Pt.y);
				switch (Id)
				{
					case IDM_CUT:
					{
						Cut();
						break;
					}
					case IDM_COPY:
					{
						Copy();
						break;
					}
					case IDM_PASTE:
					{
						Paste();
						break;
					}
				}
			}
		}
	}
	else
	{
		if (MouseSelect)
		{
			Capture(false);
			MouseSelect = FALSE;

			if (Start == End)
			{
				OnEndSelection();
			}

			SetPulse();
		}
	}
}

void GTextView::OnMouseMove(VMouse &m)
{
}

void GTextView::OnPaint(GdcBasePrimitives *pDC)
{
	#ifdef WIN32
	COLOUR Fore = GetSysColor(COLOR_BTNTEXT);
	COLOUR Back = GetSysColor(COLOR_WINDOW);
	#else if BEOS
	COLOUR Fore = Rgb24(0, 0, 0);
	COLOUR Back = Rgb24(255, 255, 255);
	#endif
	COLOUR Cursor = 0;
	int Sy = Spacer + LineHt;

	Font.Colour(Fore, Back);
	Font.SetTabs(TABSTOPS, TabStops, LeftMargin);
	Back = CBit(pDC->GetBits(), Back);

	TextLock Lock;
	Doc.Lock(&Lock, HiddenLines, DisplayLines+1);

	// update everthing
	char *StartPtr = 0, *EndPtr = 0;
	int Sel = 0;
	if (TextView::Flags & TVF_SELECTION)
	{
		Sel = 1;
		if (Start > End)
		{
			StartPtr = End;
			EndPtr = Start;
		}
		else
		{
			StartPtr = Start;
			EndPtr = End;
		}
	}

	char *Next = StartPtr;
	char *StartLock = Lock[HiddenLines];
	if (StartPtr AND (StartLock > StartPtr) AND (NOT (StartLock > EndPtr)))
	{
		Font.Colour(Font.Back(), Font.Fore());

		Next = EndPtr;
		Sel = 2;
	}

	// Set clip region
	GdcRegion r(0, 0, Pos.X()-1, Pos.Y()-1);

	#ifdef BEOS
	switch (Border)
	{
		case SUNKEN:
		case RAISED:
		{
			DrawMoveDeepEdge(pDC, r, Border);
			pDC->ClipRgn(&r);
			// pDC->SetOrigin(-r.x1, -r.y1);
			break;
		}
	}
	#endif

	int ScrY = TopMargin;
	int EndY = ScrY + Sy - 1;

	// Draw top margin
	pDC->Colour(Back);
	pDC->Rectangle(0, 0, pDC->X(), TopMargin-1);

	for (int i=0; i<=DisplayLines; i++, ScrY+=Sy, EndY+=Sy)
	{
		char *BeginLine = Lock[HiddenLines+i];
		char *c = BeginLine;

		pDC->Colour(Back);
		if (c)
		{	
			int Len = StrLen(c);
			int SpacerY = ScrY + LineHt;
			int DrawX = LeftMargin;
			int TextX = Font.X(c, Len);

			// to left margin
			pDC->Rectangle(0, ScrY, LeftMargin-1, EndY);
			
			while (Sel AND Next >= c AND Next <= c + Len)
			{
				int ThisSide = (int) (Next - c);
				int Width = Font.X(BeginLine, (int) (Next - BeginLine)) - Font.X(BeginLine, (int) (c - BeginLine));

				Font.Text(	pDC,
							DrawX,
							ScrY,
							c,
							ThisSide);

				// spacer below text
				pDC->Colour(CBit(pDC->GetBits(), Font.Back()));
				pDC->Rectangle(DrawX, SpacerY, DrawX+Width, EndY);

				Font.Colour(Font.Back(), Font.Fore());
				DrawX += Width;
				TextX -= Width;
				c += ThisSide;
				Len -= ThisSide;
				
				if (Sel == 1)
				{
					Sel = 2;
					Next = EndPtr;
				}
				else
				{
					Sel = 0;
					Next = 0;
				}
			}

			Font.Text(	pDC,
						DrawX,
						ScrY,
						c,
						Len);

			// draw space below text
			pDC->Colour(CBit(pDC->GetBits(), Font.Back()));
			pDC->Rectangle(DrawX, SpacerY, DrawX+TextX, EndY);

			int RightMargin = DrawX + TextX;
			if (Sel == 2)
			{
				// the EOL char needs to be seen when selected
				RightMargin += 5;
				pDC->Rectangle(DrawX+TextX, ScrY, RightMargin-1, EndY);
			}

			pDC->Colour(Back);

			// from right margin
			pDC->Rectangle(RightMargin, ScrY, pDC->X(), EndY);
		}
		else
		{
			pDC->Rectangle(0, ScrY, pDC->X()-1, EndY);
		}
	}

	// draw from the last text down
	pDC->Colour(Back);
	pDC->Rectangle(0, ScrY, pDC->X(), pDC->Y());

	// put the cursor back
	if (Focus())
	{
		DrawCursor(pDC, User, Cursor);
	}

	Doc.UnLock(&Lock);
	Clean();
}

void GTextView::OnFocus(bool f)
{
	if (f)
	{
		SetFlag(TextView::Flags, TVF_FOCUS);
	}
	else
	{
		ClearFlag(TextView::Flags, TVF_FOCUS);
	}
	Dirty(TVF_DIRTY_CURSOR);
}

int GTextView::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		#ifdef WIN32
		case WM_NCPAINT:
		{
			switch (Border)
			{
				case SUNKEN:
				case RAISED:
				{
					HDC hDC = GetWindowDC(hWnd);
					if (hDC)
					{
						GdcGdiScreen DC(hDC, hWnd);
						GdcRegion r(0, 0, Pos.X()-1, Pos.Y()-1);
						DrawMoveDeepEdge(&DC, r, Border);
						ReleaseDC(hWnd, hDC);
					}
					break;
				}
			}
			break;
		}
		case WM_NCCALCSIZE:
		{
			// do default processing on the message
			int Status = DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);

			if (Msg->a)
			{
				NCCALCSIZE_PARAMS *i = (NCCALCSIZE_PARAMS*) Msg->b;

				switch (Border)
				{
					case SUNKEN:
					case RAISED:
					{
						i->rgrc[0].top += 2;
						i->rgrc[0].left += 2;
						i->rgrc[0].right -= 2;
						i->rgrc[0].bottom -= 2;
						break;
					}
				}
			}
			else
			{
				RECT *i = (RECT*) Msg->b;
				i->top += 2;
				i->left += 2;
				i->right -= 2;
				i->bottom -= 2;
			}

			return Status;
			break;
		}
		case WM_GETDLGCODE:
		{
			return DLGC_WANTALLKEYS;
			break;
		}
		case WM_CREATE:
		{
			GFindReplace::SetParent(GWindow::Handle());
			GTextView::OnCreate();
			break;
		}
		case WM_MOUSEWHEEL:
		{
			if (Doc.GetLines() > DisplayLines)
			{
				short fwKeys = LOWORD(Msg->a);			// key flags
				short zDelta = (short) HIWORD(Msg->a);	// wheel rotation
				short xPos = (short) LOWORD(Msg->b);	// horizontal position of pointer
				short yPos = (short) HIWORD(Msg->b);	// vertical position of pointer
				UINT nScrollLines = 1;

				SystemParametersInfo(	SPI_GETWHEELSCROLLLINES, 
							0, 
							(PVOID) &nScrollLines, 
							0); 
				
				int NewHidden = limit(	(int) HiddenLines - (((int) zDelta / WHEEL_DELTA) * (int) nScrollLines),
							0,
							Doc.GetLines()-DisplayLines);

				if (NewHidden != HiddenLines)
				{
					OnSetHidden(NewHidden);
				}
			}
			return 0;
		}
		#endif
	}

	int Status = GWindow::OnEvent(Msg);
	GFindReplace::OnEvent(Msg);

	switch (Message(Msg))
	{
		#ifdef WIN32
		case WM_VSCROLL:
		{
			if (Doc.GetLines() > DisplayLines)
			{
				int nScrollCode = (int) LOWORD(Msg->a);
				int nPos = (short int) HIWORD(Msg->a);
				HWND hwndScrollBar = (HWND) Msg->b;

				switch (nScrollCode)
				{
					case SB_BOTTOM:
					case SB_LINEDOWN:
					{
						if (HiddenLines < Doc.GetLines()-DisplayLines)
						{
							OnSetHidden(HiddenLines+1);
						}
						break;
					}
					case SB_TOP:
					case SB_LINEUP:
					{
						if (HiddenLines > 0)
						{
							OnSetHidden(HiddenLines-1);
						}
						break;
					}
					case SB_PAGEDOWN:
					{
						int Move = DisplayLines-1;
						OnSetHidden(HiddenLines+Move);
						OnMoveCursor(0, Move);
						break;
					}
					case SB_PAGEUP:
					{
						int Move = 1-DisplayLines;
						OnSetHidden(HiddenLines+Move);
						OnMoveCursor(0, Move);
						break;
					}
					case SB_THUMBPOSITION:
					case SB_THUMBTRACK:
					{
						OnSetHidden(nPos);
						break;
					}
				}
			}
			break;
		}
		#endif
	}

	return Status;
}

bool GTextView::Open(char *Name)
{
	bool Status = TextView::Open(Name);
	if (Status)
	{
		Doc.CursorCreate(&User, 0, 0);
		SetStatus(1, "(0, 0)");

		HiddenLines = 0;
		UpdateScrollBars();
	}
	return Status;
}

bool GTextView::Name(char *Name)
{
	bool Status = FALSE;
	if (Name)
	{
		Status = Doc.Import(Name);
		UpdateScrollBars();
	}
	return Status;
}

char *GTextView::Name()
{
	Cursor C;
	if (Doc.CursorCreate(&C, 0, 0))
	{
		return (char*) C;
	}
	return NULL;
}

bool GTextView::SetScrollBars(bool x, bool y)
{
	bool Status = FALSE;
	static bool Processing = FALSE;

	#ifdef WIN32
	if (NOT Processing)
	{
		Processing = TRUE;

		if (y AND NOT VScroll)
		{
			VScroll = NEW(GScrollBar);
			if (VScroll)
			{
				ExtraFlags |= WS_VSCROLL;
				VScroll->SetVertical(TRUE);
				VScroll->SetParent(this);
				Status = TRUE;
			}
		}
		if (NOT y AND VScroll)
		{
			ExtraFlags &= ~WS_VSCROLL;
			DeleteObj(VScroll);
			Status = TRUE;
		}
		
		if (x AND NOT HScroll)
		{
			HScroll = NEW(GScrollBar);
			if (HScroll)
			{
				ExtraFlags |= WS_HSCROLL;
				HScroll->SetVertical(FALSE);
				HScroll->SetParent(this);
				Status = TRUE;
			}
		}
		if (NOT x AND HScroll)
		{
			ExtraFlags &= ~WS_HSCROLL;
			DeleteObj(HScroll);
			Status = TRUE;
		}

		Processing = FALSE;
	}
	#endif

	return Status;
}

void GTextView::OnUrl(char *Url)
{
}
