/*hdr
**      FILE:           GuiDlg.h
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998 Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#ifndef __DLG_H
#define __DLG_H

/////////////////////////////////////////////////////////////////////////////////////////
#define DialogToPixelX(i)		(((i)*Bx)/4)
#define DialogToPixelY(i)		(((i)*By)/8)
#define PixelToDialogX(i)		(((i)*4)/Bx)
#define PixelToDialogY(i)		(((i)*8)/By)

/////////////////////////////////////////////////////////////////////////////////////////
#define IDM_OK				99
#define IDM_CANCEL			98
#define IDM_STATIC			0

#define DIALOG_X			1.56
#define DIALOG_Y			1.76
#define CTRL_X				1.50
#define CTRL_Y				1.64

#define CTRL_SUNKEN			0x80000000
#define CTRL_FLAT			0x40000000
#define CTRL_RAISED			0x20000000

/////////////////////////////////////////////////////////////////////////////////////////
class GControlGroup;
class GDialog;
class GDlgFunc;
class GListItem;
class GTreeItem;

class GButton;
class GEdit;
class GCheckBox;
class GText;
class GRadioGroup;
class GRadioButton;
class GTabControl;
class GTabPage;
class GSlider;
class GCombo;
class GBitmap;
class GList;
class GTree;


/////////////////////////////////////////////////////////////////////////////////////////
class GControlInfo {
public:
	DLGITEMTEMPLATE Item;
	short Atom;
	char *Class;
	char *Text;
	GControl *Ctrl;

	GControlInfo();
	~GControlInfo();
};

class GDlgFunc {
public:
	short *StrCopy(short *A, char *N);
	short *Push(short *A, int n);
	short *PadToDWord(short *A, bool Seek = FALSE);
	void PixelToDialog(GdcRegion &r);
	void DialogToPixel(GdcRegion &r);
	short *StrSkip(short *Ptr);
	short *StrExtract(short *Ptr, char *Str, short *Atom = NULL);
	bool LoadFromResource(char *Resource, List<GWindow> &Children, GdcRegion *Pos = 0, char *Name = 0);
	virtual GControl *CreateControl(GControlInfo *Info);
};

///////////////////////////////////////////////////////////////////////////////////////////////
class GControlGroup : public GWindow {

public:
	virtual GControl *FindControl(int Id);
	virtual GControl *GetControl(HWND hCtrl);

	int OnEvent(GMessage *Msg);
};

///////////////////////////////////////////////////////////////////////////////////////////////
class GDialog : public GControlGroup, public GDlgFunc {

	friend class GControl;

private:
	bool IsResizable;
	int SizeOfControls();
	GMem *Mem;

protected:
	void MoveOnScreen();
	void MoveToCenter();
	void MoveToMouse();

	bool LoadFromResource(char *Resource);

public:
	GDialog();
	~GDialog();

	bool Resizable() { return IsResizable; }
	void Resizable(bool r) { IsResizable = r; }
	
	virtual int DoModal();
	virtual void EndModal(int Code = 0);

	virtual int DoModeless();
	virtual void EndModeless(int Code = 0);

	int OnEvent(GMessage *Msg);
};

class GControl : public GWindow, public GDlgFunc, public GId {

	friend class GDialog;
	friend class GControlGroup;

private:
	virtual char *Class() { return NULL; }
	virtual DWORD ExStyle() { return 0; }
	virtual DWORD Style() { return WS_CHILD; }
	virtual int SysOnNotify(int Code) { return 0; }

	// obselete
	virtual int SizeOf() { return 0; }
	virtual void *WriteResource(void *Ptr) { return Ptr; }

protected:
	GdcPt2 SizeOfStr(char *Str);

public:
	GControl();
	~GControl();

	bool Sunken() { return (WndFlags & CTRL_SUNKEN) != 0; }
	void Sunken(bool n) { if (n) WndFlags |= CTRL_SUNKEN; else WndFlags &= ~CTRL_SUNKEN; }
	bool Flat() { return (WndFlags & CTRL_FLAT) != 0; }
	void Flat(bool n) { if (n) WndFlags |= CTRL_FLAT; else WndFlags &= ~CTRL_FLAT; }
	bool Raised() { return (WndFlags & CTRL_RAISED) != 0; }
	void Raised(bool n) { if (n) WndFlags |= CTRL_RAISED; else WndFlags &= ~CTRL_RAISED; }

	char *Name();
	bool Name(char *n);
	int OnEvent(GMessage *Msg);
	bool Attach(GWindow *parent);
	virtual GControl *IsControl() { return this; }
	virtual GControl *GetControl(HWND hCtrl) { if (hWnd == hCtrl) return this; else return 0; }
	virtual GControl *FindControl(int Id) { if (GetId() == Id) return this; else return 0; }
};

//////////////////////////////////////////////////////////////////////////////////////
// Controls
#define LGI_BUTTON			"LGI_Button"
#define LGI_EDITBOX			"LGI_Editbox"
#define LGI_CHECKBOX			"LGI_CheckBox"
#define LGI_STATIC			"LGI_Static"
#define LGI_RADIOGROUP			"LGI_RadioGroup"
#define LGI_TABCONTROL			"LGI_TabControl"
#define LGI_TABPAGE			"LGI_TabPage"
#define LGI_SLIDER			"LGI_Slider"
#define LGI_COMBO			"LGI_Combo"
#define LGI_BITMAP			"LGI_Bitmap"
#define LGI_LIST			"LGI_ListView"
#define LGI_TREE			"LGI_TreeView"
#define LGI_PROGRESS			"LGI_Progress"

// Button
class GButton : public GControl {

	static GWin32Class WndClass;
	char *Class() { return LGI_BUTTON; }
	DWORD Style() { return ((Def) ? BS_DEFPUSHBUTTON : BS_PUSHBUTTON) |
				BS_TEXT |
				WS_CHILD |
				WS_TABSTOP; }

	bool Def;

public:
	GButton(int id, int x, int y, int cx, int cy, char *name);
	~GButton();

	bool Default();
	void Default(bool b);
	int OnEvent(GMessage *Msg);
};

// Edit
class GEdit : public GControl {

	static GWin32Class WndClass;

protected:
	bool MultiLineEdit;
	bool PasswordEdit;
	bool Enter;

	char *Class() { return LGI_EDITBOX; }
	DWORD ExStyle() { return WS_EX_CLIENTEDGE; }
	DWORD Style() { return	WS_CHILD |
				ES_AUTOHSCROLL |
				WS_TABSTOP |
				((Enter) ? ES_WANTRETURN : 0) |
				((MultiLineEdit) ? ES_MULTILINE | ES_AUTOVSCROLL | ES_WANTRETURN | WS_VSCROLL: 0) |
				((PasswordEdit) ? ES_PASSWORD : 0); }
	int SysOnNotify(int Code);

public:
	GEdit(int id, int x, int y, int cx, int cy, char *name);
	~GEdit();

	bool WantReturn() { return Enter; }
	void WantReturn(bool b) { Enter = b; }
	bool MultiLine() { return MultiLineEdit; }
	void MultiLine(bool m) { MultiLineEdit = m; }
	bool Password() { return PasswordEdit; }
	void Password(bool m) { PasswordEdit = m; }

	void Value(int i);
	int Value();
	int OnEvent(GMessage *Msg);
};

// Check box
class GCheckBox : public GControl {

	static GWin32Class WndClass;
	int Val;
	bool Three;

	char *Class() { return LGI_CHECKBOX; }
	DWORD ExStyle() { return 0; }
	DWORD Style() { return	WS_CHILD |
				WS_TABSTOP |
				((Three) ? BS_AUTO3STATE : BS_AUTOCHECKBOX); }

public:
	GCheckBox(int id, int x, int y, int cx, int cy, char *name, int InitState = FALSE);
	~GCheckBox();

	bool ThreeState() { return Three; }
	void ThreeState(bool t) { Three = t; }

	int Value();
	void Value(int b);
	int OnEvent(GMessage *Msg);
};

// Static
class GText : public GControl {

	static GWin32Class WndClass;

	char *Class() { return LGI_STATIC; }
	DWORD ExStyle() { return 0; }
	DWORD Style() { return WS_CHILD; }

public:
	GText(int id, int x, int y, int cx, int cy, char *name);
	~GText();
	int OnEvent(GMessage *Msg);
};

// Group of radio buttons
typedef List<GRadioButton> RadioButtonList;

class GRadioGroup : public GControl {

	static GWin32Class WndClass;
	static int NextId;

	char *Class() { return LGI_RADIOGROUP; }
	DWORD ExStyle() { return 0; }
	DWORD Style() { return WS_CHILD | BS_GROUPBOX; }
	bool Attach(GWindow *parent);

protected:
	RadioButtonList Buttons;
	int InitialValue;

public:
	GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GRadioGroup();

	int Value();
	void Value(int i);

	int OnEvent(GMessage *Msg);
	GRadioButton *Append(int x, int y, char *name);
};

class GRadioButton : public GControl {

	friend class GRadioGroup;

	static GWin32Class WndClass;
	char *Class() { return LGI_BUTTON; }
	DWORD Style() { return BS_AUTORADIOBUTTON | BS_TEXT | WS_CHILD | ((First) ? WS_GROUP : 0); }
	
	bool First;

public:
	GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first = FALSE);
	~GRadioButton();

	int Value();
	void Value(int i);
	int OnEvent(GMessage *Msg);

};

// Tab control
typedef List<GTabPage> TabPageList;
class GTabControl : public GControl {

	friend class GTabPage;
	static GWin32Class WndClass;

	char *Class() { return LGI_TABCONTROL; }
	DWORD Style() { return WS_CHILD | WS_TABSTOP; }
	bool Attach(GWindow *parent);
	GControl *GetControl(HWND hCtrl);
	GControl *FindControl(int Id);

	int Current;
	int NextTab;
	int SysOnNotify(int Code);

protected:
	TabPageList Tabs;
	int InitialValue;

public:
	GTabControl(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GTabControl();

	int Value();
	void Value(int i);

	int OnEvent(GMessage *Msg);
	GTabPage *Append(char *name);
	GdcRegion GetClient();
};

class GTabPage : public GObject, public GDlgFunc {

	friend class GTabControl;
	GTabControl *Parent;
	GdcRegion Pos;
	bool Active;

	int Index;
	TC_ITEM Item;
	WindowList Children;
	
	bool Attach(GTabControl *parent, int Index);
	GControl *GetControl(HWND hCtrl);

public:
	GTabPage(char *name);
	~GTabPage();

	bool Enabled();
	void Enabled(bool e);
	bool Visible();
	void Visible(bool v);
	char *Name();
	bool Name(char *n);

	int OnEvent(GMessage *Msg);
	void Append(GWindow *Wnd) { Children.Insert(Wnd); }
	bool AppendDialogFromResource(char *Name);
};

// Slider
class GSlider : public GControl {

	static GWin32Class WndClass;
	bool Vertical;
	int Min, Max;
	int Val;

	char *Class() { return LGI_SLIDER; }
	DWORD ExStyle() { return 0; }
	DWORD Style() { return WS_CHILD | WS_TABSTOP | ((Vertical) ? TBS_VERT : TBS_HORZ); }
	int SysOnNotify(int Code);

public:
	GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert);
	~GSlider();

	void Value(int i);
	int Value();
	void GetLimits(int &x, int &y);
	void SetLimits(int x, int y);
	int OnEvent(GMessage *Msg);
};

// Combo Box
class GCombo : public GControl, public List<char> {

	static GWin32Class WndClass;
	char *Class() { return LGI_COMBO; }
	DWORD Style() { return WS_VSCROLL | CBS_DISABLENOSCROLL | CBS_DROPDOWN | CBS_AUTOHSCROLL | WS_CHILD | WS_TABSTOP | ((SortItems) ? CBS_SORT : 0); }
	int SysOnNotify(int Code);

	bool SortItems;

public:
	GCombo(int id, int x, int y, int cx, int cy, char *name);
	~GCombo();

	bool Sort() { return SortItems; }
	void Sort(bool s) { SortItems = s; }
	void Value(int i);
	int Value();

	int OnEvent(GMessage *Msg);
	bool Delete();
	bool Delete(int i);
	bool Delete(char *p);
	bool Insert(char *p, int Index = -1);
};

// Bitmap
class GBitmap : public GControl {

	static GWin32Class WndClass;
	char *Class() { return LGI_BITMAP; }
	DWORD Style() { return WS_CHILD; }

	GdcBasePrimitives *pDC;

public:
	GBitmap(int id, int x, int y, char *FileName);
	~GBitmap();

	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	void OnMouseClick(VMouse &m);
};

//////////////////////////////////////////////////////////////////////////////
#define GIC_OWNER_DRAW			0x0001
#define GIC_ASK_TEXT			0x0002
#define GIC_ASK_IMAGE			0x0004
#define GIC_OWN_LIST			0x0008

class GItemContainer {
protected:
	int Flags;
	GImageList *ImageList;

public:
	GItemContainer();
	~GItemContainer();

	bool OwnerDraw() { return TestFlag(Flags, GIC_OWNER_DRAW); }
	void OwnerDraw(bool b) { if (b) SetFlag(Flags, GIC_OWNER_DRAW); else ClearFlag(Flags, GIC_OWNER_DRAW); }
	bool AskText() { return TestFlag(Flags, GIC_ASK_TEXT); }
	void AskText(bool b) { if (b) SetFlag(Flags, GIC_ASK_TEXT); else ClearFlag(Flags, GIC_ASK_TEXT); }
	bool AskImage() { return TestFlag(Flags, GIC_ASK_IMAGE); }
	void AskImage(bool b) { if (b) SetFlag(Flags, GIC_ASK_IMAGE); else ClearFlag(Flags, GIC_ASK_IMAGE); }
	bool OwnList() { return TestFlag(Flags, GIC_OWN_LIST); }
	void OwnList(bool b) { if (b) SetFlag(Flags, GIC_OWN_LIST); else ClearFlag(Flags, GIC_OWN_LIST); }
	GImageList *GetList() { return ImageList; }

	bool SetImageList(GImageList *List, bool Own = TRUE);
	bool Load(char *File);
};

class GItem : public GEvent{
protected:
public:
	// Events
	void OnMouseClick(VMouse &m) {}
	void OnPaint(GdcBasePrimitives *pDC) {}
	virtual void OnBeginDrag(VMouse &m) {}
	virtual void OnMeasure(MEASUREITEMSTRUCT *Info) {}
	virtual void Update() {}

	// State
	virtual bool Selected() { return FALSE; }
	virtual void Selected(bool b) {}

	// Data
	virtual char *GetText(int i=0) { return 0; }
	virtual bool SetText(char *s, int i=0) { return FALSE; }
	virtual int GetImage(int Flags = 0) { return -1; }
	virtual void SetImage(int i) {}
};

// List view
class GListItem : public GItem {

	friend class GList;

protected:
	GList *Parent;
	int Cols;
	char **Str;

	bool GetItem(LV_ITEM *Item);

public:
	GListItem();
	virtual ~GListItem();

	virtual int GetCols() { return Cols; }
	char *GetText(int i);
	bool SetText(char *s, int i=0);
	void Update();
	bool Selected();
	void Selected(bool b);
};

class GList : public GControl, public List<GListItem>, public GItemContainer {

	static GWin32Class WndClass;
	friend class GListItem;

protected:
	List<LV_COLUMN> Columns;
	bool ColumnHeader;
	bool EditLabels;
	int Val;

	char *Class() { return LGI_LIST; }
	DWORD Style() { return	LBS_NOTIFY |
				((ColumnHeader) ? 0 : LVS_NOCOLUMNHEADER) |
				((EditLabels) ? LVS_EDITLABELS : 0) |
				LVS_REPORT |
				WS_CHILD |
				WS_TABSTOP; }
	DWORD ExStyle() { return (WndFlags & CTRL_SUNKEN) ? WS_EX_CLIENTEDGE : 0; }

	bool SysAddItem(GListItem *Item, int Pos);
	void SysRemoveItem(int Pos);
	int SysOnNotify(int Code);
	void SysUpdate(GListItem *Item);

public:
	GList(int id, int x, int y, int cx, int cy, char *name = "List");
	~GList();

	// Events
	virtual void OnItemClick(GListItem *Item, VMouse &m);
	virtual void OnItemBeginDrag(GListItem *Item, int Flags);
	virtual void OnItemSelect(GListItem *Item) {}
	virtual void OnColumnClick(int Col) {}

	// properties
	void AddColumn(char *Name, int Width = 50, int Align = LVCFMT_LEFT);
	void EmptyColumns();
	bool ShowColumnHeader();
	void ShowColumnHeader(bool Show);
	bool AllowEditLabels() { return EditLabels; }
	void AllowEditLabels(bool b) { EditLabels = b; }

	// methods
	int OnEvent(GMessage *Msg);
	void OnMouseClick(VMouse &m);
	int Value();
	void Value(int i);
	bool Select(GListItem *Obj);
	bool GetSelection(List<GListItem> &l);
	bool GetAll(List<GListItem> &l);

	// list
	GListItem *Current();
	bool IsEmpty() { return GetItems() == 0; }
	bool Delete();
	bool Delete(int Index);
	bool Delete(GListItem *p);
	bool Insert(GListItem *p, int Index = -1);
	bool Remove(GListItem *Obj);

	void Empty();			// delete all items
	void RemoveAll();		// remove references to externally owned items
};

// Tree view
class GTreeItem : public GItem {

	friend class GTree;

protected:
	char *Str;
	GTree *Tree;
	GTreeItem *Parent;
	List<GTreeItem> Items;

	TV_ITEM Item;
	HTREEITEM hTreeItem;

	bool Attach(GTreeItem *After);
	bool AttachAll();

public:
	GTreeItem();
	virtual ~GTreeItem();

	char *GetText(int i=0) { return Str; }
	bool SetText(char *s, int i=0);
	HTREEITEM ItemHandle() { return hTreeItem; }
	GTreeItem *Insert(GTreeItem *Obj = NULL, int Pos = -1);
	void Update();
	bool Selected();
	void Selected(bool b);
	bool Expanded();
	void Expanded(bool b);

	GTreeItem *GetChild();
	GTreeItem *GetNext();
};

class GTree : public GControl, public GTreeItem, public GItemContainer {

	static GWin32Class WndClass;
	friend class GTreeItem;

protected:
	bool Lines;
	bool Buttons;
	bool LinesAtRoot;
	bool EditLabels;

	char *Class() { return LGI_TREE; }
	DWORD Style() { return	WS_CHILD |
				WS_TABSTOP |
				((Lines) ? TVS_HASLINES : 0) |
				((Buttons) ? TVS_HASBUTTONS : 0) |
				((LinesAtRoot) ? TVS_LINESATROOT : 0) |
				((EditLabels) ? TVS_EDITLABELS : 0); }

public:
	GTree(int id, int x, int y, int cx, int cy, char *name);
	~GTree();

	// Events
	virtual void OnItemClick(GTreeItem *Item, VMouse &m);
	virtual void OnItemBeginDrag(GTreeItem *Item, int Flags);
	virtual void OnItemExpand(GTreeItem *Item, bool Expand) {}
	virtual void OnItemSelect(GTreeItem *Item) {}
	
	// Procedures
	void OnMouseClick(VMouse &m);
	int OnEvent(GMessage *Msg);
	GTreeItem *Insert(GTreeItem *Obj = 0, int Pos = -1);
	bool Delete(GTreeItem *Obj);
	bool Select(GTreeItem *Obj);
	bool Remove(GTreeItem *Obj);

	// Container
	void Empty();		// delete all items
	void RemoveAll();	// remove reference to externally owned items
};

// Progress control
class GProgress : public GControl, public Progress {

	static GWin32Class WndClass;
	char *Class() { return LGI_PROGRESS; }
	DWORD Style() { return WS_CHILD | WS_TABSTOP; }

public:
	GProgress(int id, int x, int y, int cx, int cy, char *name);
	~GProgress();

	void SetLimits(int l, int h);
	void Value(int v);
	int OnEvent(GMessage *Msg);
};

#endif