/*hdr
**	FILE:		gui.h
**	AUTHOR:		Matthew Allen
**	DATE:		19/12/97
**	DESCRIPTION:	Graphical user interface
**
**
**	Copyright (C) 1997-8, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

/////////////////////////////////////////////////////////////////////////////////////
// Includes
#ifndef __GUI_H
#define __GUI_H

#include <string.h>

/////////////////////////////////////////////////////////////////////////////////////
// Class names
class VKey;
class VMouse;
class GObject;
class GTarget;
class GComponent;
class GEvent;
class GId;
class GApp;
class GAppWindow;
class GWin32Class;
class GWindow;
class GView;
class GFileType;
class GFileSelect;
class GFindReplace;
class GSubMenu;
class GMenuItem;
class GMenu;
class GToolBar;
class GToolButton;
class GSplitter;
class GStatusPane;
class GStatusBar;
class GToolColour;
class GScrollBar;
class GControl;
class GImageList;
class Progress;

/////////////////////////////////////////////////////////////////////////////////////
// Defines
#define hAppInst()			GApp::Instance()
#define SysFont				(GApp::ObjInstance()->SystemFont)
#define ExitApp()			(GApp::ObjInstance()->Exit())
#define Message(m)			(m->Msg)
#define CreateMsg(m, a, b)	GMessage(m, a, b)

#define SetFlag(i, f)		(i) |= (f)
#define ClearFlag(i, f)		(i) &= ~(f)
#define TestFlag(i, f)		((i) & (f))

#define K_CHAR				0x0

#define MK_LEFT				MK_LBUTTON
#define MK_RIGHT			MK_RBUTTON
#define MK_MIDDLE			MK_MBUTTON
#define MK_CTRL				MK_CONTROL

// Edge types
#define SUNKEN				1
#define RAISED				2
#define CHISEL				3
#define FLAT				4

// Colours
#define TOOL_LOW			GetSysColor(COLOR_3DSHADOW)
#define TOOL_MED			GetSysColor(COLOR_3DFACE)
#define TOOL_HIGH			GetSysColor(COLOR_3DHILIGHT)
#define C_WORKSPACE			GetSysColor(COLOR_3DHILIGHT)

// Menus
#define PostMenuCmd(hWnd, Cmd)		PostMessage(hWnd, WM_COMMAND, MAKELONG(Cmd, 0), (int) hWnd)

// Messages
// wParam = bool Inside; // is the mouse inside the client area?
// lParam = MAKELONG(x, y); // mouse location
#define WM_MOUSEENTER		(WM_USER+0x100)
#define WM_MOUSEEXIT		(WM_USER+0x101)

// wParam = (GWindow*) Wnd;
// lParam = (int) Flags;
#define WM_CHANGE			(WM_USER+0x102)

// wParam = (GWindow*) Wnd;
// lParam = (char*) Text; // description from window
#define WM_DESCRIBE			(WM_USER+0x103)

// return (bool)
#define WM_WANT_DIALOG_PROC	(WM_USER+0x104)

/////////////////////////////////////////////////////////////////////////////////////
// Externs
extern char *NewStr(char *s);

extern void DrawBox(GdcBasePrimitives *pDC, GdcRegion &r, bool Sunken, bool Fill);
extern void DrawMoveDeepEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type);
extern void DrawMoveShallowEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type);
extern void DrawMoveFill(GdcBasePrimitives *pDC, GdcRegion &r, int Width = -1);

extern WORD LgiGetLanguageId();
extern char *LgiLoadString(WORD Res);
extern int LgiMsg(GWindow *Parent, char *Msg, char *Title, int Type, ...);
extern bool LgiGetCursorPos(GdcPt2 *Pt);
extern bool LgiExecute(char *File, char *Arguments="", char *Dir=".");

/////////////////////////////////////////////////////////////////////////////////////
// Typedefs
typedef HFONT GFont;

/////////////////////////////////////////////////////////////////////////////////////
// Classes
class GMessage {
public:
	int Msg;
	int a;
	int b;

	GMessage()
	{
		Msg = a = b = 0;
	}

	GMessage(int M, int A = 0, int B = 0)
	{
		Msg = M;
		a = A;
		b = B;
	}
};

class VKey {
public:
	uchar c;
	int Flags;
	int Data;
	bool Down;
	bool IsChar;

	bool Ctrl() { return (Flags & MK_CONTROL) != 0; }
	bool Alt() { return (Flags & MK_ALT) != 0; }
	bool Shift() { return (Flags & MK_SHIFT) != 0; }
};

class VMouse {
public:
	int x;
	int y;
	int Flags;
	bool Down;
	bool Double;

	bool Left() { return (Flags & MK_LEFT) != 0; }
	bool Middle() { return (Flags & MK_MIDDLE) != 0; }
	bool Right() { return (Flags & MK_RIGHT) != 0; }
	bool Ctrl() { return (Flags & MK_CONTROL) != 0; }
	bool Alt() { return (Flags & MK_ALT) != 0; }
	bool Shift() { return (Flags & MK_SHIFT) != 0; }
};

class Progress {
protected:
	char *Description;
	int Start;
	int Val;
	int Low, High;
	char *Type;
	double Scale;

public:
	Progress();
	Progress(char *desc, int l, int h, char *type = NULL, double scale = 1.0);
	virtual ~Progress();

	virtual char *GetDescription() { return Description; }
	virtual void SetDescription(char *d);
	virtual void GetLimits(int *l, int *h);
	virtual void SetLimits(int l, int h);
	virtual int Value() { return Val; }
	virtual void Value(int v) { Val = v; }
	virtual double GetScale() { return Scale; }
	virtual void SetScale(double s) { Scale = s; }
	virtual char *GetType() { return Type; }
	virtual void SetType(char *t) { Type = t; }

	Progress &operator =(Progress &p);
};

class GObject {

	char *ObjName;

public:
	GObject()
	{
		ObjName = 0;
	}
	
	virtual ~GObject()
	{
		DeleteArray(ObjName);
	}

	virtual char *Name() { return ObjName; }
	virtual bool Name(char *n)
	{
		if (n == ObjName) return TRUE;
		DeleteObj(ObjName);
		if (n)
		{
			ObjName = NEW(char[strlen(n)+1]);
			if (ObjName)
			{
				strcpy(ObjName, n);
				return TRUE;
			}
		}
		return FALSE;
	}

	virtual int Sizeof() { return 0; }
	virtual bool Serialize(VFile &f, bool Write) { return FALSE; }
};

VFile &operator >>(VFile &f, GObject &o);
VFile &operator <<(VFile &f, GObject &o);

class GTarget {
public:
	virtual int OnEvent(GMessage *Msg) { return 0; }
};

class GComponent {
public:
	// Status
	virtual bool Enabled() { return FALSE; }
	virtual void Enabled(bool e) {}
	virtual bool Visible() { return FALSE; }
	virtual void Visible(bool v) {}
	virtual bool Focus() { return FALSE; }
	virtual void Focus(bool f) {}
};

class GEvent {
public:
	// Mouse related
	virtual void OnMouseClick(VMouse &m) {}
	virtual void OnMouseEnter(VMouse &m) {}
	virtual void OnMouseExit(VMouse &m) {}
	virtual void OnMouseMove(VMouse &m) {}

	// Keyboard related
	virtual void OnKey(VKey &k) {}

	// Other
	virtual void OnPaint(GdcBasePrimitives *pDC) {}
};

class GId {
protected:
	int Id;

public:
	GId() { Id = 0; }

	virtual int GetId() { return Id; }
	virtual void SetId(int i) { Id = i; }
	virtual int Value() { return 0; }
	virtual void Value(int i) {}
};

/////////////////////////////////////////////////////////////////////////////////
class GApp {

	static HINSTANCE hAppInst;
	static char *CmdLine;
	static GApp *TheApp;
	static int CmdShow;

public:
	static HWND ActiveDialog;
	GdcWin32Ttf *SystemFont;
	GAppWindow *AppWnd;

	GApp(HINSTANCE hInst, char *lpCmdLine, int nShow);
	virtual ~GApp();

	static HINSTANCE Instance() { return hAppInst; }
	static GApp *ObjInstance() { return TheApp; }
	static int Show() { return CmdShow; }

	virtual void OnCmdLine(char *Cmd);
	virtual void OnArgs(int Args, char **Arg);
	virtual void Run(bool Loop = TRUE);
	virtual void Exit(int Code = 0);
};

class GWin32Class : public GObject {

	static LRESULT CALLBACK Redir(HWND hWnd, UINT m, WPARAM a, LPARAM b);
	static LRESULT CALLBACK SubClassRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b);
	WNDPROC ParentProc;

public:
	WNDCLASS Class;

	GWin32Class();
	~GWin32Class();

	bool Create(char *ClassName);
	bool SubClass(char *ClassName, char *Parent);
	LRESULT CALLBACK CallParent(HWND hWnd, UINT m, WPARAM a, LPARAM b);
};

extern GWindow *GWindowFromHandle(HWND *hWnd);
typedef List<GWindow> WindowList;

#define GWF_ENABLE			0x0001
#define GWF_VISIBLE			0x0002
#define GWF_FOCUS			0x0004
#define GWF_OVER			0x0008
#define GWF_DROP_TARGET			0x0010
#define GWF_DIALOG_PROCESSING		0x0020

class GWindow : public GObject, public GTarget, public GComponent, public GEvent {

	friend		class GAppWindow;
	friend		class GView;
	friend		class GDialog;
	friend		void DebugCapture(GWindow *p);
	friend LRESULT CALLBACK GWin32Class::Redir(HWND hWnd, UINT m, WPARAM a, LPARAM b);
	friend LRESULT CALLBACK GWin32Class::SubClassRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b);
	friend LRESULT CALLBACK DlgRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b);
	friend		void DebugCapture(GWindow *p);

	static void CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime);

protected:
	// OS specific
	static GWindow	*Capturing;
	static GWindow	*Over;
	static HWND	hPrevCapture;
	HWND		hWnd;

	// General
	int		WndFlags;
	GdcRegion	Pos;
	GWindow		*Parent;
	GWindow		*Notify;

	virtual GWindow *FindReal();
	virtual bool HandleCapture(GWindow *Wnd, bool c);

public:
	WindowList	Children;

	GWindow();
	virtual ~GWindow();

	// Win32 API
	virtual HWND Handle() { return hWnd; }
	operator HWND() { return hWnd; }

	// Create a Win32 window
	virtual bool Create(char *Class, int Style, int x, int y, int Sx, int Sy, HWND hParent);
	virtual bool IsDropTarget() { return TestFlag(WndFlags, GWF_DROP_TARGET); }
	virtual void IsDropTarget(bool t) { if (t) SetFlag(WndFlags, GWF_DROP_TARGET); else ClearFlag(WndFlags, GWF_DROP_TARGET); }

	// Create a virtual window
	bool Create(char *name, GdcRegion &Pos);

	// Target
	int OnEvent(GMessage *Msg);

	// Flags
	bool Enabled();
	void Enabled(bool e);
	bool Visible();
	void Visible(bool v);
	bool Focus();
	void Focus(bool f);

	// Events
	// void OnMouseClick(VMouse &m);
	// void OnMouseEnter(VMouse &m);
	// void OnMouseExit(VMouse &m);
	// void OnMouseMove(VMouse &m);
	// void OnKey(VKey &k);
	// void OnPaint(GdcBasePrimitives *pDC);
	virtual void OnFocus(bool f) {}
	virtual void OnPulse() {}
	virtual void OnCreate() {}
	virtual bool OnRequestClose() { return TRUE; }
	virtual void OnPosChange() {}
	virtual int OnNotify(GControl *Ctrl, int Flags) { return 0; }

	// Font
	virtual HFONT GetFont();
	virtual void SetFont(HFONT hFont);

	// Other
	bool Name(char *n);
	char *Name();

	virtual bool Invalidate(GdcRegion *r = NULL, bool Repaint = FALSE);
	virtual bool InvalidateFrame(GdcRegion *r = NULL, bool Repaint = FALSE);
	virtual bool SetPos(GdcRegion &p, bool Repaint = FALSE);
	virtual bool Capture(bool c);
	virtual bool IsOver(VMouse &m);
	virtual GWindow *WindowFromPoint(int x, int y);
	virtual void SetPulse(int Length = -1);

	// Stubs
	virtual GWindow *GetParent() { return Parent; }
	virtual void SetParent(GWindow *p) { Parent = p; }
	virtual bool Attach(GWindow *p) { return FALSE; }
	virtual GControl *IsControl() { return 0; }
	virtual GWindow *GetNotify() { return Notify; }
	virtual void SetNotify(GWindow *n) { Notify = n; }
	virtual int X() { return Pos.X(); }
	virtual int Y() { return Pos.Y(); }
	virtual GdcRegion &GetPos() { return Pos; }
	
};

// Dialog stuff
#include "GuiDlg.h"

///////////////////////////////////////////////////////////////////////////////////////////////
#define GV_EDGE_TOP			0x0001
#define GV_EDGE_RIGHT			0x0002
#define GV_EDGE_BOTTOM			0x0004
#define GV_EDGE_LEFT			0x0008

class GView : public GControlGroup {
protected:
	static GWin32Class	Class;
	int			ExtraFlags;
	GScrollBar		*VScroll;
	GScrollBar		*HScroll;

	GdcRegion *FindLargest(GdcComplexRegion &r);
	GdcRegion *FindSmallestFit(GdcComplexRegion &r, int Sx, int Sy);
	GdcRegion *FindLargestEdge(GdcComplexRegion &r, int Edge);
	bool SetScrollBars(bool x, bool y);

public:
	GView();
	~GView();

	virtual void GetScrollPos(int &x, int &y);
	virtual void SetScrollPos(int x, int y);
	int OnEvent(GMessage *Msg);

	virtual bool Attach(GWindow *Wnd);
	virtual bool Detach();
	virtual bool Pour(GdcComplexRegion &r) { return FALSE; }
};

///////////////////////////////////////////////////////////////////////////////////////////////
// Menu wrappers
typedef List<GMenuItem> MenuItemList;

class GSubMenu : public GObject, public GTarget, public GComponent {

	friend class GMenuItem;
	friend class GMenu;

	bool LoadFromMenu(HMENU hmenu);

protected:
	GMenu		*Menu;
	GMenuItem	*Parent;

	HMENU		hMenu;
	MenuItemList	List;

public:
	GSubMenu(bool Popup = TRUE);
	~GSubMenu();

	HMENU Handle() { return hMenu; }
	GMenuItem *AppendItem(char *Str, int Id, bool Enabled);
	GMenuItem *AppendSeparator();
	GSubMenu *AppendSub(char *Str);

	int Float(GWindow *Parent, int x, int y, bool Left = TRUE);
	GMenuItem *FindItem(int i);
};

class GMenuItem : public GObject, public GTarget, public GComponent {

	friend class GSubMenu;
	friend class GMenu;

private:
	bool Insert(int Pos);
	bool Update();

protected:
	GMenu		*Menu;
	GSubMenu	*Parent;
	GSubMenu	*Child;

	int		Position;
	MENUITEMINFO	Info;

public:
	GMenuItem();
	GMenuItem(GMenu *m, GSubMenu *p, int Pos);
	~GMenuItem();

	GSubMenu *Create();
	
	void Id(int i);
	void Separator(bool s);
	void Checked(bool c);
	bool Name(char *n);
	void Enabled(bool e);
	void Focus(bool f);
	void Sub(GSubMenu *s);

	int Id() { return Info.wID; }
	bool Separator() { return (Info.fType & MFT_SEPARATOR) != 0; }
	bool Checked() { return (Info.fState & MF_CHECKED) != 0; }
	bool Enabled() { return (Info.fState & MFS_ENABLED) != 0; }
	bool Visible() { return TRUE; }
	bool Focus() { return (Info.fState & MFS_HILITE) != 0; }
	GSubMenu *Sub() { return Child; }
};

class GMenu : public GSubMenu {

	friend class GSubMenu;
	friend class GMenuItem;

protected:
	GWindow *Window;

public:
	GMenu();
	~GMenu();

	bool Attach(GWindow *p);
	bool Detach();
	bool Load(char *Res);
};

///////////////////////////////////////////////////////////////////////////////////////////////////
typedef List<GView> ViewList;

class GAppWindow : public GControlGroup {
protected:
	GMenu		*Menu;
	ViewList	Views;

public:
	GAppWindow();
	~GAppWindow();

	// Methods
	void Pour();

	// Properties
	bool Visible();
	void Visible(bool v);
	GMenu *GetMenu() { return Menu; }

	// Events
	int OnEvent(GMessage *Msg);
	virtual int OnCommand(int Cmd, int Event, HWND hWnd) { return 0; }
	virtual void OnReceiveFiles(List<char> &Files) {}
};

///////////////////////////////////////////////////////////////////////////////////////////////////
// File select dialog
class GFileType : public GObject {

	char *Ext;

public:
	GFileType()
	{
		Ext = 0;
	}

	~GFileType()
	{
		DeleteObj(Ext);
	}

	char *Extension() { return Ext; }
	bool Extension(char *e)
	{
		return (Ext = NewStr(e)) != 0;
	}

	char *Description() { return Name(); }
	bool Description(char *d) { return Name(d); }
};

typedef List<GFileType> FileTypeList;

class GFileSelect : public GObject {

	int		Size;
	char		*FileStr;

	OPENFILENAME	Info;
	FileTypeList	TypeList;
	GWindow		*ParentWnd;

	char		*TypeStr();

public:
	GFileSelect();
	~GFileSelect();

	char *Name();
	bool Name(char *n);
	char *operator [](int i);
	int Names();
	char *File();
	char *Extension();

	int Types();
	GFileType *Type(int n);
	bool Type(char *Description, char *Extension);

	GWindow *Parent();
	void Parent(GWindow *Window);

	bool MultiSelect();
	void MultiSelect(bool Multi);

	char *InitialDir();
	void InitialDir(char *InitDir);

	char *Title();
	void Title(char *Title);

	char *DefaultExtension();
	void DefaultExtension(char *DefExt);

	bool Open();
	bool Save();
};

class GFindReplace {

	HWND hDlg;
	FINDREPLACE FindInfo;
	int FindMsgID;
	bool FindOp;

public:
	char Find[256];
	char Replace[256];

	GFindReplace(HWND hParent = NULL);
	virtual ~GFindReplace();
	virtual HWND Handle() { return hDlg; }

	void SetParent(HWND hParent);

	bool OpenFind();
	bool OpenReplace();
	int OnEvent(GMessage *Msg);
	virtual bool OnFind();
	virtual bool OnReplace();

	bool CaseSensitive();
	void CaseSensitive(bool i);
	bool WholeWord();
	void WholeWord(bool i);
	bool BtnFindNext();
	bool BtnReplace();
	bool BtnReplaceAll();
};

/////////////////////////////////////////////////////////////////////////////////////////////
#define IMGLST_SELECTED				0x0001
#define IMGLST_DISABLED				0x0002
#define IMGLST_GDC				0x0004

class GImageList : public GdcDibSection {

	HIMAGELIST hImgList;
	int Sx, Sy;
	GdcBasePrimitives *pCreate;

public:
	GImageList(int x, int y, GdcBasePrimitives *pDC = NULL);
	~GImageList();

	operator const HIMAGELIST () { return hImgList; }
	HIMAGELIST Handle() { return hImgList; }

	int TileX() { return Sx; }
	int TileY() { return Sy; }

	bool Create(int x, int y, int Bits, int LineLen = 0, bool KeepData = FALSE);
	void Update(int Flags);
	void Draw(GdcBasePrimitives *pDest, int x, int y, int Image, int Flags);
};

///////////////////////////////////////////////////////////////////////////////////////////////
// Toolbar stuff

// Tool button types
#define TBT_PUSH			0
#define TBT_RADIO			1
#define TBT_TOGGLE			2

// Default tool button sizes
#define TBS_X				19
#define TBS_Y				18

// Commands
#define IDM_NONE			0
#define IDM_SEPARATOR			-1
#define IDM_BREAK			-2

// Dimensions
#define BORDER_SHADE			1
#define BORDER_SPACER			1
#define BORDER_SEPARATOR		4
#define BORDER_BUTTON			1

class GToolButton : public GControl {

	friend class GToolBar;

protected:
	int		Type;		// Button Type

	bool		Clicked;
	bool		Down;
	int		ImgIndex;
	
	virtual	void SetDown(bool d) { Down = d; }

public:
	GToolButton(int Bx, int By);
	~GToolButton();

	int Value() { return Down; }
	void Value(int i);
	int Image() { return ImgIndex; }
	void Image(int i) { ImgIndex = i; }

	void OnPaint(GdcBasePrimitives *pDC);

	void OnMouseClick(VMouse &m);
	void OnMouseEnter(VMouse &m);
	void OnMouseExit(VMouse &m);

	virtual void OnCommand(VMouse &m);
	virtual void SendNotify(int Flags);
	virtual bool GetDimension(int &x, int &y) { return FALSE; }
};

class GToolBar : public GView {

	friend class GToolButton;

protected:
	int Bx, By;
	int Sx, Sy;
	bool Vertical;
	int LastIndex;
	GImageList *ImgList;

	virtual int PostCommand(int Cmd);
	virtual int PostDescription(GWindow *Ctrl, char *Text);

public:
	GToolBar();
	~GToolBar();

	int X();
	int Y();

	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);

	void OnMouseClick(VMouse &m);
	void OnMouseEnter(VMouse &m);
	void OnMouseExit(VMouse &m);
	void OnMouseMove(VMouse &m);

	bool IsVertical() { return Vertical; }
	void IsVertical(bool v) { Vertical = v; }
	bool SetBitmap(char *File, int Bx, int By);
	bool SetDC(GdcBasePrimitives *pDC, int Bx, int By);

	GToolButton *AppendButton(char *Tip, int Id, int Type = TBT_PUSH, int Enabled = TRUE, int IconId = -1);
	bool AppendControl(GControl *Ctrl);
	bool AppendSeparator();
	bool AppendBreak();

	void Empty();
	bool Pour(GdcComplexRegion &r);
};

////////////////////////////////////////////////////////////////////////////////////////////////
class GSplitter : public GView {

	bool Vertical;
	bool Moving;
	int SplitPos;
	GdcRegion Bar;
	int Offset;

	GdcRegion PosA;
	GdcRegion PosB;
	GWindow *ViewA;
	GWindow *ViewB;

	void CalcRegions(GdcRegion Rect);

public:
	GSplitter();
	~GSplitter();

	bool IsVertical() { return Vertical; }
	void IsVertical(bool v) { Vertical = v; }
	int Split() { return SplitPos; }
	GWindow *GetViewA() { return ViewA; }
	GWindow *GetViewB() { return ViewB; }
	void SetViewA(GWindow *a);
	void SetViewB(GWindow *b);
	void Split(int s);
	GControl *GetControl(HWND hCtrl);

	bool Pour(GdcComplexRegion &r);
	void OnPaint(GdcBasePrimitives *pDC);

	void OnMouseClick(VMouse &m);
	void OnMouseMove(VMouse &m);
};

////////////////////////////////////////////////////////////////////////////////////////////////
#define STATUSBAR_SEPARATOR			4
#define GSP_SUNKEN				0x0001

typedef GStatusPane *GStatusPanePtr;

class GStatusBar : public GView, private PtrArray {

	friend class GStatusPane;

protected:
	GdcWin32Ttf *Font;
	void RePour();

public:
	GStatusBar();
	~GStatusBar();

	bool Pour(GdcComplexRegion &r);
	void OnPaint(GdcBasePrimitives *pDC);

	GStatusPane *AppendPane(char *Text, int Width);
	bool AppendPane(GStatusPane *Pane);
};

class GStatusPane : public GWindow {

	friend class GStatusBar;

protected:
	int		Flags;
	int		Width;
	GdcBasePrimitives *pDC;

public:
	GStatusPane();
	~GStatusPane();

	char *Name() { return GObject::Name(); }
	bool Name(char *n);
	void OnPaint(GdcBasePrimitives *pDC);

	int X();
	void X(int x);
	bool Sunken();
	void Sunken(bool s);
	GdcBasePrimitives *Bitmap();
	void Bitmap(GdcBasePrimitives *pdc);
};

////////////////////////////////////////////////////////////////////////////////////////////////
class GScrollBar : public GTarget, public GComponent {

	int		Type;		// vertical, horzontal or control
	GWindow		*Parent;
	SCROLLINFO	Info;

	void Update();
	void SetParentFlag(bool Bool);

public:
	GScrollBar();
	~GScrollBar();

	// data
	bool Vertical();
	void SetVertical(bool v);
	void SetParent(GWindow *p);
	int Pos();
	void SetPos(int p);
	void Limits(int &Low, int &High);
	void SetLimits(int Low, int High);
	int Page();
	void SetPage(int p);
	void Invalidate();
	
	// events
	int OnEvent(GMessage *Msg);

	// overridables
	virtual void OnChange(int Pos) {}
};

/////////////////////////////////////////////////////////////////////////////////////////////
class GCommand : public GObject, public GComponent {

	int		Flags;

public:
	int		Id;
	GToolButton	*ToolButton;
	GMenuItem	*MenuItem;
	VKey		*Accelerator;
	char		*TipHelp;
	
	GCommand();
	~GCommand();

	bool Enabled();
	void Enabled(bool e);
	bool Value();
	void Value(bool v);
};

/////////////////////////////////////////////////////////////////////////////////////////////
class GClipBoard {

	GWindow *Owner;
	bool Open;

public:
	GClipBoard(GWindow *o);
	~GClipBoard();

	bool Empty();

	// Text
	bool Text(char *Str);
	char *Text();

	// Bitmap
	bool Bitmap(GdcBasePrimitives *pDC);
	GdcBasePrimitives *Bitmap();
	GdcBasePrimitives *ConvertFromPtr(void *Ptr);
};

/////////////////////////////////////////////////////////////////////////////////////////
class ProgressList : public List<Progress> {

	bool InUse;

public:
	ProgressList();

	bool Lock();
	void Unlock();
};

class GProgressPane : public Progress, public GView {

	friend class GProgressDlg;

	Progress *Ref;

protected:
	bool Wait;
	bool Canceled;

	GText *Desc;
	GText *ValText;
	GText *Rate;
	GProgress *Bar;
	GButton *But;

public:
	GProgressPane();
	~GProgressPane();

	char *GetDescription();
	void SetDescription(char *d);
	void GetLimits(int *l, int *h) { Progress::GetLimits(l, h); }
	void SetLimits(int l, int h);
	int Value() { return Progress::Value(); }
	void Value(int v);
	bool IsCanceled() { return Canceled; }

	void OnCreate();
	int OnNotify(GControl *Ctrl, int Flags);
	void OnPaint(GdcBasePrimitives *pDC);
	bool Pour(GdcComplexRegion &r);
	HFONT GetFont();
};

class GProgressDlg : public GDialog {
protected:
	bool Wait;
	List<GProgressPane> Progri;

public:
	GProgressDlg(GWindow *parent = 0, bool wait = FALSE);
	~GProgressDlg();

	GProgressPane *Push();
	void Pop(GProgressPane *p = 0);
	char *GetDescription();
	void SetDescription(char *d);
	void GetLimits(int *l, int *h);
	void SetLimits(int l, int h);
	int Value();
	void Value(int v);
	bool IsCanceled();
	char *GetType();
	void SetType(char *t);

	void OnSync(ProgressList *Prg);

	int OnEvent(GMessage *Msg);
	void OnPaint(GdcBasePrimitives *pDC);
	void OnCreate();
};

class GThread {

	uint ThreadId;
	HANDLE hThread;

public:
	GThread(bool Sleep = FALSE, int StackSize = 8 << 10);
	virtual ~GThread();

	// Properties
	HANDLE Handle() { return hThread; }
	DWORD ExitCode();
	bool IsExited();
	void Run(bool b);
	void Terminate();
	
	int Priority();
	void Priority(int p);

	// override to do something
	virtual int Main();
};

///////////////////////////////////////////////////////////////////////////
class GTrayIcon : public GObject, public GComponent {
protected:
	static int Id;

	int Val;		// which icon visible
	HICON *hIcon;		// array of loaded icons
	int Icons;		// size of array
	int MyId;
	int Msg;
	GWindow *Notify;

public:
	NOTIFYICONDATA	TrayIcon;

	GTrayIcon(int Icons = 1);
	~GTrayIcon();

	bool LoadIcon(char *Str, int Index = 0);
	bool Visible();
	void Visible(bool v);

	int Value();
	void Value(int v);
	int NotifyMsg() { return Msg; }
	void NotifyMsg(int i) { Msg = i; }
	GWindow *GetNotify() { return Notify; }
	void SetNotify(GWindow *n) { Notify = n; }
};

#endif