/*hdr
**      FILE:           Gui.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**              fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"

#include "gdc2.h"
#include "gui.h"

bool Foo = FALSE;

////////////////////////////////////////////////////////////////////////////////////////////////////
char *NewStr(char *s)
{
	if (s)
	{
		char *c = NEW(char[strlen(s)+1]);
		if (c)
		{
			strcpy(c, s);
			return c;
		}
	}

	return 0;	
}

VMouse AdjustMouseInfoForWindow(VMouse &Info, GWindow *Wnd)
{
	VMouse Temp = Info;
	if (GWindow)
	{
		if (NOT Wnd->Handle())
		{
			GdcRegion r = Wnd->GetPos();
			Temp.x -= r.x1;
			Temp.y -= r.y1;
		}
	}
	return Temp;
}

void DrawRect(HWND hwnd)
{
	HDC hDC = GetDC(hwnd);
	if (hDC)
	{
		Foo = NOT Foo;
		HPEN hPen = CreatePen(PS_SOLID, 1, (Foo) ? Rgb24(255, 0, 0) : Rgb24(0, 0, 0));
		HBRUSH hBrush = GetStockObject(NULL_BRUSH);
		hPen = SelectObject(hDC, hPen);
		hBrush = SelectObject(hDC, hBrush);
		
		RECT rc;
		GetWindowRect(hwnd, &rc);
		Rectangle(hDC, 0, 0, rc.right-rc.left-1, rc.bottom-rc.top-1);
		
		hBrush = SelectObject(hDC, hBrush);
		hPen = SelectObject(hDC, hPen);
		DeleteObject(hPen);
		DeleteObject(hBrush);

		ReleaseDC(hwnd, hDC);
	}
}

int LgiMsg(GWindow *Parent, char *Msg, char *Title, int Type, ...)
{
	int Result = 0;
	if (Msg)
	{
		char Buffer[1025];

		va_list Arg;
		va_start(Arg, Type);
		vsprintf(Buffer, Msg, Arg);
		va_end(Arg);

		Result = MessageBox((Parent)?Parent->Handle():0, Buffer, (Title)?Title:"Message", Type);
	}

	return Result;
}

bool LgiGetCursorPos(GdcPt2 *Pt)
{
	POINT p;
	if (Pt AND GetCursorPos(&p))
	{
		Pt->x = p.x;
		Pt->y = p.y;
		return TRUE;
	}
	return FALSE;
}

bool LgiExecute(char *File, char *Arguments, char *Dir)
{
	return ((int)ShellExecute(NULL, "open", File, Arguments, Dir, 5)) > 32;
}

/////////////////////////////////////////////////////////////////////////////////////////////
void DrawBox(GdcBasePrimitives *pDC, GdcRegion &r, bool Sunken, bool Fill)
{
	if (Fill)
	{
		pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
		pDC->Rectangle(r.x1+1, r.y1+1, r.x2-1, r.y2-1);
	}

	pDC->Colour(CBit(pDC->GetBits(), (Sunken) ? TOOL_HIGH : TOOL_LOW));
	pDC->Line(r.x2, r.y2, r.x2, r.y1);
	pDC->Line(r.x2, r.y2, r.x1, r.y2);

	pDC->Colour(CBit(pDC->GetBits(), (Sunken) ? TOOL_LOW : TOOL_HIGH));
	pDC->Line(r.x1, r.y1, r.x1, r.y2);
	pDC->Line(r.x1, r.y1, r.x2, r.y1);
}

void DrawMoveDeepEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	if (NOT pDC) return;
	COLOUR Old = pDC->Colour();
	COLOUR VLow = CBit(pDC->GetBits(), GetSysColor(COLOR_3DDKSHADOW));
	COLOUR Low = CBit(pDC->GetBits(), GetSysColor(COLOR_3DSHADOW));
	COLOUR High = CBit(pDC->GetBits(), GetSysColor(COLOR_3DLIGHT));
	COLOUR VHigh = CBit(pDC->GetBits(), GetSysColor(COLOR_3DHILIGHT));

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(Low);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(VLow);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(High);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VHigh);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case RAISED:
		{
			pDC->Colour(VHigh);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High);
			pDC->Line(r.x1+1, r.y1+1, r.x2-1, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-1);

			pDC->Colour(Low);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(VLow);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);
			break;
		}
		case CHISEL:
		{
			pDC->Colour(Low);
			pDC->Line(r.x1, r.y1, r.x2-1, r.y1);
			pDC->Line(r.x1, r.y1, r.x1, r.y2-1);

			pDC->Colour(High);
			pDC->Line(r.x1+1, r.y1+1, r.x2-2, r.y1+1);
			pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2-2);

			pDC->Colour(Low);
			pDC->Line(r.x2-1, r.y2-1, r.x2-1, r.y1+1);
			pDC->Line(r.x2-1, r.y2-1, r.x1+1, r.y2-1);

			pDC->Colour(High);
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y1, r.x1, r.y2);
			break;
		}
	}

	r.Size(2, 2);
	pDC->Colour(Old);
}

void DrawMoveShallowEdge(GdcBasePrimitives *pDC, GdcRegion &r, int Type)
{
	if (NOT pDC) return;
	COLOUR Old = pDC->Colour();

	switch (Type)
	{
		case SUNKEN:
		{
			pDC->Colour(CBit(pDC->GetBits(), TOOL_HIGH));
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(CBit(pDC->GetBits(), TOOL_LOW));
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
		case RAISED:
		{
			pDC->Colour(CBit(pDC->GetBits(), TOOL_LOW));
			pDC->Line(r.x2, r.y2, r.x2, r.y1);
			pDC->Line(r.x2, r.y2, r.x1, r.y2);

			pDC->Colour(CBit(pDC->GetBits(), TOOL_HIGH));
			pDC->Line(r.x1, r.y1, r.x1, r.y2);
			pDC->Line(r.x1, r.y1, r.x2, r.y1);

			r.Size(1, 1);
			break;
		}
	}

	pDC->Colour(Old);
}

void DrawMoveFill(GdcBasePrimitives *pDC, GdcRegion &r, int Width)
{
	pDC->Colour(CBit(pDC->GetBits(), TOOL_MED));
	if (Width < 1 OR r.X() < (2 * Width) OR r.Y() < (2 * Width))
	{
		pDC->Rectangle(&r);
		r.ZOff(-1, -1);
	}
	else
	{
		pDC->Rectangle(r.x1, r.y1, r.x2, r.y1+Width-1);
		pDC->Rectangle(r.x1, r.y2-Width+1, r.x2, r.y2);
		pDC->Rectangle(r.x1, r.y1+Width, r.x1+Width-1, r.y2-Width);
		pDC->Rectangle(r.x2-Width+1, r.y1+Width, r.x2, r.y2-Width);
		r.Size(Width, Width);
	}
}

WORD LgiGetLanguageId()
{
	return MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT);
	// return MAKELANGID(LANG_FRENCH, SUBLANG_FRENCH);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GWin32Class GView::Class;

////////////////////////////////////////////////////////////////////////////////////////////////////
HHOOK	hMouseHook = 0;
HWND	hMouseOver = 0;

LRESULT CALLBACK MouseHook(int nCode, WPARAM wParam, LPARAM lParam)
{
	// check whether the mouse has moved off the last window
	MOUSEHOOKSTRUCT *Ms = (MOUSEHOOKSTRUCT*) lParam;
	if (Ms)
	{
		RECT rc;
		GetClientRect(Ms->hwnd, &rc);
		ScreenToClient(Ms->hwnd, &Ms->pt);

		// is the mouse inside the client area?
		bool Inside = NOT      (Ms->pt.x < 0 OR
					Ms->pt.y < 0 OR
					Ms->pt.x >= rc.right-rc.left OR
					Ms->pt.y >= rc.bottom-rc.top);

		HWND hWnd = (Inside) ? Ms->hwnd : 0;

		if (hWnd != hMouseOver)
		{
			// window has changed
			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is LOSING mouse
				SendMessage(hMouseOver, WM_MOUSEEXIT, 0, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}

			// set current window
			hMouseOver = hWnd;

			if (hMouseOver AND IsWindow(hMouseOver))
			{
				// window is GETTING mouse
				SendMessage(hMouseOver, WM_MOUSEENTER, Inside, MAKELONG((short) Ms->pt.x, (short) Ms->pt.y));
			}
		}
	}

	return CallNextHookEx(hMouseHook, nCode, wParam, lParam);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK GWin32Class::Redir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		
		GWindow *NewWnd = (GWindow*) Info->lpCreateParams;
		NewWnd->hWnd = hWnd;
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GWindow *Wnd = (GWindow*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	return DefWindowProc(hWnd, m, a, b);
}

LRESULT CALLBACK GWin32Class::SubClassRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		GWindow *NewWnd = 0;
		if (Info->lpCreateParams)
		{
			NewWnd = (GWindow*) Info->lpCreateParams;
			NewWnd->hWnd = hWnd;
		}
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GWindow *Wnd = (GWindow*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	return DefWindowProc(hWnd, m, a, b);
}

GWin32Class::GWin32Class()
{
	Class.style = CS_DBLCLKS;
	Class.lpfnWndProc = (WNDPROC) Redir;
	Class.cbClsExtra = 0;
	Class.cbWndExtra = 0;
	Class.hInstance = hAppInst();
	Class.hIcon = NULL;
	Class.hCursor = LoadCursor(hAppInst(), MAKEINTRESOURCE(IDC_ARROW));
	Class.hbrBackground = NULL;
	Class.lpszMenuName = NULL;
	Class.lpszClassName = NULL;

	ParentProc = 0;
}

GWin32Class::~GWin32Class()
{
	UnregisterClass(Name(), hAppInst());
	Class.lpszClassName = NULL;
}

bool GWin32Class::Create(char *ClassName)
{
	bool Status = FALSE;

	if (NOT Class.lpszClassName)
	{
		Name(ClassName);
		Class.lpszClassName = ClassName;
		Status = RegisterClass(&Class) != 0;
	}
	else
	{
		Status = TRUE;
	}

	return Status;
}

bool GWin32Class::SubClass(char *ClassName, char *Parent)
{
	bool Status = FALSE;

	if (GetClassInfo(hAppInst(), Parent, &Class))
	{
		Name(ClassName);
		ParentProc = Class.lpfnWndProc;
		Class.lpfnWndProc = (WNDPROC) SubClassRedir;
		Class.lpszClassName = ClassName;
		Status = RegisterClass(&Class) != 0;
	}

	return Status;
}

LRESULT CALLBACK GWin32Class::CallParent(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (NOT ParentProc) return 0;
	return CallWindowProc((FARPROC)ParentProc, hWnd, m, a, b);
}

//////////////////////////////////////////////////////////////////////////////
GWindow *GWindowFromHandle(HWND *hWnd)
{
	return (GWindow*) GetWindowLong(hWnd, GWL_USERDATA);
}

//////////////////////////////////////////////////////////////////////////////
GWindow *GWindow::Capturing = 0;
GWindow *GWindow::Over = 0;
HWND GWindow::hPrevCapture = 0;

GWindow::GWindow()
{
	hWnd = 0;
	WndFlags = GWF_ENABLE | GWF_VISIBLE;
	Pos.ZOff(0, 0);
	Parent = 0;
	Notify = 0;
}

GWindow::~GWindow()
{
	if (Parent)
	{
		Parent->Children.Delete(this);
	}

	GWindow *c;
	while (c = Children.First())
	{
		DeleteObj(c);
	}

	if (hWnd AND IsWindow(hWnd))
	{
		DestroyWindow(hWnd);
	}
	
	hWnd = 0;
	WndFlags = 0;
	if (Over == this)
	{
		Over = 0;
	}
}

HFONT GWindow::GetFont()
{
	if (hWnd)
	{
		return (HFONT) SendMessage(hWnd, WM_GETFONT, 0, 0);
	}
	return 0;
}

void GWindow::SetFont(HFONT hFont)
{
	if (hWnd)
	{
		SendMessage(hWnd, WM_SETFONT, (WPARAM) hFont, 0);
	}
}

bool GWindow::IsOver(VMouse &m)
{
	return	(m.x >= 0) AND
		(m.y >= 0) AND
		(m.x < Pos.X()) AND
		(m.y < Pos.Y());
}

GWindow *GWindow::WindowFromPoint(int x, int y)
{
	for (GWindow *c = Children.First(); c; c = Children.Next())
	{
		if (c->Pos.Overlap(x, y))
		{
			return c;
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

bool GWindow::HandleCapture(GWindow *Wnd, bool c)
{
	bool Status = FALSE;

	if (hWnd)
	{
		if (c)
		{
			if (Capturing)
			{
				Capturing->HandleCapture(NULL, FALSE);
			}

			hPrevCapture = SetCapture(hWnd);
			Status = TRUE;
			Capturing = Wnd;
		}
		else
		{
			if (Capturing)
			{
				ReleaseCapture();
				Capturing = 0;
			}
		}
	}
	else
	{
		if (Parent)
		{
			Status = Parent->HandleCapture(Wnd, c);
		}
	}

	return Status;
}

bool GWindow::Capture(bool c)
{
	return HandleCapture(this, c);
}

bool GWindow::Create(char *Class, int Style, int x, int y, int Sx, int Sy, HWND hParent)
{
	if (NOT (WndFlags & GWF_ENABLE))
	{
		Style |= WS_DISABLED;
	}
	if (WndFlags & GWF_VISIBLE)
	{
		Style |= WS_VISIBLE;
	}

	CreateWindowEx(	((IsDropTarget()) ? WS_EX_ACCEPTFILES : 0),
			Class,
			Name(),
			Style,
			x, y,
			Sx, Sy,
			hParent,
			NULL,
			hAppInst(),
			this);

	return hWnd != NULL;
}

bool GWindow::Create(char *name, GdcRegion &pos)
{
	GObject::Name(name);
	Pos = pos;
	return TRUE;
}

bool GWindow::SetPos(GdcRegion &p, bool Repaint)
{
	bool Status = TRUE;
	Pos = p;
	if (hWnd)
	{
		Status = SetWindowPos(	hWnd,
					NULL,
					Pos.x1,
					Pos.y1,
					Pos.X(),
					Pos.Y(),
					((NOT Repaint) ? SWP_NOREDRAW : 0) |
					SWP_NOZORDER);
	}
	else if (Repaint)
	{
		Invalidate();
	}
	return Status;
}

bool GWindow::Invalidate(GdcRegion *r, bool Repaint)
{
	if (hWnd)
	{
		return InvalidateRect(hWnd, (r) ? &((RECT)*r) : NULL, Repaint);
	}
	else
	{
		GdcRegion Up;
		GWindow *p = this;

		if (r)
		{
			Up = *r;
		}
		else
		{
			Up.Set(0, 0, Pos.X()-1, Pos.Y()-1);
		}

		while (p AND NOT p->Handle())
		{
			Up.Offset(p->Pos.x1, p->Pos.y1);
			p = p->Parent;
		}

		if (p AND p->Handle())
		{
			return p->Invalidate(&Up);
		}
	}
	return FALSE;
}

bool GWindow::InvalidateFrame(GdcRegion *r, bool Repaint)
{
	if (hWnd)
	{
		RedrawWindow(	hWnd,
				NULL,
				NULL,
				RDW_FRAME |
				RDW_INVALIDATE |
				RDW_ALLCHILDREN |
				((Repaint) ? RDW_UPDATENOW : 0));

		return TRUE;
	}

	return FALSE;
}

void
CALLBACK
GWindow::TimerProc(HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
	GView *View = (GView*) idEvent;
	if (View)
	{
		View->OnPulse();
	}
}

void GWindow::SetPulse(int Length)
{
	if (hWnd)
	{
		static UINT Id = 0;

		if (Length > 0)
		{
			Id = SetTimer(hWnd, (UINT) this, Length, (TIMERPROC) TimerProc);
		}
		else
		{
			KillTimer(hWnd, Id);
		}
	}
}

#define SetKeyFlag(v, k, f)		if (GetKeyState(k)&0xFF00) { v |= f; }

int GWindow::OnEvent(GMessage *Msg)
{
	if (hWnd)
	{
		switch (Msg->Msg)
		{
			case WM_NCDESTROY:
			{
				hWnd = 0;
				break;
			}
	 		case WM_CLOSE:
			{
				if (OnRequestClose())
				{
					DestroyWindow(hWnd);
				}
				break;
			}
			case WM_DESTROY:
			{
				break;
			}
			case WM_CREATE:
			{
				OnCreate();
				break;
			}
			case WM_SETFOCUS:
			{
				OnFocus(TRUE);
				break;
			}
			case WM_KILLFOCUS:
			{
				OnFocus(FALSE);
				break;
			}
			case WM_WANT_DIALOG_PROC:
			{
				return (WndFlags & GWF_DIALOG_PROCESSING) != 0;
			}
			case WM_WINDOWPOSCHANGED:
			{
				WINDOWPOS *Info = (LPWINDOWPOS) Msg->b;
				if (Info)
				{
					Pos.ZOff(Info->cx-1, Info->cy-1);
					Pos.Offset(Info->x, Info->y);
				}
				OnPosChange();

				return DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);
				break;
			}
			case WM_CAPTURECHANGED:
			{
				Capturing = 0;
				break;
			}
			case WM_MOUSEENTER:
			{
				if (Msg->a AND NOT TestFlag(WndFlags, GWF_OVER))
				{
					VMouse Ms;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;
					Ms.Down = FALSE;
					Ms.Double = FALSE;

					GWindow *MouseOver = WindowFromPoint(Ms.x, Ms.y);
					if (Capturing)
					{
						if (MouseOver == Capturing)
						{
							Capturing->OnMouseEnter(AdjustMouseInfoForWindow(Ms, Capturing));
						}
					}
					else
					{
						if (Over != MouseOver)
						{
							if (Over)
							{
								Over->OnMouseExit(AdjustMouseInfoForWindow(Ms, Over));
							}

							Over = MouseOver;

							if (Over)
							{
								Over->OnMouseEnter(AdjustMouseInfoForWindow(Ms, Over));
							}
						}
					}

					SetFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEEXIT:
			{
				if (NOT Msg->a AND TestFlag(WndFlags, GWF_OVER))
				{
					VMouse Ms;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;
					Ms.Down = FALSE;
					Ms.Double = FALSE;

					if (Capturing)
					{
						Capturing->OnMouseExit(AdjustMouseInfoForWindow(Ms, Capturing));
					}
					else
					{
						if (Over)
						{
							Over->OnMouseExit(AdjustMouseInfoForWindow(Ms, Over));
							Over = 0;
						}
						else
						{
							OnMouseExit(Ms);
						}
					}

					ClearFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEMOVE:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = Msg->a;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down = ((Msg->a & (MK_LBUTTON|MK_MBUTTON|MK_RBUTTON)) != 0);
				Ms.Double = FALSE;

				GWindow *MouseOver = WindowFromPoint(Ms.x, Ms.y);
				if (Over != MouseOver)
				{
					if (Over)
					{
						Over->OnMouseExit(AdjustMouseInfoForWindow(Ms, Over));
					}

					Over = MouseOver;

					if (Over)
					{
						Over->OnMouseEnter(AdjustMouseInfoForWindow(Ms, Over));
					}
				}

				if (Capturing)
				{
					Capturing->OnMouseMove(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseMove(AdjustMouseInfoForWindow(Ms, Over));
				}
				break;
			}
			case WM_LBUTTONDBLCLK:
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = Msg->a | MK_LEFT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down = Msg->Msg != WM_LBUTTONUP;
				Ms.Double = Msg->Msg == WM_LBUTTONDBLCLK;

				if (Capturing)
				{
					Capturing->OnMouseClick(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(AdjustMouseInfoForWindow(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_RBUTTONDBLCLK:
			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = Msg->a | MK_RIGHT;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down = Msg->Msg != WM_RBUTTONUP;
				Ms.Double = Msg->Msg == WM_RBUTTONDBLCLK;

				if (Capturing)
				{
					Capturing->OnMouseClick(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(AdjustMouseInfoForWindow(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_MBUTTONDBLCLK:
			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP:
			{
				VMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = Msg->a | MK_MIDDLE;
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
 				Ms.Down = Msg->Msg != WM_MBUTTONUP;
				Ms.Double = Msg->Msg == WM_MBUTTONDBLCLK;

				if (Capturing)
				{
					Capturing->OnMouseClick(AdjustMouseInfoForWindow(Ms, Capturing));
				}
				else if (Over)
				{
					Over->OnMouseClick(AdjustMouseInfoForWindow(Ms, Over));
				}
				else
				{
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_KEYDOWN:
			case WM_KEYUP:
			{
				VKey Key;
				Key.c = Msg->a;		// virtual-key code 
				Key.Flags = 0;
				SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
				SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
				SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
				Key.Data = Msg->b;
				Key.Down = Msg->Msg == WM_KEYDOWN;
				Key.IsChar = FALSE;

				OnKey(Key);
				break;
			}
			case WM_CHAR:
			{
				VKey Key;
				Key.c = Msg->a;		// virtual-key code 
				Key.Flags = 0;
				SetKeyFlag(Key.Flags, VK_MENU, MK_ALT);
				SetKeyFlag(Key.Flags, VK_SHIFT, MK_SHIFT);
				SetKeyFlag(Key.Flags, VK_CONTROL, MK_CTRL);
				Key.Data = Msg->b;
				Key.Down = TRUE;
				Key.IsChar = TRUE;

				OnKey(Key);
				break;
			}
			case WM_PAINT:
			{
				GdcGdiScreen DeviceContext(hWnd);
				OnPaint(&DeviceContext);
				break;
			}
			default:
			{
				return DefWindowProc(hWnd, Msg->Msg, Msg->a, Msg->b);
			}
		}
	}

	return 0;
}

bool GWindow::Name(char *n)
{
	bool Status = GObject::Name(n);
	if (hWnd) SetWindowText(hWnd, Name());
	return TRUE;
}

char *GWindow::Name()
{
	return GObject::Name();
}

bool GWindow::Enabled()
{
	if (hWnd)
	{
		return IsWindowEnabled(hWnd);
	}
	return TestFlag(WndFlags, GWF_ENABLE);
}

void GWindow::Enabled(bool e)
{
	if (e) SetFlag(WndFlags, GWF_ENABLE);
	else ClearFlag(WndFlags, GWF_ENABLE);

	if (hWnd)
	{
		EnableWindow(hWnd, e);
	}
	else
	{
		Invalidate();
	}
}

bool GWindow::Visible()
{
	if (hWnd)
	{
		return IsWindowVisible(hWnd);
	}

	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GWindow::Visible(bool v)
{
	if (v) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);

	if (hWnd)
	{
		ShowWindow(hWnd, (v) ? SW_SHOWNORMAL : SW_HIDE);
		Invalidate();
	}
	else
	{
		Invalidate();
	}
}

bool GWindow::Focus()
{
	if (hWnd)
	{
		return GetFocus() == hWnd;
	}
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GWindow::Focus(bool f)
{
	if (f) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);

	if (hWnd)
	{
		SetFocus((f) ? hWnd : GetDesktopWindow());
	}
	else
	{
		Invalidate();
	}
}

GWindow *GWindow::FindReal()
{
	GWindow *p = Parent;
	while (p AND NOT p->Handle())
	{
		p = p->Parent;
	}
	
	if (p AND p->Handle())
	{
		return p;
	}

	return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////
GAppWindow::GAppWindow()
{
	Menu = 0;
}

GAppWindow::~GAppWindow()
{
	if (Menu)
	{
		Menu->Detach();
		DeleteObj(Menu);
	}

	for (GView *v = Views.First(); v; v = Views.Next())
	{
		DeleteObj(v);
	}
}

bool GAppWindow::Visible()
{
	return GWindow::Visible();
}

void GAppWindow::Visible(bool v)
{
	if (v) Pour();
	GWindow::Visible(v);
}

void GAppWindow::Pour()
{
	RECT rcClient;

	if (NOT GetClientRect(hWnd, &rcClient))
	{
		return;
	}

	GdcComplexRegion Client(rcClient);
	GdcComplexRegion Update;

	for (GView *v = Views.First(); v; v = Views.Next())
	{
		GdcRegion OldPos = v->Pos;
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(TRUE);
			}

			if (OldPos != v->Pos)
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->Pos);
			Update.Subtract(&v->Pos);
		}
		else
		{
			// make the view not visible
			v->Visible(FALSE);
		}
	}

	for (int i=0; i<Update.GetSize(); i++)
	{
		Invalidate(Update[i]);
	}
}

int GAppWindow::OnEvent(GMessage *Msg)
{
	switch (Message(Msg))
	{
		case WM_CREATE:
		{
			Pour();
			break;
		}
	 	case WM_CLOSE:
		{
			if (OnRequestClose())
			{
				if (Menu)
				{
					Menu->Detach();
					DeleteObj(Menu);
				}
				ExitApp();
			}
			return 0;
			break;
		}
		case WM_DESTROY:
		{
			break;
		}
		case WM_COMMAND:
		{
			return OnCommand(LOWORD(Msg->a), HIWORD(Msg->a), (HWND) Msg->b);
		}
		case WM_DROPFILES:
		{
			HDROP hDrop = (HDROP) Msg->a;
			if (hDrop)
			{
				List<char> FileNames;
				int Count = DragQueryFile(hDrop, -1, NULL, 0);;
				char FileName[256];

				for (int i=0; i<Count; i++)
				{
					if (DragQueryFile(hDrop, i, FileName, sizeof(FileName)-1) > 0)
					{
						FileNames.Insert(NewStr(FileName));
					}
				}

				OnReceiveFiles(FileNames);

				for (char *n = FileNames.First(); n; n = FileNames.Next())
				{
					DeleteObj(n);
				}
			}
			break;
		}
	}

	int Status = GWindow::OnEvent(Msg);

	switch (Message(Msg))
	{
		case WM_WINDOWPOSCHANGED:
		{
			Pour();
			break;
		}
	}

	return Status;
}

/////////////////////////////////////////////////////////////////////////////
HINSTANCE GApp::hAppInst = 0;
GApp *GApp::TheApp = 0;
char *GApp::CmdLine = 0;
int GApp::CmdShow = 0;

GApp::GApp(HINSTANCE hInst, char *lpCmdLine, int nShow)
{
	CmdShow = nShow;
	TheApp = this;
	hAppInst = hInst;
	CmdLine = lpCmdLine;
	AppWnd = 0;

	hMouseHook = SetWindowsHookEx(WH_MOUSE, (HOOKPROC) MouseHook, NULL, (DWORD) GetCurrentThreadId());

	SystemFont = NEW(GdcWin32Ttf);
	if (SystemFont)
	{
		if (SystemFont->CreateFont(	-10,
						0,
						0,
						0,
						0,
						0,
						0,
						0,
						ANSI_CHARSET,
						OUT_TT_PRECIS,
						CLIP_DEFAULT_PRECIS,
						DEFAULT_QUALITY,
						DEFAULT_PITCH,
						"Ms Sans Serif"))
		{
			SystemFont->Fore(0);
			SystemFont->Transparent(TRUE);
		}
		else
		{
			// error: do what???
		}
	}
}

GApp::~GApp()
{
	DeleteObj(AppWnd);	
	DeleteObj(SystemFont);
	if (hMouseHook)
	{
		UnhookWindowsHookEx(hMouseHook);
	}
	TheApp = 0;
}

void GApp::OnCmdLine(char *Cmd)
{
	if (Cmd AND strlen(Cmd) > 0)
	{
		int Args = 0;
		for (char *c = Cmd; c; )
		{
			Args++;
			c = strchr(c, ' ');
			if (c) c++;
		}

		char **Arg = NEW(char*[Args]);
		if (Arg)
		{
			char *c = Cmd;
			for (int i=0; i<Args; i++)
			{
				Arg[i] = c;
				c = strchr(c, ' ');
				if (c)
				{
					*c = 0;
					c++;
				}
			}

			OnArgs(Args, Arg);
			DeleteArray(Arg);
		}
	}
}

void GApp::OnArgs(int Args, char **Arg)
{
	if (Arg AND AppWnd)
	{
		List<char> Files;
		char File[256] = "";

		for (int i=0; i<Args; i++)
		{
			if (strlen(File) == 0)
			{
				strcpy(File, Arg[i]);
				if (FileExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
			else
			{
				strcat(File, " ");
				strcat(File, Arg[i]);
				if (FileExists(File))
				{
					Files.Insert(NewStr(File));
					File[0] = 0;
				}
			}
		}

		AppWnd->OnReceiveFiles(Files);

		for (char *s = Files.First(); s; s = Files.Next())
		{
			DeleteArray(s);
		}
	}
}

HWND GApp::ActiveDialog = NULL;
void GApp::Run(bool Loop)
{
	MSG Msg;
	while (GetMessage(&Msg, NULL, 0, 0) > 0)
	{
		if (ActiveDialog)
		{
			if (IsDialogMessage(ActiveDialog, &Msg))
			{
				continue;
			}
		}

		TranslateMessage(&Msg);
		DispatchMessage(&Msg);

		if (NOT Loop) break;
	}
}

void GApp::Exit(int Code)
{
	DeleteObj(AppWnd);
	PostQuitMessage(Code);
}

//////////////////////////////////////////////////////////////////////////////
#include <process.h>

uint WINAPI EntryPoint(int i)
{
#ifdef	_MT
	if (i)
	{
		int Status = ((GThread*)i)->Main();
		_endthreadex(Status);
	}
#endif

	return 0;
}

GThread::GThread(bool Sleep, int StackSize)
{
	hThread = 0;
	ThreadId = 0;

#ifdef	_MT
	hThread = (HANDLE) _beginthreadex(NULL,
				StackSize,
				(unsigned int (__stdcall *)(void *)) EntryPoint,
				(LPVOID) this,
				(Sleep) ? CREATE_SUSPENDED : 0,
				&ThreadId);
#else
	MessageBox(NULL, "You are tryed to use a Thread without the multithreaded libs.", "Error", MB_OK);
#endif
}

GThread::~GThread()
{
}

bool GThread::IsExited()
{
	return (ExitCode() != STILL_ACTIVE);
}

void GThread::Run(bool b)
{
	DWORD Status = 0;
	DWORD Error = 0;

	if (b)
	{
		Status = ResumeThread(hThread);
	}
	else
	{
		Status = SuspendThread(hThread);
	}
}

void GThread::Terminate()
{
	if (hThread)
	{
		TerminateThread(hThread, 0);
	}
}

DWORD GThread::ExitCode()
{
	DWORD Code = 0;
	GetExitCodeThread(hThread, &Code);
	return Code;
}

int GThread::Main()
{
	return 0;
}

int GThread::Priority()
{
	int p = 0;
	if (hThread)
	{
		p = GetThreadPriority(hThread);
	}
	return p;
}

void GThread::Priority(int p)
{
	if (hThread)
	{
		SetThreadPriority(hThread, p);
	}
}

//////////////////////////////////////////////////////////////////////////////
int GTrayIcon::Id = 1;

GTrayIcon::GTrayIcon(int icons)
{
	ZeroObj(TrayIcon);
	Val = 0;
	Msg = 0;
	Notify = 0;
	MyId = 0;
	Icons = max(icons, 1);
	hIcon = NEW(HICON[Icons]);
}

GTrayIcon::~GTrayIcon()
{
	Visible(FALSE);

	if (hIcon)
	{
		for (int i=0; i<Icons; i++)
		{
			DeleteObject(hIcon[i]);
		}

		DeleteArray(hIcon);
	}
}

bool GTrayIcon::LoadIcon(char *Str, int Index)
{
	if (hIcon AND Index >= 0 AND Index < Icons)
	{
		hIcon[Index] = ::LoadIcon(hAppInst(), Str);
		return hIcon[Index] != NULL;
	}

	return FALSE;
}

bool GTrayIcon::Visible()
{
	return TrayIcon.cbSize != 0;
}

void GTrayIcon::Visible(bool v)
{
	if (Visible() != v)
	{
		if (v)
		{
			if (Notify AND hIcon)
			{
				ZeroObj(TrayIcon);
				TrayIcon.cbSize = sizeof(TrayIcon);
				TrayIcon.hWnd = Notify->Handle();
				TrayIcon.uID = MyId = Id++;
				TrayIcon.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
				TrayIcon.uCallbackMessage = Msg;
				TrayIcon.hIcon = hIcon[Val];
				strncpy(TrayIcon.szTip, Name(), sizeof(TrayIcon.szTip));

				Shell_NotifyIcon(NIM_ADD, &TrayIcon);
			}
		}
		else
		{
			Shell_NotifyIcon(NIM_DELETE, &TrayIcon);
			ZeroObj(TrayIcon);
		}
	}
}

int GTrayIcon::Value()
{
	return Val;
}

void GTrayIcon::Value(int v)
{
	Val = v;

	if (Visible() AND hIcon)
	{
		ZeroObj(TrayIcon);
		TrayIcon.cbSize = sizeof(TrayIcon);
		TrayIcon.hWnd = Notify->Handle();
		TrayIcon.uID = MyId;
		TrayIcon.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
		TrayIcon.uCallbackMessage = Msg;
		TrayIcon.hIcon = hIcon[Val];
		strncpy(TrayIcon.szTip, Name(), sizeof(TrayIcon.szTip));

		Shell_NotifyIcon(NIM_MODIFY, &TrayIcon);
	}
}

