/*
**	FILE:		GuiToolbar.cpp
**	AUTHOR:		Matthew Allen
**	DATE:		18/7/98
**	DESCRIPTION:	Toolbar classes
**
**	Copyright (C) 1998, Matthew Allen
**		55 Kensignton Rd, Bolwarra, 2320, NSW, Australia
**		fret@ozemail.com.au
*/

#include <stdio.h>

#include "defs.h"
#include "memdev.h"
#include "file.h"
#include "gdc2.h"
#include "gui.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////
bool BltBmpToBmp(HBITMAP hDest,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HBITMAP hSrc,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = FALSE;
	HDC DestDC = CreateCompatibleDC(0);
	HDC SrcDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hDest = SelectObject(DestDC, hDest);
		hSrc = SelectObject(SrcDC, hSrc);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hDest = SelectObject(DestDC, hDest);
		hSrc = SelectObject(SrcDC, hSrc);
	}

	if (DestDC)
	{
		DeleteDC(DestDC);
	}
	if (SrcDC)
	{
		DeleteDC(SrcDC);
	}

	return Status;
}

bool BltBmpToDc(HDC DestDC,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HBITMAP hSrc,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = FALSE;
	HDC SrcDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hSrc = SelectObject(SrcDC, hSrc);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hSrc = SelectObject(SrcDC, hSrc);
	}

	if (SrcDC)
	{
		DeleteDC(SrcDC);
	}

	return Status;
}

bool BltDcToBmp(HBITMAP hDest,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HDC SrcDC,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = FALSE;
	HDC DestDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hDest = SelectObject(DestDC, hDest);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hDest = SelectObject(DestDC, hDest);
	}

	if (DestDC)
	{
		DeleteDC(DestDC);
	}

	return Status;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GImageList::GImageList(int x, int y, GdcBasePrimitives *pDC)
{
	Sx = x;
	Sy = y;
	hImgList = 0;
	pCreate = pDC;
}

GImageList::~GImageList()
{
	if (hImgList)
	{
		ImageList_Destroy(hImgList);
		hImgList = 0;
	}
}

bool GImageList::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	bool Status = GdcDibSection::Create(x, y, Bits, LineLen, KeepData);
	if (Status)
	{
		if (pCreate)
		{
			RGBQUAD *Rgb = Info->bmiColors;

			if (Bits <= 8)
			{
				// create palette system palette
				PALETTEENTRY Entries[256];
				HDC hDC = CreateCompatibleDC(0);
				GetSystemPaletteEntries(hDC, 0, 256, Entries);
				DeleteDC(hDC);

				GdcPalette *Pal = NEW(GdcPalette(0, 256));
				if (Pal)
				{
					for (int i=0; i<256; i++)
					{
						GdcRGB *p = (*Pal)[i];
						if (p)
						{
							Rgb[i].rgbRed = p->R = Entries[i].peRed;
							Rgb[i].rgbGreen = p->G = Entries[i].peGreen;
							Rgb[i].rgbBlue = p->B = Entries[i].peBlue;
							Rgb[i].rgbReserved = 0;
						}
					}
	
					Palette(Pal);
				}
			}

			if (pCreate->GetBits() == 8 AND GetBits() == 8)
			{
				// Set the palette
				HDC hDC = StartDC();
				if (hDC)
				{
					SetDIBColorTable(hDC, 0, 256, Rgb);
					EndDC();
				}

				// gotta remap here
				char Remap[256];
				GdcPalette *SrcPal = pCreate->Palette();
				GdcPalette *DstPal = Palette();
				if (SrcPal AND DstPal)
				{
					// build remap table
					ZeroObj(Remap);
					for (int i=0; i<256; i++)
					{
						GdcRGB *From = (*SrcPal)[i];
						if (From)
						{
							Remap[i] = DstPal->MatchRgb(Rgb24(From->R, From->G, From->B));
						}
					}

					// do blt
					for (int y=0; y<Y(); y++)
					{
						uchar *s = (*pCreate)[y];
						uchar *d = (*this)[y];
						for (int x=0; x<X(); x++)
						{
							*d++ = Remap[*s++];
						}
					}
				}
			}
			else
			{
				// just standard copy blt
				Blt(0, 0, pCreate);
			}
		}

		int Flags = 0;
		switch (GetBits())
		{
			case 8:
			{
				Flags = ILC_COLOR8;
				break;
			}
			case 16:
			{
				Flags = ILC_COLOR16;
				break;
			}
			case 24:
			{
				Flags = ILC_COLOR24;
				break;
			}
			case 32:
			{
				Flags = ILC_COLOR32;
				break;
			}
		}

		int Images = X() / Sx;
		hImgList = ImageList_Create(	Sx,
						Sy,
						Flags | ILC_MASK,
						Images,
						0);
		if (hImgList)
		{
			int Width = ((X() + 15) / 16) * 2;
			uchar *Data = NEW(uchar[Width * pCreate->Y()]);
			if (Data)
			{
				// prepare some 2-byte aligned monochrome data for CreateBitmap
				memset(Data, 0, Width * pCreate->Y());

				COLOUR Key = pCreate->Get(0, 0);
				for (int y = 0; y < pCreate->Y(); y++)
				{
					int Iy = pCreate->Y() - y - 1;
					uchar *d = Data + (y * Width);
					uchar Mask = 0x80;

					for (int x = 0; x < pCreate->X(); x++)
					{
						COLOUR c = pCreate->Get(x, y);
						if (c == Key)
						{
							*d |= Mask;
						}

						Mask >>= 1;
						if (NOT Mask)
						{
							Mask = 0x80;
							d++;
						}
					}
				}
			}

			int Index = -1;
			HDC hDC = StartDC();
			HBITMAP bmp = CreateCompatibleBitmap(hDC, X(), Y());
			HBITMAP mask = CreateBitmap(X(), Y(), 1, 1, Data);
			if (bmp)
			{
				BltDcToBmp(bmp, 0, 0, X(), Y(), hDC, 0, 0, SRCCOPY);
				EndDC();
				Index = ImageList_Add(hImgList, bmp, mask);
				DeleteObject(bmp);
				if (mask) DeleteObject(mask);
			}
		}
	}
	return Status;
}

void GImageList::Update(int Flags)
{
	/*
	IMAGEINFO Info;
	ZeroObj(Info);
	if (ImageList_GetImageInfo(hImgList, 0, &Info))
	{
		HDC hSrc = StartDC();
		HDC hDest = CreateCompatibleDC(hSrc);
		if (hSrc AND hDest)
		{
			HBITMAP Old = SelectObject(hDest, Info.hbmImage);
			
			// set the initial image
			bool Temp = BitBlt(hDest, 0, 0, X(), Y(), hSrc, 0, 0, SRCCOPY);
			
			SelectObject(hDest, Old);
		}

		// clean up DCs
		EndDC();
		if (hSrc)
		{
			DeleteDC(hSrc);
		}
	}
	*/
}

void GImageList::Draw(GdcBasePrimitives *pDest, int Dx, int Dy, int Image, int Flags)
{
	if (Flags & IMGLST_GDC)
	{
		if (pDest)
		{
			COLOUR Key = Get(0, 0);
			GdcRegion r;
			r.ZOff(Sx-1, Sy-1);
			r.Offset(Image * Sx, 0);

			if (Flags & IMGLST_DISABLED)
			{
				COLOUR Low = CBit(pDest->GetBits(), TOOL_LOW);
				COLOUR High = CBit(pDest->GetBits(), TOOL_HIGH);
				
				for (int y=0; y<r.Y(); y++)
				{
					for (int x=0; x<r.X(); x++)
					{
						if (Get(r.x1 + x, r.y1 + y) != Key)
						{
							pDest->Colour(Low);
							pDest->Set(Dx + x, Dy + y);
							pDest->Colour(High);
							pDest->Set(Dx + x + 1, Dy + y + 1);
						}
					}
				}
			}
			else
			{
				for (int y=0; y<r.Y(); y++)
				{
					for (int x=0; x<r.X(); x++)
					{
						COLOUR c = Get(r.x1 + x, r.y1 + y);
						if (c != Key)
						{
							pDest->Colour(c);
							pDest->Set(Dx + x, Dy + y);
						}
					}
				}
			}
		}
	}
	else
	{
		HDC hdc = pDest->StartDC();
		if (hdc)
		{
			int Mode = ILD_NORMAL;
			if (Flags & IMGLST_SELECTED)
			{
				Mode = ILD_BLEND50;
			}
			if (Flags & IMGLST_DISABLED)
			{
				Mode = ILD_FOCUS;
			}

			int Ox, Oy;
			pDest->GetOrigin(Ox, Oy);

			bool Status = ImageList_DrawEx(	hImgList,
							Image,
							hdc,
							Ox + Dx,
							Oy + Dy,
							Sx,
							Sy,
							0, // Background colour CLR_NONE
							TOOL_HIGH, // Foreground colour
							Mode);

			pDest->EndDC();
		}
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolButton::GToolButton(int Bx, int By)
{
	Name("");
	Type = TBT_PUSH;
	Id = IDM_NONE;
	SetDown(FALSE);
	Clicked = FALSE;
	ImgIndex = 0;

	Pos.ZOff(Bx+1, By+1);
	Parent = 0;
}

GToolButton::~GToolButton()
{
}

HPALETTE GetSystemPalette()
{
	HPALETTE hPal = 0;
	LOGPALETTE *Log = (LOGPALETTE*) NEW(uchar[sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * 255)]);
	if (Log)
	{
		Log->palVersion = 0x300;
		Log->palNumEntries = 256;

		HDC hDC = CreateCompatibleDC(0);
		GetSystemPaletteEntries(hDC, 0, 256, Log->palPalEntry);
		DeleteDC(hDC);

		hPal = CreatePalette(Log);
	}

	return hPal;
}

COLOUR Map(GdcBasePrimitives *pDC, COLOUR c)
{
	if (pDC)
	{
		if (pDC->GetBits() <= 8)
		{
			if (pDC->IsScreen())
			{
				c = CBit(24, c);
			}
			else
			{
				HPALETTE hPal = GetSystemPalette();
				if (hPal)
				{
					c = GetNearestPaletteIndex(hPal, c);
					DeleteObject(hPal);
				}
			}
			return c;
		}
		else
		{
			return CBit(pDC->GetBits(), c);
		}
	}
	return 0;
}

void GToolButton::OnPaint(GdcBasePrimitives *pDC)
{
	GToolBar *Par = (GToolBar*) Parent;
	if (Par AND Par->ImgList)
	{
		GdcDibSection *pMemDC = NEW(GdcDibSection);
		if (pMemDC AND pMemDC->Create(Pos.X(), Pos.Y(), GdcD->GetBits()))
		{
			int Px = Pos.X()-1;
			int Py = Pos.Y()-1;

			pMemDC->Colour(Map(pMemDC, TOOL_MED));
			pMemDC->Rectangle();
			if (((Type == TBT_RADIO) OR (Type == TBT_TOGGLE)) AND Down)
			{
				// draw highlight
				pMemDC->Colour(Map(pMemDC, TOOL_HIGH));

				for (int y=0; y<pMemDC->Y(); y++)
				{
					for (int x=y&1; x<pMemDC->X(); x+=2)
					{
						pMemDC->Set(x, y);
					}
				}
			}

			int Off = BORDER_BUTTON + ((Down) ? 1 : 0);
			
			if (Id >= 0)
			{
				// button
				int Flags = IMGLST_GDC;

				if (Enabled())
				{
					// draw normal
					Par->ImgList->Draw(pMemDC, Off, Off, ImgIndex, Flags);
				}
				else
				{
					// draw disabled
					Par->ImgList->Draw(pMemDC, Off, Off, ImgIndex, IMGLST_DISABLED | Flags);
				}

				// draw border
				if (Down)
				{
					pMemDC->Colour(Map(pMemDC, (Down) ? TOOL_HIGH : TOOL_LOW));
					pMemDC->Line(Px, Py, Px, 0);
					pMemDC->Line(Px, Py, 0, Py);

					pMemDC->Colour(Map(pMemDC, (Down) ? TOOL_LOW : TOOL_HIGH));
					pMemDC->Line(0, 0, 0, Py);
					pMemDC->Line(0, 0, Px, 0);
				}
			}
			else
			{
				// separator
				if (X() > Y())
				{
					int c = Y()/2-1;
					pMemDC->Colour(Map(pMemDC, TOOL_LOW));
					pMemDC->Line(2, c, Px-2, c);
					pMemDC->Colour(Map(pMemDC, TOOL_HIGH));
					pMemDC->Line(2, c+1, Px-2, c+1);
				}
				else
				{
					int c = X()/2-1;
					pMemDC->Colour(Map(pMemDC, TOOL_LOW));
					pMemDC->Line(c, 2, c, Py-2);
					pMemDC->Colour(Map(pMemDC, TOOL_HIGH));
					pMemDC->Line(c+1, 2, c+1, Py-2);
				}
			}
			pDC->Blt(0, 0, pMemDC);
		}

		DeleteObj(pMemDC);
	}
}

void GToolButton::Value(int b)
{
	switch (Type)
	{
		case TBT_PUSH:
		{
			// do nothing... can't set value
			break;
		}

		case TBT_TOGGLE:
		{
			SetDown(b);
			Invalidate();
			break;
		}

		case TBT_RADIO:
		{
			if (Parent AND b)
			{
				// Clear any other radio buttons that are down
				for (	GToolButton *But = (GToolButton*) Parent->Children.Find(this);
					But AND But->Id >= 0;
					But = (GToolButton*) Parent->Children.Next())
				{
					if (	But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(FALSE);
						But->Invalidate();
					}
				}

				for (	But = (GToolButton*) Parent->Children.Find(this);
					But AND But->Id >= 0;
					But = (GToolButton*) Parent->Children.Prev())
				{
					if (	But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(FALSE);
						But->Invalidate();
					}
				}
			}

			SetDown(b);
			Invalidate();
			break;
		}
	}
}

void GToolButton::OnCommand(VMouse &m)
{
	if (Parent)
	{
		((GToolBar*) Parent)->PostCommand(Id);
	}
}

void GToolButton::SendNotify(int Flags)
{
	GWindow *To = (Notify) ? Notify : Parent;
	if (To)
	{
		To->OnEvent(&GMessage(WM_CHANGE, (int)((GControl*)this), Flags));
	}
}

void GToolButton::OnMouseClick(VMouse &m)
{
	if (Id >= 0 AND Enabled())
	{
		switch (Type)
		{
			case TBT_PUSH:
			{
				bool Old = Down;

				Clicked = m.Down;
				Capture(m.Down);

				if (Old AND IsOver(m))
				{
					if (m.Flags & MK_LEFT)
					{
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}

				SetDown(m.Down);
				if (Old != Down)
				{
					Invalidate();
				}
				break;
			}

			case TBT_TOGGLE:
			{
				if (m.Down)
				{
					if (m.Left())
					{
						Value(NOT Down);
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}
				break;
			}

			case TBT_RADIO:
			{
				if (m.Down)
				{
					if (NOT Down AND m.Left())
					{
						Value(TRUE);
						OnCommand(m);
					}
					SendNotify(m.Flags);
				}
				break;
			}
		}
	}
}

void GToolButton::OnMouseEnter(VMouse &m)
{
	if (Clicked)
	{
		SetDown(TRUE);
		Invalidate();
	}
	else
	{
		((GToolBar*) Parent)->PostDescription(this, Name());
	}
}

void GToolButton::OnMouseExit(VMouse &m)
{
	if (Clicked)
	{
		SetDown(FALSE);
		Invalidate();
	}
	else
	{
		((GToolBar*) Parent)->PostDescription(this, "");
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolBar::GToolBar()
{
	Name("LGI_Toolbar");
	Bx = By = 0;
	Sx = Sy = 10;
	Vertical = FALSE;

	LastIndex = 0;
	ImgList = 0;
}

GToolBar::~GToolBar()
{
	DeleteObj(ImgList);
}

bool GToolBar::Pour(GdcComplexRegion &r)
{
	if (ImgList)
	{
		int PosX = BORDER_SHADE + BORDER_SPACER;
		int PosY = BORDER_SHADE + BORDER_SPACER;
		int SrcX = 0;
		int SrcY = 0;
		int EndX = 0;
		int EndY = 0;

		GdcRegion ButPos;
		GControl *But = (GControl*) Children.First();
		while (But)
		{
			ButPos = But->GetPos();

			if (But->GetId() == IDM_SEPARATOR)
			{
				// separator
				if (Vertical)
				{
					ButPos.ZOff(Bx+(2*BORDER_BUTTON)-1, BORDER_SEPARATOR+1);
				}
				else
				{
					ButPos.ZOff(BORDER_SEPARATOR+1, By+(2*BORDER_BUTTON)-1);
				}
			}
			else if (But->GetId() >= 0)
			{
				// other control
			}

			ButPos.Offset(PosX - ButPos.x1, PosY - ButPos.y1);

			if (But->GetId() == IDM_BREAK)
			{
				ButPos.ZOff(0, 0);
				if (Vertical)
				{
					PosX = EndX;
					PosY = BORDER_SHADE + BORDER_SPACER;
				}
				else
				{
					PosX = BORDER_SHADE + BORDER_SPACER;
					PosY = EndY;
				}
			}
			else
			{
				if (Vertical)
				{
					EndX = max(EndX, ButPos.x2 + 1);
					PosY = ButPos.y2 + 1;
				}
				else
				{
					PosX = ButPos.x2 + 1;
					EndY = max(EndY, ButPos.y2 + 1);
				}
			}

			But->SetPos(ButPos);
			But = (GControl*) Children.Next();
		}

		if (Vertical)
		{
			PosX = EndX;
		}
		else
		{
			PosY = EndY;
		}

		Sx = PosX + BORDER_SHADE + BORDER_SPACER;
		Sy = PosY + BORDER_SHADE + BORDER_SPACER;
		Pos.ZOff(Sx-1, Sy-1);

		GdcRegion *Best = FindLargest(r);
		if (Best)
		{
			Pos.Offset(Best->x1, Best->y1);
			Pos.Bound(Best);
			SetPos(Pos);

			return TRUE;
		}
	}

	return FALSE;
}

int GToolBar::PostCommand(int Cmd)
{
	if (GWindow::Parent)
	{
		return GWindow::Parent->OnEvent(&GMessage(WM_COMMAND, MAKELONG(Cmd, 0), (int) hWnd));
	}
	return 0;
}

int GToolBar::PostDescription(GWindow *Ctrl, char *Text)
{
	if (GWindow::Parent)
	{
		return GWindow::Parent->OnEvent(&GMessage(WM_DESCRIBE, (int) Ctrl, (int) Text));
	}
	return 0;
}

int GToolBar::X()
{
	return Sx;
}

int GToolBar::Y()
{
	return Sy;
}

int GToolBar::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		case WM_CHANGE:
		{
			return GWindow::Parent->OnEvent(Msg);
		}
	}

	return GWindow::OnEvent(Msg);
}

void GToolBar::OnPaint(GdcBasePrimitives *pScr)
{
	GdcRegion r(0, 0, X()-1, Y()-1);

	DrawMoveShallowEdge(pScr, r, RAISED);
	pScr->Colour(Map(pScr, TOOL_MED));
	pScr->Rectangle(&r);

	for (GWindow *But = Children.First(); But; But = Children.Next())
	{
		GdcRegion r = But->GetPos();
		pScr->SetOrigin(-r.x1, -r.y1);
		But->OnPaint(pScr);
	}
}

void GToolBar::OnMouseClick(VMouse &m)
{
}

void GToolBar::OnMouseEnter(VMouse &m)
{
}

void GToolBar::OnMouseExit(VMouse &m)
{
}

void GToolBar::OnMouseMove(VMouse &m)
{
}

bool GToolBar::SetBitmap(char *File, int bx, int by)
{
	bool Status = FALSE;
	GdcBasePrimitives *pDC = LoadDC(File);
	Status = SetDC(pDC, bx, by);
	DeleteObj(pDC);
	return Status;
}

bool GToolBar::SetDC(GdcBasePrimitives *pNewDC, int bx, int by)
{
	if (pNewDC)
	{
		DeleteObj(ImgList);
		ImgList = NEW(GImageList(bx, by, pNewDC));
		if (	ImgList AND
			ImgList->Create(pNewDC->X(), pNewDC->Y(), GdcD->GetBits()))
		{
			COLOUR Key = ImgList->Get(0, 0);

			/*
			for (int y=0; y<pColour->Y(); y++)
			{
				for (int x=0; x<pColour->X(); x++)
				{
					COLOUR c = pColour->Get(x, y);
					if (c == Key)
					{
						// tranparent
						pMask->Colour(-1);
						pColour->Colour(0);
						pColour->Set(x, y);
					}
					else
					{
						// colour
						pMask->Colour(0);
					}

					pMask->Set(x, y);
				}
			}
			*/

			Bx = bx;
			By = by;
			return TRUE;
		}
	}
	return FALSE;
}

GToolButton *GToolBar::AppendButton(char *Tip, int Id, int Type, int Enabled, int IconId)
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->Name(Tip);
		But->Id = Id;
		But->Type = Type;
		But->Parent = this;
		But->Enabled(Enabled);
		
		if (IconId >= 0)
		{
			But->ImgIndex = IconId;
		}
		else
		{
			But->ImgIndex = LastIndex++;
		}
		Children.Insert(But);
	}
	return But;
}

bool GToolBar::AppendSeparator()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->Id = IDM_SEPARATOR;
		But->Parent = this;
		Children.Insert(But);
		return TRUE;
	}
	return FALSE;
}

bool GToolBar::AppendBreak()
{
	GToolButton *But = NEW(GToolButton(Bx, By));
	if (But)
	{
		But->Id = IDM_BREAK;
		But->Parent = this;
		Children.Insert(But);
		return TRUE;
	}
	return FALSE;
}

bool GToolBar::AppendControl(GControl *Ctrl)
{
	bool Status = FALSE;
	if (Ctrl)
	{
		Ctrl->SetParent(this);
		Children.Insert(Ctrl);
		Status = TRUE;
	}
	return Status;
}

void GToolBar::Empty()
{
	for (GWindow *But = Children.First(); But; But = Children.Next())
	{
		DeleteObj(But);
	}
}

